unit author;

{$ifdef fpc}
  {$mode objfpc}{$H+}
{$endif}

interface

uses
  {$ifdef Unix}
    Unix,
    LCLType,
  {$else}
    Windows,
  {$endif}
  {$ifdef fpc}
    LConvEncoding,
    LResources,
    LCLIntf,
  {$endif}
  SysUtils,
  Forms,
  Controls,
  StdCtrls,
  Graphics,
  Classes,
  ExtCtrls,
  Buttons,
  copyrite,
  Hexagon,
  WolfUtil,
  zusatz;

type

  { TAuthorForm }

  TAuthorForm            = class(TForm)
    Author               : TLabel;
    AutorVal             : TLabel;
    Anschrift            : TLabel;
    Strasse              : TLabel;
    Ort                  : TLabel;
    Telefon              : TLabel;
    TelefonRegion        : TLabel;
    TelefonVal           : TLabel;
    Telefax              : TLabel;
    TelefaxVal           : TLabel;
    eMail                : TLabel;
    eMailVal             : TLabel;
    Internet             : TLabel;
    InternetVal          : TLabel;
    Compiler             : TLabel;
    LblCompiler          : TLabel;
    Pflege               : TLabel;
    Datum                : TLabel;
    Hinweis              : TLabel;
    Image                : TImage;
    Button               : TButton;
    procedure              FormCreate(Sender: TObject);
    procedure              InternetValClick(Sender: TObject);
    procedure              ToFront(Sender: TObject);
    procedure              FormClose(Sender: TObject; var CloseAction: TCloseAction);
    procedure              eMailValClick(Sender: TObject);
  end;

var
  AuthorForm : TAuthorForm;

implementation

const
  {$ifdef Unix}
    ReleaseTag   =    9;
    ReleaseMonat =    5;
    ReleaseJahr  = 2011;
  {$else}
    ReleaseTag   =   16;
    ReleaseMonat =    5;
    ReleaseJahr  = 2011;
  {$endif}

{$ifdef fpc}
  {$R *.lfm}
{$else}
  {$R *.dfm}
{$endif}

procedure TAuthorForm.FormCreate(Sender: TObject);
  begin
    case SysLocale.PriLangID of
      Lang_German:
        Datum.Caption := CP1252toSystem('Mrz 1996 bis ') + IntToStr(ReleaseTag)
                         + '. ' + GermanLongMonthNames[ReleaseMonat]
                         + ' '  + IntToStr(ReleaseJahr);
    else
      begin
        Datum.Caption     := 'March 1996 until '
                             + EnglishLongMonthNames[ReleaseMonat]
                             + ' ' + Ordinal(ReleaseTag)
                             + ' ' + IntToStr(ReleaseJahr);
        Caption           := 'Hextension program''s author';
        Author.Caption    := 'Author';
        Anschrift.Caption := 'Address';
        Telefon.Caption   := 'Telephone / Answ. mach.';
        Pflege.Caption    := 'Program maintained';
        Hinweis.Caption   :=
          'I came to know the game in a box named "Take it easy". ' +
          'It was awkward to calculate the scores and the winner ' +
          'on ending the game. Therefore I brought the sense of the game ' +
          'to the computer screen and step by step I improved the handling.';
        InternetVal.Hint  := 'Check for an upgrade and visit the web site';
      end;
    end;
    AutorVal.Caption      := CopyrightName;
    Strasse.Caption       := CopyrightStrasse;
    Ort.Caption           := CopyrightOrt;
    TelefonRegion.Caption := CopyrightTelefonRegion;
    TelefonVal.Caption    := CopyrightTelefon;
    TelefaxVal.Caption    := CopyrightTelefax;
    eMailVal.Caption      := CopyrighteMail;
    InternetVal.Caption   := CopyrightInternet;
    LblCompiler.Caption   := CompilerName;
    LoadFormFromIni(Self, Name);
  end;

procedure TAuthorForm.ToFront(Sender: TObject);
  begin
    BringToFront;
  end;

procedure TAuthorForm.FormClose(Sender: TObject; var CloseAction: TCloseAction);
  begin
    if CloseAction = caFree then;
    SaveFormToIni(Self, Name);
  end;

procedure TAuthorForm.eMailValClick(Sender: TObject);
  begin
    {$ifdef Unix}
      fpSystem('xdg-email --subject "Unix TakeEasy" mailto:'
               + (Sender as tLabel).Caption);
    {$else}
      {$ifdef fpc}
        OpenURL('mailto:' + (Sender as tLabel).Caption
                + ' ? subject=Lazarus%20TakeEasy');
      {$else}
        Exec(0, 'mailto:' + (Sender as tLabel).Caption
             + ' ? subject=Delphi%20TakeEasy', '', sw_ShowNormal);
      {$endif}
    {$endif}
  end;

procedure TAuthorForm.InternetValClick(Sender: TObject);
  const   MyName = 'TakeEasy';
  var     b, ok  : Boolean;
          s      : String;
  begin
    Screen.Cursor := crHourGlass;
    {$ifdef Unix}
      s           := MyName;
    {$else}
      s           := MyName + '.exe';
    {$endif}
    b             := false;
    ok            := InetUpdate(s, ReleaseJahr, ReleaseMonat, ReleaseTag, b);
    Screen.Cursor := crDefault;
    if not ok then
      case SysLocale.PriLangID of
        Lang_German:
          MsgBox(CP1252ToSystem(
            'Konnte die Version dieses Programmes nicht prfen!'),
            Caption, mb_ok);
      else
        MsgBox('Could not check the release of this program!', Caption, mb_ok);
      end
    else
      begin
        if b then
          case SysLocale.PriLangID of
            Lang_German:
              b := MsgBox('Es gibt eine neue Version dieses Programmes! '
                          + CP1252ToSystem('Mchten Sie diese jetzt laden?'),
                          Caption, mb_YesNo) = idYes;
          else
            b := MsgBox('There is a new release of this program! '
                        + 'Would you like to download it now?',
                        Caption, mb_YesNo) = idYes;
          end
        else
          case SysLocale.PriLangID of
            Lang_German:
              MsgBox('Dieses Programm ist auf dem neuesten Stand.',
                     Caption, mb_ok);
          else
            MsgBox('This program is the latest release.', Caption, mb_ok);
          end;
      end;
    Screen.Cursor := crHourGlass;
    {$ifdef fpc}
      if b then
        OpenURL(s)
      else
        OpenURL((Sender as tLabel).Caption);
    {$else}
      if b then
        Exec(Handle, s, '', sw_ShowNormal)
      else
        Exec(Handle, (Sender as tLabel).Caption, '', sw_ShowNormal);
    {$endif}
    Screen.Cursor := crDefault;
  end;

end.
