unit chat;

{$ifdef fpc}
  {$mode objfpc}{$H+}
{$endif}

{$ifdef MSWINDOWS}
  { Delphi 2005 and newer }
  {$define DelphiMin4}
{$endif}

{$ifdef Ver140}
  { Kylix 3 }
  {$define DelphiMin4}
{$endif}

{$ifdef Ver130}
  { Delphi 5 }
  {$define DelphiMin4}
{$endif}

{$ifdef Ver120}
  { Delphi 4 }
  {$define DelphiMin4}
{$endif}

{$ifdef FPC}
  {$define DelphiMin4}
{$endif}

interface

uses
  {$ifdef fpc}
    LResources,
  {$else}
    IniFiles,
  {$endif}
  {$ifdef Unix}
    LIBC,
    WolfUtil,
  {$else}
    Windows,
    WinSock,
  {$endif}
  SysUtils,
  Classes,
  Graphics,
  Controls,
  Forms,
  Dialogs,
  StdCtrls,
  Zusatz;

type
  tChatMessage    = Class
                      Ip    : LongInt;
                      StdIp : LongInt;
                      Txt   : String;
                    end;

  TChatForm       = class(TForm)
    ComboBox      : TComboBox;
    LblTo         : TLabel;
    Edit          : TEdit;
    BtnSenden     : TButton;
    Memo          : TMemo;
    BtnClear      : TButton;
    procedure       FormResize(Sender: TObject);
    procedure       FormCreate(Sender: TObject);
    procedure       BtnClearClick(Sender: TObject);
    procedure       BtnSendenClick(Sender: TObject);
    procedure       ChatChange(Sender: TObject);
    procedure       FormClose(Sender: TObject; var CloseAction: TCloseAction);
    procedure       FormPaint(Sender: TObject);
    procedure       FormDestroy(Sender: TObject);
  private
    ComboBoxWidth : Integer;
    Start         : Boolean;
    procedure       NewSize;
    procedure       ResizeHosts;
  public
    Running       : Boolean;
    Messages      : tList;
    Procedure       Chat(ChatIp : LongInt; ChatStdIp : LongInt; ChatText : String);
    procedure       AddHost(s : String);
    procedure       DeleteHost(s : String);
    procedure       Schliessen;
  end;

var
  ChatForm: TChatForm;

implementation

{$ifdef fpc}
  {$R *.lfm}
{$else}
  {$R *.dfm}
{$endif}

uses
  Hexagon,
  Thread,
  Main;

Procedure tChatForm.Chat(ChatIp : LongInt; ChatStdIp : LongInt; ChatText : String);
  var s, t : String;
      i    : Integer;
      m    : tChatMessage;
  begin
    case SysLocale.PriLangID of
      Lang_German:
        s := 'Von';
    else
      s := 'From';
    end;
    i := Messages.Count - 1;
    while i >= 0 do
      begin
        m := tChatMessage(Messages[i]);
        if m.StdIp = ChatStdIp then
          begin
            if m.Ip <> ChatIp then
              if m.Txt = ChatText then
                s := '';
            i := -1;
          end
        else
          dec(i);
      end;
    if s > '' then
      begin
        t := Inet_NtoA(tInAddr(ChatIp));
        if ComboBox.Items.IndexOf(t) < 0 then
          ComboBox.Items.Add(t);
        Memo.Lines.Add(s + ' ' + t + ': ' + ChatText);
        BringToFront;
        m := tChatMessage.Create;
        if m <> Nil then
          begin
            m.Ip    := ChatIp;
            m.StdIp := ChatStdIp;
            m.Txt   := ChatText;
            i := Messages.Add(m) - 1;
            while i >= 0 do
              begin
                m := tChatMessage(Messages[i]);
                if m.StdIp = ChatStdIp then
                  begin
                    Messages.Delete(i);
                    i := -1;
                  end
                else
                  dec(i);
              end;
          end;
      end;
  end;

procedure TChatForm.FormResize(Sender: TObject);
  begin
    NewSize;
  end;

procedure TChatForm.NewSize;
  begin
    Edit.Width  := ClientWidth  - Edit.Left - BtnClear.Left;
    Memo.Height := ClientHeight - Memo.Top;
    Memo.Width  := ClientWidth;
    SaveFormToIni(Self, Name);
  end;

procedure TChatForm.FormCreate(Sender: TObject);
  begin
    case SysLocale.PriLangID of
      Lang_German:
    else
      Caption           := 'Messages';
      LblTo.Caption     := 'To';
      BtnSenden.Caption := '&Send';
      BtnClear.Caption  := '&Clear';
    end;
    Start          := True;
    ComboBoxWidth  := ComboBox.Width;
    Edit.MaxLength := sizeof(tChatText) - 1;
  end;

procedure TChatForm.BtnClearClick(Sender: TObject);
  begin
    Memo.Clear;
  end;

procedure TChatForm.BtnSendenClick(Sender: TObject);
  var     Telegramm : tTelegramm;
          IP        : LongInt;
  begin
    Telegramm.Kennung := TelegrammChat;
    if MainForm.MyIps <> Nil then
      begin
        IP               := GetIpAddr(ComboBox.Text);
        Telegramm.ChatIp := MainForm.GetStdIP(IP, MainForm.MyIps);
        StrPCopy(Telegramm.ChatText, Edit.Text);
        MainForm.SendTelegramm(Telegramm, IP);
      end;
    case SysLocale.PriLangID of
      Lang_German:
        Memo.Lines.Add('An ' + ComboBox.Text + ': ' + Edit.Text);
    else
      Memo.Lines.Add('To ' + ComboBox.Text + ': ' + Edit.Text);
    end;
  end;

procedure TChatForm.ChatChange(Sender: TObject);
  begin
    BtnSenden.enabled := (ComboBox.Text > '') and (Edit.Text > '');
  end;

procedure TChatForm.FormClose(Sender: TObject; var CloseAction: TCloseAction);
  begin
    if CloseAction = caFree then;
    Schliessen;
  end;

procedure TChatForm.Schliessen;
  begin
    Running := False;
    MainForm.ShowPanel(True);
    SaveFormToIni(Self, Name);
    MainForm.ChildSchliessen(Self);
    Release;
  end;

procedure TChatForm.FormPaint(Sender: TObject);
  var     i         : Integer;
          {$ifndef fpc}
            IniFile : tIniFile;
            L, T    : Integer;
          {$endif} 
  begin
    if Start then
      begin
        Start   := False;
        Running := True;
        {$ifdef fpc}
          LoadFormFromIni(Self);
        {$else}
          IniFile := tIniFile.Create(IniName);
          L       := (MainForm.ClientWidth  - Width) div 2;
          T       := (MainForm.ClientHeight - MainForm.ToolBar.Height
                      - Height) div 2;
          MoveFormTo(Self,
                     IniFile.ReadInteger(Name, 'Left',   L),
                     IniFile.ReadInteger(Name, 'Top',    T),
                     IniFile.ReadInteger(Name, 'Width',  Width),
                     IniFile.ReadInteger(Name, 'Height', Height),
                     tWindowState(
                       IniFile.ReadInteger(
                         Name, 'WindowState', ord(WindowState))));
          IniFile.Free;
        {$endif}
        for i := 0 to ComboBox.Items.Count - 1 do
          if Canvas.TextWidth(ComboBox.Items[i]) + ComboBoxWidth
             > ComboBox.Width
          then
            ComboBox.Width := Canvas.TextWidth(ComboBox.Items[i])
                              + ComboBoxWidth;
        ResizeHosts;
      end;
  end;

procedure TChatForm.ResizeHosts;
  begin
    Edit.Left      := ComboBox.Left  + ComboBox.Width  + BtnClear.Left;
    BtnSenden.Left := Edit.Left;
    {$ifdef DelphiMin4}
      if BtnSenden.Left + BtnSenden.Width + BtnClear.Left - ClientWidth + Width
         > Constraints.MinWidth
      then
        Constraints.MinWidth := BtnSenden.Left + BtnSenden.Width
          + BtnClear.Left - ClientWidth + Width;
    {$endif}
    NewSize;
  end;

procedure TChatForm.AddHost(s : String);
  begin
    if ComboBox.Items.IndexOf(s) < 0 then
      begin
        ComboBox.Items.Add(s);
        if not Start then
          if Canvas.TextWidth(s) + ComboBoxWidth > ComboBox.Width then
            begin
              ComboBox.Width := Canvas.TextWidth(s) + ComboBoxWidth;
              ResizeHosts;
            end;
      end;
  end;

procedure TChatForm.DeleteHost(s : String);
  var i : Integer;
  begin
    i := ComboBox.Items.IndexOf(s);
    if i >= 0 then
      ComboBox.Items.Delete(i);
  end;

procedure TChatForm.FormDestroy(Sender: TObject);
  begin
    Messages.free;
  end;

end.
