unit hexagon;

{$ifdef fpc}
  {$mode objfpc}{$H+}
{$endif}

interface

uses
  {$ifdef fpc}
    LResources,
  {$else}
    Windows,
  {$endif}
  Classes,
  SysUtils,
  Forms,
  Controls,
  Graphics,
  Dialogs,
  StdCtrls,
  ExtCtrls;

Const
  GesamtSteine  = 3 * 3 * 3;
  ZuSetzen      = 19;
  HexName       = 'Hexagon';
  OptimumPkte   = 307;
  OptimumAnz    = 16;
  IniChild      = 'Child';
  IniChat       = 'Chat';
  IniState      = 'State';
  IniTop        = 'Top';
  IniLeft       = 'Left';
  IniHeight     = 'Height';
  IniWidth      = 'Width';
  IniColors     = 'Colors';
  IniHighscore  = 'HighScore';
  IniPoints     = 'Points';
  IniNames      = 'Names';
  IniPlayer     = 'Player';
  IniLine       = 'Line';
  IniBackground = 'Background';
  IniShadow     = 'Shadow';
  IniLight      = 'Light';
  IniStartMenue = 'StartMenu';

Type
  thxStyle      = (hxEmpty, hxHidden, hxVisible, hxHighlight);
  thxNummer     = 0 .. GesamtSteine - 1;
  tRichtung     = (rHoch, rAuf, rAb);
  tRGB          = (RedVal, GreenVal, BlueVal);
  tBewertung    = (bewPunkte, bewAnfang, bewEnde, bewDemoPunkte, bewDemoEnde);
  tLineColors   = Array[1 .. 9] of tColor;

  tHexagon      =
    class(TPaintBox)
      Constructor  Create(AOwner: TComponent); override;
      procedure    Paint; override;
      procedure    Occupy(Hex : tHexagon);
      procedure    FreeOccupation;
      function     GetOccupation : tHexagon;
      procedure    Init(Nummer : Word; InitStyle : thxStyle);
      procedure    LayAt(NewTop, NewLeft : Integer);
      procedure    HexResize(Hoehe, Weite : LongInt);
      procedure    NewColors(NewLineColors : tLineColors;
                             NewBackground, NewLight, NewShadow : tColor);
      function     GetLine(Richtung : tRichtung) : word;
    private
      hxStyle    : thxStyle;
      hxNummer   : thxNummer;
      Occupying  : tHexagon;
      LineColors : tLineColors;
      Background,
      Light,
      Shadow     : tColor;
      procedure    ResizeLinks( Hoehe, Weite : LongInt; HexName : String);
      procedure    ResizeRechts(Hoehe, Weite : LongInt; HexName : String);
      procedure    SetStyle(Style : thxStyle);
      procedure    SetNummer(Nummer : thxNummer);
    published
      property     Style  : thxStyle  read hxStyle  write SetStyle;
      property     Nummer : thxNummer read hxNummer write SetNummer;
    end;

  TPkteLabel =
    class(tLabel)
      Constructor  Create(AOwner: TComponent); override;
      function     Calculate(Hex1, Hex2, Hex3, Hex4, Hex5 : tHexagon;
                             Richtung  : tRichtung;
                             Bewertung : tBewertung) : word;
      procedure    NewColors;
    private
      LineColors : tLineColors;
      Background,
      Light      : tColor;
      Line       : Integer;
      procedure    AddLine(Hex : tHexagon; var Steine : Word;
                           Richtung : tRichtung);
    end;

  tBar =
    class(tPaintBox)
      Constructor  Create(AOwner: TComponent); override;
      procedure    Paint; override;
      procedure    HiLight;
      procedure    UnLight;
      procedure    Down;
      function     Up : Boolean;
      procedure    Mark;
      procedure    NewColors(NewLineColors : tLineColors;
                             NewBackground, NewLight, NewShadow : tColor);
    private
      State      : (None, HiLighted, Pressed);
      Marked     : Boolean;
      LineColors : tLineColors;
      Background,
      Light,
      Shadow     : tColor;
    end;

  tOptimumListe =
    array[1 .. OptimumAnz, 0 .. ZuSetzen - 1] of
      record
        Hexagon,
        Wert    : Word;
      end;

const
  OptimumListe  : tOptimumListe =
{1} (((Hexagon: 113; Wert:  4),
      (Hexagon: 124; Wert: 13),
      (Hexagon: 135; Wert: 19),
      (Hexagon: 212; Wert: 23),
      (Hexagon: 223; Wert:  5),
      (Hexagon: 234; Wert: 11),
      (Hexagon: 245; Wert: 26),
      (Hexagon: 311; Wert: 12),
      (Hexagon: 322; Wert: 21),
      (Hexagon: 333; Wert:  0),
      (Hexagon: 344; Wert: 15),
      (Hexagon: 355; Wert: 24),
      (Hexagon: 421; Wert: 14),
      (Hexagon: 432; Wert: 20),
      (Hexagon: 443; Wert:  8),
      (Hexagon: 454; Wert: 17),
      (Hexagon: 531; Wert: 10),
      (Hexagon: 542; Wert: 25),
      (Hexagon: 553; Wert:  7)),

{2}  ((Hexagon: 113; Wert:  4),
      (Hexagon: 124; Wert: 16),
      (Hexagon: 135; Wert: 10),
      (Hexagon: 212; Wert: 23),
      (Hexagon: 223; Wert:  8),
      (Hexagon: 234; Wert: 11),
      (Hexagon: 245; Wert: 14),
      (Hexagon: 311; Wert: 21),
      (Hexagon: 322; Wert: 24),
      (Hexagon: 333; Wert:  0),
      (Hexagon: 344; Wert: 12),
      (Hexagon: 355; Wert: 15),
      (Hexagon: 421; Wert: 26),
      (Hexagon: 432; Wert: 20),
      (Hexagon: 443; Wert:  5),
      (Hexagon: 454; Wert: 17),
      (Hexagon: 531; Wert: 19),
      (Hexagon: 542; Wert: 22),
      (Hexagon: 553; Wert:  7)),

{3}  ((Hexagon: 113; Wert:  4),
      (Hexagon: 124; Wert: 22),
      (Hexagon: 135; Wert: 10),
      (Hexagon: 212; Wert: 14),
      (Hexagon: 223; Wert:  5),
      (Hexagon: 234; Wert: 20),
      (Hexagon: 245; Wert: 17),
      (Hexagon: 311; Wert: 21),
      (Hexagon: 322; Wert: 12),
      (Hexagon: 333; Wert:  0),
      (Hexagon: 344; Wert: 24),
      (Hexagon: 355; Wert: 15),
      (Hexagon: 421; Wert: 23),
      (Hexagon: 432; Wert: 11),
      (Hexagon: 443; Wert:  8),
      (Hexagon: 454; Wert: 26),
      (Hexagon: 531; Wert: 19),
      (Hexagon: 542; Wert: 16),
      (Hexagon: 553; Wert:  7)),

{4}  ((Hexagon: 113; Wert:  4),
      (Hexagon: 124; Wert: 22),
      (Hexagon: 135; Wert: 10),
      (Hexagon: 212; Wert: 23),
      (Hexagon: 223; Wert:  5),
      (Hexagon: 234; Wert: 20),
      (Hexagon: 245; Wert: 17),
      (Hexagon: 311; Wert: 12),
      (Hexagon: 322; Wert: 21),
      (Hexagon: 333; Wert:  0),
      (Hexagon: 344; Wert: 24),
      (Hexagon: 355; Wert: 15),
      (Hexagon: 421; Wert: 13),
      (Hexagon: 432; Wert: 19),
      (Hexagon: 443; Wert:  7),
      (Hexagon: 454; Wert: 25),
      (Hexagon: 531; Wert: 11),
      (Hexagon: 542; Wert: 26),
      (Hexagon: 553; Wert:  8)),

{5}  ((Hexagon: 113; Wert:  4),
      (Hexagon: 124; Wert: 25),
      (Hexagon: 135; Wert: 10),
      (Hexagon: 212; Wert: 22),
      (Hexagon: 223; Wert:  7),
      (Hexagon: 234; Wert: 19),
      (Hexagon: 245; Wert: 13),
      (Hexagon: 311; Wert: 12),
      (Hexagon: 322; Wert: 24),
      (Hexagon: 333; Wert:  0),
      (Hexagon: 344; Wert: 21),
      (Hexagon: 355; Wert: 15),
      (Hexagon: 421; Wert: 17),
      (Hexagon: 432; Wert: 20),
      (Hexagon: 443; Wert:  5),
      (Hexagon: 454; Wert: 26),
      (Hexagon: 531; Wert: 11),
      (Hexagon: 542; Wert: 23),
      (Hexagon: 553; Wert:  8)),

{6}  ((Hexagon: 113; Wert:  4),
      (Hexagon: 124; Wert: 25),
      (Hexagon: 135; Wert: 19),
      (Hexagon: 212; Wert: 14),
      (Hexagon: 223; Wert:  8),
      (Hexagon: 234; Wert: 20),
      (Hexagon: 245; Wert: 23),
      (Hexagon: 311; Wert: 12),
      (Hexagon: 322; Wert: 15),
      (Hexagon: 333; Wert:  0),
      (Hexagon: 344; Wert: 21),
      (Hexagon: 355; Wert: 24),
      (Hexagon: 421; Wert: 17),
      (Hexagon: 432; Wert: 11),
      (Hexagon: 443; Wert:  5),
      (Hexagon: 454; Wert: 26),
      (Hexagon: 531; Wert: 10),
      (Hexagon: 542; Wert: 13),
      (Hexagon: 553; Wert:  7)),

{7}  ((Hexagon: 113; Wert:  5),
      (Hexagon: 124; Wert: 23),
      (Hexagon: 135; Wert: 11),
      (Hexagon: 212; Wert: 22),
      (Hexagon: 223; Wert:  4),
      (Hexagon: 234; Wert: 19),
      (Hexagon: 245; Wert: 16),
      (Hexagon: 311; Wert: 12),
      (Hexagon: 322; Wert: 21),
      (Hexagon: 333; Wert:  0),
      (Hexagon: 344; Wert: 24),
      (Hexagon: 355; Wert: 15),
      (Hexagon: 421; Wert: 14),
      (Hexagon: 432; Wert: 20),
      (Hexagon: 443; Wert:  8),
      (Hexagon: 454; Wert: 26),
      (Hexagon: 531; Wert: 10),
      (Hexagon: 542; Wert: 25),
      (Hexagon: 553; Wert:  7)),

{8}  ((Hexagon: 113; Wert:  5),
      (Hexagon: 124; Wert: 26),
      (Hexagon: 135; Wert: 11),
      (Hexagon: 212; Wert: 23),
      (Hexagon: 223; Wert:  8),
      (Hexagon: 234; Wert: 20),
      (Hexagon: 245; Wert: 14),
      (Hexagon: 311; Wert: 12),
      (Hexagon: 322; Wert: 24),
      (Hexagon: 333; Wert:  0),
      (Hexagon: 344; Wert: 21),
      (Hexagon: 355; Wert: 15),
      (Hexagon: 421; Wert: 16),
      (Hexagon: 432; Wert: 19),
      (Hexagon: 443; Wert:  4),
      (Hexagon: 454; Wert: 25),
      (Hexagon: 531; Wert: 10),
      (Hexagon: 542; Wert: 22),
      (Hexagon: 553; Wert:  7)),

{9}  ((Hexagon: 113; Wert:  7),
      (Hexagon: 124; Wert: 13),
      (Hexagon: 135; Wert: 10),
      (Hexagon: 212; Wert: 26),
      (Hexagon: 223; Wert:  5),
      (Hexagon: 234; Wert: 11),
      (Hexagon: 245; Wert: 17),
      (Hexagon: 311; Wert: 24),
      (Hexagon: 322; Wert: 21),
      (Hexagon: 333; Wert:  0),
      (Hexagon: 344; Wert: 15),
      (Hexagon: 355; Wert: 12),
      (Hexagon: 421; Wert: 23),
      (Hexagon: 432; Wert: 20),
      (Hexagon: 443; Wert:  8),
      (Hexagon: 454; Wert: 14),
      (Hexagon: 531; Wert: 19),
      (Hexagon: 542; Wert: 25),
      (Hexagon: 553; Wert:  4)),

{10} ((Hexagon: 113; Wert:  7),
      (Hexagon: 124; Wert: 16),
      (Hexagon: 135; Wert: 19),
      (Hexagon: 212; Wert: 26),
      (Hexagon: 223; Wert:  8),
      (Hexagon: 234; Wert: 11),
      (Hexagon: 245; Wert: 23),
      (Hexagon: 311; Wert: 15),
      (Hexagon: 322; Wert: 24),
      (Hexagon: 333; Wert:  0),
      (Hexagon: 344; Wert: 12),
      (Hexagon: 355; Wert: 21),
      (Hexagon: 421; Wert: 17),
      (Hexagon: 432; Wert: 20),
      (Hexagon: 443; Wert:  5),
      (Hexagon: 454; Wert: 14),
      (Hexagon: 531; Wert: 10),
      (Hexagon: 542; Wert: 22),
      (Hexagon: 553; Wert:  4)),

{11} ((Hexagon: 113; Wert:  7),
      (Hexagon: 124; Wert: 22),
      (Hexagon: 135; Wert: 10),
      (Hexagon: 212; Wert: 25),
      (Hexagon: 223; Wert:  4),
      (Hexagon: 234; Wert: 19),
      (Hexagon: 245; Wert: 16),
      (Hexagon: 311; Wert: 15),
      (Hexagon: 322; Wert: 21),
      (Hexagon: 333; Wert:  0),
      (Hexagon: 344; Wert: 24),
      (Hexagon: 355; Wert: 12),
      (Hexagon: 421; Wert: 14),
      (Hexagon: 432; Wert: 20),
      (Hexagon: 443; Wert:  8),
      (Hexagon: 454; Wert: 23),
      (Hexagon: 531; Wert: 11),
      (Hexagon: 542; Wert: 26),
      (Hexagon: 553; Wert:  5)),

{12} ((Hexagon: 113; Wert:  7),
      (Hexagon: 124; Wert: 22),
      (Hexagon: 135; Wert: 19),
      (Hexagon: 212; Wert: 17),
      (Hexagon: 223; Wert:  5),
      (Hexagon: 234; Wert: 20),
      (Hexagon: 245; Wert: 26),
      (Hexagon: 311; Wert: 15),
      (Hexagon: 322; Wert: 12),
      (Hexagon: 333; Wert:  0),
      (Hexagon: 344; Wert: 24),
      (Hexagon: 355; Wert: 21),
      (Hexagon: 421; Wert: 14),
      (Hexagon: 432; Wert: 11),
      (Hexagon: 443; Wert:  8),
      (Hexagon: 454; Wert: 23),
      (Hexagon: 531; Wert: 10),
      (Hexagon: 542; Wert: 16),
      (Hexagon: 553; Wert:  4)),

{13} ((Hexagon: 113; Wert:  7),
      (Hexagon: 124; Wert: 25),
      (Hexagon: 135; Wert: 10),
      (Hexagon: 212; Wert: 17),
      (Hexagon: 223; Wert:  8),
      (Hexagon: 234; Wert: 20),
      (Hexagon: 245; Wert: 14),
      (Hexagon: 311; Wert: 24),
      (Hexagon: 322; Wert: 15),
      (Hexagon: 333; Wert:  0),
      (Hexagon: 344; Wert: 21),
      (Hexagon: 355; Wert: 12),
      (Hexagon: 421; Wert: 26),
      (Hexagon: 432; Wert: 11),
      (Hexagon: 443; Wert:  5),
      (Hexagon: 454; Wert: 23),
      (Hexagon: 531; Wert: 19),
      (Hexagon: 542; Wert: 13),
      (Hexagon: 553; Wert:  4)),

{14} ((Hexagon: 113; Wert:  7),
      (Hexagon: 124; Wert: 25),
      (Hexagon: 135; Wert: 10),
      (Hexagon: 212; Wert: 26),
      (Hexagon: 223; Wert:  8),
      (Hexagon: 234; Wert: 20),
      (Hexagon: 245; Wert: 14),
      (Hexagon: 311; Wert: 15),
      (Hexagon: 322; Wert: 24),
      (Hexagon: 333; Wert:  0),
      (Hexagon: 344; Wert: 21),
      (Hexagon: 355; Wert: 12),
      (Hexagon: 421; Wert: 16),
      (Hexagon: 432; Wert: 19),
      (Hexagon: 443; Wert:  4),
      (Hexagon: 454; Wert: 22),
      (Hexagon: 531; Wert: 11),
      (Hexagon: 542; Wert: 23),
      (Hexagon: 553; Wert:  5)),

{15} ((Hexagon: 113; Wert:  8),
      (Hexagon: 124; Wert: 23),
      (Hexagon: 135; Wert: 11),
      (Hexagon: 212; Wert: 26),
      (Hexagon: 223; Wert:  5),
      (Hexagon: 234; Wert: 20),
      (Hexagon: 245; Wert: 17),
      (Hexagon: 311; Wert: 15),
      (Hexagon: 322; Wert: 21),
      (Hexagon: 333; Wert:  0),
      (Hexagon: 344; Wert: 24),
      (Hexagon: 355; Wert: 12),
      (Hexagon: 421; Wert: 13),
      (Hexagon: 432; Wert: 19),
      (Hexagon: 443; Wert:  7),
      (Hexagon: 454; Wert: 22),
      (Hexagon: 531; Wert: 10),
      (Hexagon: 542; Wert: 25),
      (Hexagon: 553; Wert:  4)),

{16} ((Hexagon: 113; Wert:  8),
      (Hexagon: 124; Wert: 26),
      (Hexagon: 135; Wert: 11),
      (Hexagon: 212; Wert: 25),
      (Hexagon: 223; Wert:  7),
      (Hexagon: 234; Wert: 19),
      (Hexagon: 245; Wert: 13),
      (Hexagon: 311; Wert: 15),
      (Hexagon: 322; Wert: 24),
      (Hexagon: 333; Wert:  0),
      (Hexagon: 344; Wert: 21),
      (Hexagon: 355; Wert: 12),
      (Hexagon: 421; Wert: 17),
      (Hexagon: 432; Wert: 20),
      (Hexagon: 443; Wert:  5),
      (Hexagon: 454; Wert: 23),
      (Hexagon: 531; Wert: 10),
      (Hexagon: 542; Wert: 22),
      (Hexagon: 553; Wert:  4)));

var
  GlobLineColors : tLineColors;
  GlobBackground,
  GlobShadow,
  GlobLight      : tColor;

procedure Register;
procedure InitLinien(var LineColors : tLineColors);
procedure InitBackground(var Background, Light, Shadow : tColor);

implementation

constructor tHexagon.Create(AOwner: TComponent);
  begin
    inherited Create(AOwner);
    NewColors(GlobLineColors, GlobBackground, GlobLight, GlobShadow);
    FreeOccupation;
  end;

procedure tHexagon.NewColors(NewLineColors : tLineColors;
                             NewBackground, NewLight, NewShadow : tColor);
  begin
    LineColors := NewLineColors;
    Background := NewBackground;
    Light      := NewLight;
    Shadow     := NewShadow;
  end;

function tHexagon.GetLine(Richtung : tRichtung) : word;
         {
         Hoch 1,5,9
         Auf  2,6,7
         Ab   3,4,8
         }
  var    l : Word;
  begin
    case Richtung of
      rHoch : l := (hxNummer mod 3) * 4 + 1;
      rAuf  : begin
                l := (hxNummer div 3) mod 3 + 2;
                if l > 2 Then
                  Inc(l, 3);
              end;
    else
      begin
        l := hxNummer div 9 + 3;
        if l = 5 Then
          l := 8;
      end;
    end;
    GetLine := l;
  end;

procedure tHexagon.SetStyle(Style : thxStyle);
  begin
    hxStyle := Style;
    invalidate;
  end;

procedure tHexagon.SetNummer(Nummer : thxNummer);
  begin
    hxNummer := Nummer;
    invalidate;
  end;

procedure tHexagon.Init(Nummer : Word; InitStyle : thxStyle);
  begin
    hxNummer := Nummer;
    hxStyle  := InitStyle;
    invalidate;
  end;

procedure tHexagon.FreeOccupation;
  begin
    Occupying := Nil;
  end;

procedure tHexagon.Occupy(Hex : tHexagon);
  begin
    Occupying := Hex;
  end;

function tHexagon.GetOccupation : tHexagon;
  begin
    GetOccupation := Occupying;
  end;

procedure tHexagon.Paint;
  var s : String;
  begin
    Canvas.Pen.Width   := 2;
    Canvas.Pen.Color   := Shadow;
    Canvas.Brush.Style := bsSolid;
    if Style = hxHighlight Then
      Canvas.Brush.Color := Light
    else
      Canvas.Brush.Color := Background;
    Canvas.Polygon([Point(Width - 2,             Height div 2),
                    Point((Width * 3) div 4 - 1, Height - 2),
                    Point(Width div 4 + 1,       Height - 2),
                    Point(1,                     Height Div 2),
                    Point(Width div 4 + 1,       1),
                    Point((Width * 3) div 4 - 1, 1),
                    Point(Width - 2,             Height div 2)]);
    Canvas.Pen.Color := Light;
    case Style of
      hxVisible,
      hxHidden,
      hxHighlight:
        Canvas.PolyLine([Point(Width div 4 + 1,       Height - 2),
                         Point(1,                     Height Div 2),
                         Point(Width div 4 + 1,       1),
                         Point((Width * 3) div 4 - 1, 1)]);
      hxEmpty:
        Canvas.PolyLine([Point((Width * 3) div 4 - 1, 1),
                         Point(Width - 2,             Height div 2),
                         Point((Width * 3) div 4 - 1, Height - 2),
                         Point(Width div 4 + 1,       Height - 2)]);
    end;
    Canvas.Pen.Width  := 1;
    Canvas.Font.Size  := Height div 4;
    Canvas.Font.Style := [fsBold];
    case Style of
      hxVisible,
      hxHighlight:
        begin
          Canvas.Brush.Style := bsSolid;
          Canvas.Brush.Color := LineColors[GetLine(rHoch)];
          Canvas.Pen.Color   := Canvas.Brush.Color;
          Canvas.Polygon([Point((Width * 7) div 16, 3),
                          Point((Width * 9) div 16, 3),
                          Point((Width * 9) div 16, Height - 4),
                          Point((Width * 7) div 16, Height - 4)]);
          Canvas.Pen.Color   := clWhite;
          Canvas.MoveTo((Width * 7) div 16, 3);
          Canvas.LineTo((Width * 7) div 16, Height - 4);
          Canvas.MoveTo((Width * 9) div 16, 3);
          Canvas.LineTo((Width * 9) div 16, Height - 4);
          str(GetLine(rHoch), s);
          Canvas.Brush.Style := bsClear;
          Canvas.Font.Color  := clWhite;
          Canvas.TextOut(Width  div 2 - Canvas.TextWidth(s)  div 2 - 1,
                         Height div 4 - Canvas.TextHeight(s) div 2 - 1, s);
          Canvas.TextOut(Width  div 2 - Canvas.TextWidth(s)  div 2 - 1,
                         Height div 4 - Canvas.TextHeight(s) div 2 + 1, s);
          Canvas.TextOut(Width  div 2 - Canvas.TextWidth(s)  div 2 + 1,
                         Height div 4 - Canvas.TextHeight(s) div 2 + 1, s);
          Canvas.TextOut(Width  div 2 - Canvas.TextWidth(s)  div 2 + 1,
                         Height div 4 - Canvas.TextHeight(s) div 2 - 1, s);
          Canvas.Font.Color  := clBlack;
          Canvas.TextOut(Width  div 2 - Canvas.TextWidth(s)  div 2,
                         Height div 4 - Canvas.TextHeight(s) div 2, s);

          Canvas.Brush.Style := bsSolid;
          Canvas.Brush.Color := LineColors[GetLine(rAuf)];
          Canvas.Pen.Color   := Canvas.Brush.Color;
          Canvas.Polygon([Point((Width * 10) div 64, (height * 43) div 64),
                          Point((Width * 50) div 64, (height * 15) div 64),
                          Point((Width * 54) div 64, (height * 21) div 64),
                          Point((Width * 14) div 64, (height * 49) div 64)]);
          Canvas.Pen.Color   := clWhite;
          Canvas.MoveTo((Width * 10) div 64, (height * 43) div 64);
          Canvas.LineTo((Width * 50) div 64, (height * 15) div 64);
          Canvas.MoveTo((Width * 14) div 64, (height * 49) div 64);
          Canvas.LineTo((Width * 54) div 64, (height * 21) div 64);
          str(GetLine(rAuf), s);
          Canvas.Brush.Style := bsClear;
          Canvas.Font.Color  := clWhite;
          Canvas.TextOut((Width  * 5) div 16 - Canvas.TextWidth(s)  div 2 - 1,
                         (Height * 5) div  8 - Canvas.TextHeight(s) div 2 - 1, s);
          Canvas.TextOut((Width  * 5) div 16 - Canvas.TextWidth(s)  div 2 - 1,
                         (Height * 5) div  8 - Canvas.TextHeight(s) div 2 + 1, s);
          Canvas.TextOut((Width  * 5) div 16 - Canvas.TextWidth(s)  div 2 + 1,
                         (Height * 5) div  8 - Canvas.TextHeight(s) div 2 + 1, s);
          Canvas.TextOut((Width  * 5) div 16 - Canvas.TextWidth(s)  div 2 + 1,
                         (Height * 5) div  8 - Canvas.TextHeight(s) div 2 - 1, s);
          Canvas.Font.Color  := clBlack;
          Canvas.TextOut((Width  * 5) div 16 - Canvas.TextWidth(s)  div 2,
                         (Height * 5) div  8 - Canvas.TextHeight(s) div 2, s);

          Canvas.Brush.Style := bsSolid;
          Canvas.Brush.Color := LineColors[GetLine(rAb)];
          Canvas.Pen.Color   := Canvas.Brush.Color;
          Canvas.Polygon([Point((Width * 14) div 64, (height * 15) div 64),
                          Point((Width * 54) div 64, (height * 43) div 64),
                          Point((Width * 50) div 64, (height * 49) div 64),
                          Point((Width * 10) div 64, (height * 21) div 64)]);
          Canvas.Pen.Color   := clWhite;
          Canvas.MoveTo((Width * 14) div 64, (height * 15) div 64);
          Canvas.LineTo((Width * 54) div 64, (height * 43) div 64);
          Canvas.MoveTo((Width * 10) div 64, (height * 21) div 64);
          Canvas.LineTo((Width * 50) div 64, (height * 49) div 64);
          str(GetLine(rAb), s);
          Canvas.Brush.Style := bsClear;
          Canvas.Font.Color  := clWhite;
          Canvas.TextOut((Width  * 11) div 16 - Canvas.TextWidth(s)  div 2 - 1,
                         (Height *  5) div  8 - Canvas.TextHeight(s) div 2 - 1, s);
          Canvas.TextOut((Width  * 11) div 16 - Canvas.TextWidth(s)  div 2 - 1,
                         (Height *  5) div  8 - Canvas.TextHeight(s) div 2 + 1, s);
          Canvas.TextOut((Width  * 11) div 16 - Canvas.TextWidth(s)  div 2 + 1,
                         (Height *  5) div  8 - Canvas.TextHeight(s) div 2 + 1, s);
          Canvas.TextOut((Width  * 11) div 16 - Canvas.TextWidth(s)  div 2 + 1,
                         (Height *  5) div  8 - Canvas.TextHeight(s) div 2 - 1, s);
          Canvas.Font.Color  := clBlack;
          Canvas.TextOut((Width  * 11) div 16 - Canvas.TextWidth(s)  div 2,
                         (Height *  5) div  8 - Canvas.TextHeight(s) div 2, s);
        end;
    end;
  end;

procedure tHexagon.ResizeLinks(Hoehe, Weite : LongInt; HexName : String);
  var     N,
          i, j     : Integer;
  begin
    Val(Copy(HexName, 8, $ff), N, i);
    i := (N - 1) mod 9;
    j := (N - 1) div 9;
    if i > 3 then
      begin
        Top  :=  (6 * Hoehe) div 100 + ((19 * Hoehe) div 100) * (8 - i);
        Left :=  (7 * Weite) div 100 + ((14 * Weite) div 100) * j;
      end
    else
      begin
        Top  :=  (15 * Hoehe) div 100 + ((19 * Hoehe) div 100) * (3 - i);
        Left := ((14 * Weite) div 100) * j;
      end;
  end;

procedure tHexagon.ResizeRechts(Hoehe, Weite : LongInt; HexName : String);
  var     N,
          i, j     : Integer;
  begin
    Val(Copy(HexName, 8, $ff), N, i);
    i    :=  N div 100;
    j    := (N div 10) mod 10 - i div 2;
    Left :=  (38 * Weite) div 100 +  ((8 * Weite) div 100) * i;
    Top  := ((15 - (i mod 2) * 9) * Hoehe) div 100
             + ((19 * Hoehe) div 100) * j;
  end;

procedure tHexagon.HexResize(Hoehe, Weite : LongInt);
  var     N, i      : Integer;
          Occupied  : tHexagon;
  begin
    Height := Hoehe div 7;
    Width  := Weite div 12;
    case Style of
      hxEmpty :
        ResizeRechts(Hoehe, Weite, Name);
      hxHidden :
        ResizeLinks(Hoehe, Weite, Name);
      hxVisible,
      hxHighlight :
        begin
          Occupied := GetOccupation;
          if Occupied <> Nil then
            ResizeRechts(Hoehe, Weite, Occupied.Name)
          else
            begin
              Val(Copy(Name, 8, $ff), N, i);
              if N >= 100 then
                ResizeRechts(Hoehe, Weite, Name)
              else
                ResizeLinks(Hoehe, Weite, Name);
            end;
        end;
    end;
  end;

procedure tHexagon.LayAt(NewTop, NewLeft : Integer);
  begin
    Top  := NewTop;
    Left := NewLeft;
    BringToFront;
  end;

constructor tPkteLabel.Create(AOwner: TComponent);
  begin
    inherited Create(AOwner);
    NewColors;
    Line := 0;
  end;

procedure tPkteLabel.NewColors;
  begin
    LineColors := GlobLineColors;
    Background := GlobBackground;
    Light      := GlobLight;
    if (1 <= Line) and (Line <= 9) then
      Font.Color := LineColors[Line];
    if Line = -1 then
      Color := Light
    else
      Color := Background;
  end;

procedure tPkteLabel.AddLine(Hex : tHexagon; var Steine : Word;
                             Richtung : tRichtung);
  begin
    if Hex <> Nil then
      if Hex.Style <> HxEmpty then
        begin
          inc(Steine);
          if Line = 0 then
            Line := Hex.GetLine(Richtung)
          else
            if Line <> Hex.GetLine(Richtung) then
              Line := -1;
        end;
  end;

function tPkteLabel.Calculate(Hex1, Hex2, Hex3, Hex4, Hex5 : tHexagon;
                              Richtung  : tRichtung;
                              Bewertung : tBewertung) : word;
  var    Steine,
         Punkte,
         Felder  : Word;
         s       : String;
  begin
    Steine := 0;
    Line   := 0;
    Punkte := 0;
    case Bewertung of
      BewDemoPunkte,
      BewDemoEnde:
        begin
          AddLine(Hex1, Steine, Richtung);
          AddLine(Hex2, Steine, Richtung);
          AddLine(Hex3, Steine, Richtung);
          Felder := 3;
          if Hex4 <> Nil then
            begin
              Felder := 4;
              AddLine(Hex4, Steine, Richtung);
            end;
          if Hex5 <> Nil then
            begin
              Felder := 5;
              AddLine(Hex5, Steine, Richtung);
            end;
        end;
      else
        begin
          AddLine(Hex1.GetOccupation, Steine, Richtung);
          AddLine(Hex2.GetOccupation, Steine, Richtung);
          AddLine(Hex3.GetOccupation, Steine, Richtung);
          Felder := 3;
          if Hex4 <> Nil then
            begin
              Felder := 4;
              AddLine(Hex4.GetOccupation, Steine, Richtung);
            end;
          if Hex5 <> Nil then
            begin
              Felder := 5;
              AddLine(Hex5.GetOccupation, Steine, Richtung);
            end;
        end;
      end;
    if Line <> -1 then
      case Bewertung of
        bewAnfang     : Punkte := integer(Steine) * integer(Steine) * Line;
        bewEnde,
        BewDemoEnde   : Punkte := integer(Felder) * integer(Steine) * Line;
        bewPunkte,
        BewDemoPunkte : if Steine = Felder then
                          Punkte := Steine * Line;
      end;
    if (1 <= Line) and (Line <= 9) then
      begin
        Font.Color := LineColors[Line];
        color      := Background;
        if Steine = Felder then
          Font.Style := [fsBold]
        else
          Font.Style := [fsItalic, fsBold];
      end
    else
      begin
        Font.Color := clWindowText;
        if Line = 0 Then
          begin
            color      := Background;
            Font.Style := [fsItalic, fsBold]
          end
        else
          begin
            color      := Light;
            Font.Style := [fsBold];
          end;
      end;
    Str(Punkte, s);
    Caption   := s;
    Calculate := Punkte;
  end;

constructor tBar.Create(AOwner: TComponent);
  begin
    inherited Create(AOwner);
    NewColors(GlobLineColors, GlobBackground, GlobLight, GlobShadow);
    State  := None;
    Marked := False;
  end;

procedure tBar.NewColors(NewLineColors : tLineColors;
                         NewBackground, NewLight, NewShadow : tColor);
  begin
    LineColors := NewLineColors;
    Background := NewBackground;
    Light      := NewLight;
    Shadow     := NewShadow;
  end;

procedure tBar.Paint;
  var     s     : String;
          i,
          error : Integer;
  begin
    s := Copy(Name, 4, $ff); {Bar#}
    val(s, i, Error);
    if Error = 0 then
      Canvas.Brush.Color := LineColors[i]
    else
      if s = 'Background1' then
        Canvas.Brush.Color := Background
      else if s = 'Background2' then
        Canvas.Brush.Color := Background
      else if s = 'Shadow' then
        Canvas.Brush.Color := Shadow
      else if s = 'Light1' then
        Canvas.Brush.Color := Light
      else if s = 'Light2' then
        Canvas.Brush.Color := Light;
    Canvas.FillRect(Rect(0, 0, Width, Height));
    Case State of
      None:
        if Error <> 0 Then
          begin
            Canvas.Brush.Color := clBlack;
            Canvas.FrameRect(Rect(0, 0, Width, Height));
          end;
      HiLighted : HiLight;
      Pressed   : Down;
    end;
  end;

procedure tBar.Mark;
  begin
    Marked := True;
  end;

procedure tBar.HiLight;
  begin
    if State = None Then
      begin
        Canvas.Pen.Color := Light;
        Canvas.MoveTo(0,         Height - 1);
        Canvas.LineTo(0,         0);
        Canvas.LineTo(Width - 1, 0);
        Canvas.Pen.Color := Shadow;
        Canvas.LineTo(Width - 1, Height - 1);
        Canvas.LineTo(0,         Height - 1);
        State := HiLighted;
      end;
    Marked := False;
  end;

function tBar.Up : Boolean;
  begin
    Up    := State = Pressed;
    State := None;
    HiLight;
  end;

procedure tBar.Down;
  begin
    Canvas.Pen.Color := Shadow;
    Canvas.MoveTo(0,         Height - 1);
    Canvas.LineTo(0,         0);
    Canvas.LineTo(Width - 1, 0);
    Canvas.Pen.Color := Light;
    Canvas.LineTo(Width - 1, Height - 1);
    Canvas.LineTo(0,         Height - 1);
    State := Pressed;
  end;

procedure tBar.UnLight;
  var     s     : String;
          i,
          error : Integer;
  begin
    if not Marked Then
      if State <> None Then
        begin
          s := Copy(Name, 4, $ff); {Bar#}
          val(s, i, Error);
          if Error = 0 then
            Canvas.Brush.Color := LineColors[i]
          else
            Canvas.Brush.Color := clBlack;
          Canvas.FrameRect(Rect(0, 0, Width, Height));
          State := None;
        end;
  end;

procedure Register;
  begin
    RegisterComponents('Take it easy', [tHexagon, tPkteLabel, tBar]);
  end;

procedure InitLinien(var LineColors : tLineColors);
  begin
    LineColors[1] := clGray;
    LineColors[2] := clFuchsia;
    LineColors[3] := clPurple;
    LineColors[4] := clAqua;
    LineColors[5] := clGreen;
    LineColors[6] := clRed;
    LineColors[7] := clLime;
    LineColors[8] := clMaroon;
    LineColors[9] := clYellow;
  end;

procedure InitBackground(var Background, Light, Shadow : tColor);
  begin
    {$ifdef Unix}
      Background := $CCCCCC;
    {$else}
      Background := clBtnFace;
    {$endif}
    Light        := clBtnHighlight;
    Shadow       := clBtnShadow;
  end;

begin
  InitLinien(GlobLineColors);
  InitBackground(GlobBackground, GlobLight, GlobShadow);
end.
