unit link;

interface

uses
  Windows,
  SysUtils,
  Classes,
  Controls,
  Forms,
  StdCtrls,
  IniFiles,
  FileCtrl,
  Registry,
  Hexagon,
  Zusatz,
  WolfUtil;

type
  TLinkForm            = class(TForm)
    CheckBoxDesktop    : TCheckBox;
    GroupBox           : TGroupBox;
    CheckBoxStartmenue : TCheckBox;
    BtnNeu             : TButton;
    EditOrdner         : TEdit;
    BtnOk              : TButton;
    DirectoryListBox   : TDirectoryListBox;
    procedure            FormCreate(Sender: TObject);
    procedure            EditOrdnerChange(Sender: TObject);
    procedure            BtnNeuClick(Sender: TObject);
    procedure            BtnOkClick(Sender: TObject);
    procedure            FormShow(Sender: TObject);
    procedure FormClose(Sender: TObject; var CloseAction: TCloseAction);
  private
    Startmenue         : String;
    Desktop            : Boolean;
    procedure            SeekDir(var Dir : String; SubDir : String);
    function             LinkName : String;
    procedure            MakeLink(Pfad: String);
  end;

var
  LinkForm : TLinkForm;

implementation

{$R *.DFM}

function tLinkForm.LinkName : String;
  begin
    case SysLocale.PriLangID of
      Lang_German:
        LinkName := 'Take it easy';
    else
        LinkName := 'Hextension';
    end;
  end;

procedure TLinkForm.SeekDir(var Dir : String; SubDir : String);
  var     NewDir : String;
  begin
    NewDir := MakePath(Dir, SubDir);
    if DirectoryExists(NewDir) then
      Dir := NewDir;
  end;

procedure TLinkForm.FormShow(Sender: TObject);
  var     IniFile  : tIniFile;
          Registry : tRegistry;
          s        : String;
  begin
    CheckBoxStartmenue.Checked := True;
    CheckboxDesktop.Checked    := False;
    Registry                   := tRegistry.Create;
    Registry.RootKey           := HKEY_Current_User;
    if Registry.openkey(HKCU_ShellFolders, False) then
      begin
        try
          s := Registry.ReadString(ProgramsFolder);
          SeekDir(s, CP1252ToSystem('Zubehr'));
          SeekDir(s, 'Accessories');
          SeekDir(s, 'Spiele');
          SeekDir(s, 'Games');
          DirectoryListBox.Directory := s;
        except
        end;
        try
          Desktop :=
            FileExists(MakePath(Registry.ReadString(DesktopFolder),
                                LinkName + LinkExt));
          CheckboxDesktop.Checked := Desktop;
        except
        end;
      end;
    Registry.Free;
    IniFile    := tIniFile.Create(IniName);
    Startmenue := Inifile.ReadString(Name, IniStartmenue, '*');
    if Startmenue = '*' then
      Startmenue := ''
    else
      if Startmenue = '' then
        CheckBoxStartmenue.Checked := False
      else
        try
          DirectoryListBox.Directory := Startmenue;
        except
          CheckBoxStartmenue.Checked := False;
          Startmenue                 := '';
        end;
    IniFile.free;
  end;

procedure TLinkForm.EditOrdnerChange(Sender: TObject);
  begin
    BtnNeu.enabled := EditOrdner.Text > '';
  end;

procedure TLinkForm.BtnNeuClick(Sender: TObject);
  var     Error : Boolean;
          s     : String;
  begin
    Error := False;
    s := MakePath(DirectoryListBox.Directory, EditOrdner.Text);
    {$I+}
    try
      MkDir(s);
      DirectoryListBox.Update;
    except
      Error := True;
    end;
    {$I-}
    if Error then
      case SysLocale.PriLangID of
        Lang_German:
          MsgBox('kann ' + s + ' nicht anlegen', Caption, mb_IconHand);
      else
        MsgBox('cannot create ' + s, Caption, mb_IconHand);
      end;
  end;

procedure TLinkForm.MakeLink(Pfad: String);
  var s : String;
  begin
    s := LinkName;
    s := MkLink(Application.ExeName, s, StartDir, MakePath(Pfad, s + LinkExt));
    if s > '' then
      MsgBox(s, Caption, mb_IconHand);
  end;

procedure TLinkForm.BtnOkClick(Sender: TObject);
  var     s        : String;
          IniFile  : tIniFile;
          Registry : tRegistry;
  begin
    if CheckBoxStartmenue.Checked then
      s := DirectoryListBox.Directory
    else
      s := '';
    IniFile := tIniFile.Create(IniName);
    IniFile.WriteString(Name, IniStartMenue, s);
    IniFile.Free;
    if s <> StartMenue then
      begin
        MakeLink(s);
        if StartMenue <> '' then
          DeleteFile(MakePath(StartMenue, LinkName + LinkExt));
      end;
    if CheckBoxDesktop.Checked <> Desktop then
      begin
        Registry         := tRegistry.Create;
        Registry.RootKey := HKEY_Current_User;
        if Registry.openkey(HKCU_ShellFolders, False) then
          try
            s := Registry.ReadString(DesktopFolder);
            if CheckBoxDesktop.Checked then
              MakeLink(s)
            else
              DeleteFile(MakePath(s, LinkName + LinkExt));
          except
          end;
        Registry.Free;
      end;
    ModalResult := mrOk;
  end;

procedure TLinkForm.FormCreate(Sender: TObject);
  var     IniFile  : tIniFile;
          MustShow : Boolean;
  begin
    case SysLocale.PriLangID of
      Lang_German:
    else
      Caption                    := 'Create shortcuts';
      CheckBoxStartmenue.Caption := 'in the selected folder of the &start menu';
      CheckBoxDesktop.Caption    := 'on the &desktop';
      BtnNeu.Caption             := '&New folder';
    end;
    IniFile  := tIniFile.Create(IniName);
    MustShow := Inifile.ReadString(Name, IniStartmenue, '*') = '*';
    IniFile.free;
    LoadFormFromIni(Self, Name);
    if MustShow then
      ShowModal;
  end;

procedure TLinkForm.FormClose(Sender: TObject; var CloseAction: TCloseAction);
  begin
    SaveFormToIni(Self, Name);
  end;

end.
