unit names;

{$ifdef fpc}
  {$mode objfpc}{$H+}
{$endif}

interface

uses
  {$ifndef Unix}
    Windows,
  {$endif}
  {$ifdef fpc}
    LResources,
  {$endif}
  SysUtils,
  Classes,
  Graphics,
  Controls,
  Forms,
  Dialogs,
  StdCtrls,
  Buttons,
  Zusatz,
  WolfUtil,
  hexagon,
  inifiles;

type

  TPlayerForm    = class(TForm)
    Spieler      : TComboBox;
    Label1       : TLabel;
    OkButton     : TButton;
    CancelButton : TButton;
    ClearButton  : TButton;
    Anleitung    : TLabel;
    procedure      Enable(Sender: TObject);
    procedure      FormCreate(Sender: TObject);
    procedure      Loesche(Sender: TObject);
    procedure      FormClose(Sender: TObject; var CloseAction: TCloseAction);
    procedure      ToFront(Sender: TObject);
  private
    procedure      SaveToIni;
  end;

var
  PlayerForm : TPlayerForm;

implementation

{$ifdef fpc}
  {$R *.lfm}
{$else}
  {$R *.dfm}
{$endif}

uses Thread;

procedure TPlayerForm.Enable(Sender: TObject);
  begin
    if trim(Spieler.Text) = 'PC' Then
      begin
        OkButton.Enabled    := True;
        ClearButton.Enabled := False;
      end
    else
      if Copy(trim(Spieler.Text), 1, 2) = '\\' then
        begin
          OkButton.Enabled :=
            (length(trim(Spieler.Text)) > 2)
            and (pos('\', Copy(trim(Spieler.Text), 3, length(Spieler.Text))) = 0);
          ClearButton.Enabled := True;
        end
      else
        if Pos('\', Spieler.Text) > 0 then
          begin
            OkButton.Enabled    := False;
            ClearButton.Enabled := True;
          end
        else
          begin
            OkButton.Enabled    := trim(Spieler.Text) > '';
            ClearButton.Enabled := trim(Spieler.Text) > '';
          end;
  end;

procedure TPlayerForm.FormCreate(Sender: TObject);
var       i       : Integer;
          s       : String;
          IniFile : TIniFile;
begin
  Spieler.MaxLength := sizeof(tPlayerName) - 1;
  LoadFormFromIni(Self, Name);
  IniFile := tIniFile.Create(IniName);
  i       := 0;
  repeat
    s := IniFile.ReadString(IniNames, IntToDez(i, 2, False), '');
    if s > '' Then
      Spieler.Items.Add(s);
    inc(i);
  until s = '';
  IniFile.Free;
  if Spieler.Items.IndexOf('PC') < 0 Then
    Spieler.Items.Add('PC');
  case SysLocale.PriLangID of
    Lang_German:
  else
    Caption              := 'Hextension player';
    CancelButton.Caption := '&Cancel';
    ClearButton.Caption  := '&Delete entry';
    Anleitung.Caption    := 'Enter your name or choose from the list. ' +
                            '"PC" will make the computer play. ' +
                            '"\\"(PC Name) invites another computer ' +
                            'to take part. ' +
                            'Call this dialogue again if more players ' +
                            'shall compete.';
  end;
end;

procedure TPlayerForm.SaveToIni;
  var     i       : Integer;
          IniFile : tIniFile;
  begin
    IniFile := tIniFile.Create(IniName);
    IniFile.EraseSection(IniNames);
    for i := 0 to Spieler.Items.Count - 1 do
      IniFile.WriteString(IniNames, IntToDez(i, 2, False),
                          Spieler.Items.Strings[i]);
    IniFile.Free;
  end;

procedure TPlayerForm.Loesche(Sender: TObject);
  var     i : Integer;
  begin
    for i := Spieler.Items.Count - 1 downto 0 do
      if Spieler.Items[i] = Spieler.Text Then
        Spieler.Items.Delete(i);
    SaveToIni;
    Spieler.Text := '';
    Enable(Sender);
  end;

procedure TPlayerForm.FormClose(Sender: TObject; var CloseAction: TCloseAction);
  begin
    if CloseAction = caFree then;
    if trim(Spieler.Text) <> '' Then
      if Spieler.Items.IndexOf(trim(Spieler.Text)) < 0 Then
        begin
          Spieler.Items.Add(trim(Spieler.Text));
          SaveToIni;
        end;
    SaveFormToIni(Self, Name);
  end;

procedure TPlayerForm.ToFront(Sender: TObject);
  begin
    BringToFront;
  end;

end.
