unit port;

{$ifdef fpc}
  {$mode objfpc}{$H+}
{$endif}

interface

uses
  {$ifdef Unix}
    LibC,
  {$else}
    Windows,
    WinSock,
  {$endif}
  {$ifdef fpc}
    LCLType,
    LResources,
  {$endif}
  SysUtils,
  Classes,
  Graphics,
  Controls,
  Forms,
  Dialogs,
  StdCtrls,
  Hexagon,
  Main,
  Zusatz,
  WolfUtil;

type

  { TPortForm }

  TPortForm          = class(TForm)
    BeschreibungPort : TLabel;
    Port             : TEdit;
    PortName         : TLabel;
    Label2           : TLabel;
    BtnOk            : TButton;
    BtnCancel        : TButton;
    MyIpPrompt       : TLabel;
    MyIpList         : TLabel;
    procedure          FormCreate(Sender: TObject);
    procedure          BtnOkClick(Sender: TObject);
    procedure          FormClose(Sender: TObject; var CloseAction: TCloseAction);
  end;

var
  PortForm           : TPortForm;

implementation

{$ifdef fpc}
  {$R *.lfm}
{$else}
  {$R *.dfm}
{$endif}

procedure TPortForm.FormCreate(Sender: TObject);
  begin
    case SysLocale.PriLangID of
      Lang_German:
    else
      Caption                  := 'Hextension socket port';
      BeschreibungPort.Caption :=
        'Please agree with the other players and the network administrator '
        + 'about a UDP port, enter the same on all PCs and open the firewall '
        + 'or leave the data field blank for not playing in the network.';
      BtnCancel.Caption        := '&Cancel';
    end;
    SaveFormToIni(Self, Name);
    PortName.Caption := UDPService;
  end;

procedure TPortForm.BtnOkClick(Sender: TObject);
  const   {$ifdef Unix}
            sServices = '/etc/services';
          {$else}
            sServices = 'Services';
          {$endif}
  var     UDPPort     : Integer;
          e           : Integer;
          ServEnt     : pServEnt;
          b           : Boolean;
  begin
    b := True;
    if Port.Text = '' then
      UdpPort := 0
    else
      begin
        val(Port.Text, UdpPort, e);
        if (e <> 0) or (UDPPort <= 0) or (UDPPort > $FFFF) then
          begin
            b := False;
            case SysLocale.PriLangID of
              Lang_German:
                MsgBox(CP1252ToSystem('ungltige Portnummer'),
                       Caption, mb_IconHand);
            else
              MsgBox('invalid port number', Caption, mb_IconHand);
            end;
          end;
      end;
    if b and (UdpPort <> 0) then
      begin
        ServEnt := GetServByPort(htons(UDPPort), 'udp');
        if ServEnt <> Nil then
          begin
            b := False;
            if ServEnt^.s_name = UDPService then
              ModalResult := mrOk
            else
              case SysLocale.PriLangID of
                Lang_German:
                  MsgBox('Port bereits vergeben', Caption, mb_IconHand);
                else
                  MsgBox('port already in use', Caption, mb_IconHand);
              end;
          end;
      end;
    if b then
      begin
        b := False;
        case SetServEnt(UDPService, 'udp', UDPPort, '') of
          SetServEntOk:
            b := True;
          SetServEntNotFound:
            case SysLocale.PriLangID of
              Lang_German:
                MsgBox(sServices + ' Datei nicht gefunden',
                       Caption, mb_IconHand);
            else
              MsgBox(sServices + ' file not found', Caption, mb_IconHand);
            end;
          SetServEntErrCreateTemp:
            case SysLocale.PriLangID of
              Lang_German:
                MsgBox(CP1252ToSystem('Kann temporre ') + sServices
                       + ' Datei nicht erzeugen',
                       Caption, mb_IconHand);
            else
              MsgBox('Cannot create temporary ' + sServices + ' file',
                     Caption, mb_IconHand);
            end;
          SetServEntErrDelPrev:
            case SysLocale.PriLangID of
              Lang_German:
                MsgBox('Kann ' + sServices + ' Datei nicht beschreiben',
                       Caption, mb_IconHand);
            else
              MsgBox('Cannot write to ' + sServices + ' file',
                     Caption, mb_IconHand);
            end;
          SetServEntErrRename:
            case SysLocale.PriLangID of
              Lang_German:
                MsgBox(CP1252ToSystem('Kann temporre ' + sServices
                         + ' Datei nicht in endgltige umbenennen'),
                       Caption, mb_IconHand);
            else
              MsgBox('Cannot rename temporary ' + sServices
                     + ' file back to original',
                     Caption, mb_IconHand);
            end;
          SetServEntInvalidPathLen:
            case SysLocale.PriLangID of
              Lang_German:
                MsgBox(CP1252ToSystem(
                  'ungltige Pfad Lnge beim Beschreiben der ') + sServices
                  + ' Datei', Caption, mb_IconHand);
            else
              MsgBox('invalid path len on writing to the ' + sServices
                     + ' file', Caption, mb_IconHand);
            end;
        else
          case SysLocale.PriLangID of
            Lang_German:
              MsgBox('unbekannter Fehler beim Eintragen in die ' + sServices
                     + ' Datei', Caption, mb_IconHand);
          else
            MsgBox('unknown error on writing to the ' + sServices + ' file',
                   Caption, mb_IconHand);
          end;
        end;
        if b then
          ModalResult := mrOk
        else
          ModalResult := mrCancel;
      end;
  end;

procedure TPortForm.FormClose(Sender: TObject; var CloseAction: TCloseAction);
  begin
    if CloseAction = caFree then;
    SaveFormToIni(Self, Name);
  end;

end.
