unit thread;

{$ifdef fpc}
  {$mode objfpc}{$H+}
{$endif}
interface

uses
  {$ifdef Unix}
    Sockets,
    LibC,
    BaseUnix,
    WolfUtil,
  {$else}
    {$ifdef fpc}
      Sockets,
    {$endif}
    Windows,
    WinSock,
  {$endif}
  Forms,
  Controls,
  Graphics,
  Dialogs,
  Classes,
  SysUtils,
  Zusatz,
  Hexagon;

const
  TelegrammAnfrageOhneAntwort     =  0;
  TelegrammMitspielerAlt          =  1;
  TelegrammBereitsBegonnen        =  2;
  TelegrammNochNichtBegonnen      =  3;
  TelegrammEnde                   =  4;
  TelegrammSteinGesetzt           =  5;
  TelegrammAnfrageMitAntwort      =  6;
  TelegrammSteinTest              =  7;
  TelegrammMitspieler             =  8;
  TelegrammChat                   =  9;
  TelegrammNotifyAddress          = 10;

  {$ifdef fpc}
    Delay                         = 1000;
  {$else}
    Delay                         = 50; { msec }
  {$endif}

type
  tNetworkState                   = (NetworkNotAvailable,
                                     NetworkStarted,
                                     NetworkReady,
                                     NetworkBlocked,
                                     NetworkTerminating,
                                     NetworkTerminated);

  tPlayerName                     = Array[0 .. 19] of char;
  tChatText                       = Array[0 .. 59] of Char;
  tSteine                         = Array[1 .. GesamtSteine] of word;

  {$ifdef fpc}
    tTelegramm                      = Packed Record
      case                             Kennung        : Byte of
        TelegrammNotifyAddress,
        TelegrammAnfrageOhneAntwort,
        TelegrammAnfrageMitAntwort  : (Filler1        : Array[1..3] of Byte;
                                       AnfrageIp      : LongInt);
                                       TelegrammMitspielerAlt,
        TelegrammMitspieler         : (Filler2        : Byte;
                                       MitspielSteine : tSteine;
                                       MitspielIp     : LongInt;
                                       MitspielName   : tPlayerName;
                                       MitspielHandle : tHandle;
                                       // TelegrammMitspielerAlt nur bis hier
                                       MitspielStdIp  : LongInt);
        TelegrammBereitsBegonnen,
        TelegrammNochNichtBegonnen  : (Filler4        : Array[1..3] of Byte;
                                       AblehnerIp     : LongInt);
  
        TelegrammEnde               : (Filler5        : Array[1..3] of Byte;
                                       EndeIp         : LongInt;
                                       EndeHandle     : tHandle);
        TelegrammSteinGesetzt,
        TelegrammSteinTest          : (Filler6        : Array[1..3] of Byte;
                                       GesetztIp      : LongInt;
                                       GesetztHandle  : tHandle;
                                       GesetztNr      : word;
                                       GesetztZiel    : word);
        TelegrammChat               : (Filler7        : Array[1..3] of Byte;
                                       ChatIp         : LongInt;
                                       ChatStdIp      : LongInt;
                                       ChatText       : tChatText);
    end;
  {$else}
    tTelegramm                      = Record
      case                             Kennung        : Byte of
        TelegrammNotifyAddress,
        TelegrammAnfrageOhneAntwort,
        TelegrammAnfrageMitAntwort  : (AnfrageIp      : LongInt);
        TelegrammMitspielerAlt,
        TelegrammMitspieler         : (MitspielSteine : tSteine;
  
  
                                       MitspielIp     : LongInt;
                                       MitspielName   : tPlayerName;
                                       MitspielHandle : tHandle;
                                       { TelegrammMitspielerAlt nur bis hier }
                                       MitspielStdIp  : LongInt);
        TelegrammBereitsBegonnen,
        TelegrammNochNichtBegonnen  : (AblehnerIp     : LongInt);
        TelegrammEnde               : (EndeIp         : LongInt;
  
  
  
                                       EndeHandle     : tHandle);
        TelegrammSteinGesetzt,
        TelegrammSteinTest          : (GesetztIp      : LongInt;
  
                                       GesetztHandle  : tHandle;
                                       GesetztNr      : ShortInt;
                                       GesetztZiel    : word);
        TelegrammChat               : (ChatIp         : LongInt;
  
                                       ChatStdIp      : LongInt;
                                       ChatText       : tChatText);
    end;
  {$endif}

  tNetworkThread   = class(TThread)
  public
    procedure        Execute; override;
  private
    UDPSocket      : tSocket;
    NetworkState   : tNetworkState;
    UDPPort        : Word;
    UDPProto       : ShortInt;
    Caption        : String;
    MainFormHandle : tHandle;
    Telegramm      : tTelegramm;
    MessageText    : String;
    procedure        AddPlayer; Register;
    procedure        AnfrageOhneAntwort; Register;
    procedure        AnfrageMitAntwort; Register;
    procedure        NotifyAddress; Register;
    procedure        GetParams; Register;
    procedure        DeletePlayer; Register;
    procedure        SteinGesetzt; Register;
    procedure        SteinTest; Register;
    procedure        Chat; Register;
    procedure        Ready; Register;
    procedure        Blocked; Register;
    procedure        ThreadMessage; Register;
    procedure        ProcessTelegramm;
    procedure        BlockMessage(s : String);
    procedure        NormalMessage(s : String);
  end;

implementation

uses
  Main;

procedure tNetworkThread.Blocked;
  begin
    MainForm.ThreadBlocked;
  end;

procedure tNetworkThread.Ready;
  begin
    MainForm.ThreadReady;
  end;

procedure tNetworkThread.AddPlayer;
  begin
    If Telegramm.MitspielIp <> 0 then
      If Telegramm.MitspielStdIp <> 0 then
        If MainForm.ReceiveSteine(Telegramm.MitspielIp,
                                  Telegramm.MitspielSteine) then
          MainForm.NeuerSpieler(Telegramm.MitspielIp,
                                Telegramm.MitspielStdIp,
                                Telegramm.MitspielName,
                                Telegramm.MitspielHandle);
  end;

procedure tNetworkThread.DeletePlayer;
  begin
    if Telegramm.EndeIp <> 0 then
      MainForm.DeletePlayer(Telegramm.EndeIp, Telegramm.EndeHandle);
  end;

procedure tNetworkThread.AnfrageOhneAntwort;
  begin
    if Telegramm.AnfrageIp <> 0 then
      MainForm.Anfrage(Telegramm.AnfrageIp, False);
  end;

procedure tNetworkThread.AnfrageMitAntwort;
  begin
    if Telegramm.AnfrageIp <> 0 then
      MainForm.Anfrage(Telegramm.AnfrageIp, True);
  end;

procedure tNetworkThread.NotifyAddress;
  begin
    if Telegramm.AnfrageIp <> 0 then
      MainForm.NotifyAddress(Telegramm.AnfrageIp);
  end;

procedure tNetworkThread.SteinGesetzt;
  begin
    if Telegramm.GesetztIp <> 0 then
      MainForm.SteinGesetzt(Telegramm.GesetztIp,
                            Telegramm.GesetztHandle,
                            Telegramm.GesetztNr,
                            Telegramm.GesetztZiel,
                            True);
  end;

procedure tNetworkThread.SteinTest;
  begin
    if Telegramm.GesetztIp <> 0 then
      MainForm.SteinGesetzt(Telegramm.GesetztIp,
                            Telegramm.GesetztHandle,
                            Telegramm.GesetztNr,
                            Telegramm.GesetztZiel,
                            False);
  end;

procedure tNetworkThread.Chat;
  begin
    if Telegramm.ChatIp <> 0 then
      MainForm.Chat(Telegramm.ChatIp, Telegramm.ChatStdIp, Telegramm.ChatText);
  end;

procedure tNetworkThread.ThreadMessage;
  begin
    MainForm.ThreadMessage(MessageText);
  end;

procedure tNetworkThread.GetParams;
  begin
    NetworkState   := MainForm.NetworkState;
    UDPProto       := MainForm.UDPProto;
    UDPPort        := MainForm.UDPPort;
    Caption        := MainForm.Caption;
    MainFormHandle := MainForm.Handle;
  end;

procedure tNetworkThread.ProcessTelegramm;
  begin
    case Telegramm.Kennung of
      {$ifdef fpc}
        TelegrammAnfrageOhneAntwort:
          Synchronize(@AnfrageOhneAntwort);
        TelegrammAnfrageMitAntwort:
          Synchronize(@AnfrageMitAntwort);
        TelegrammMitspielerAlt:
          begin
            Telegramm.MitspielStdIp := Telegramm.MitspielIp;
            Synchronize(@AddPlayer);
          end;
        TelegrammEnde:
          Synchronize(@DeletePlayer);
        TelegrammSteinGesetzt:
          Synchronize(@SteinGesetzt);
        TelegrammSteinTest:
          Synchronize(@SteinTest);
        TelegrammMitspieler:
          Synchronize(@AddPlayer);
        TelegrammChat:
          Synchronize(@Chat);
        TelegrammNotifyAddress:
          Synchronize(@NotifyAddress);
      {$else}
        TelegrammAnfrageOhneAntwort:
          Synchronize(AnfrageOhneAntwort);
        TelegrammAnfrageMitAntwort:
          Synchronize(AnfrageMitAntwort);
        TelegrammMitspielerAlt:
          begin
            Telegramm.MitspielStdIp := Telegramm.MitspielIp;
            Synchronize(AddPlayer);
          end;
        TelegrammEnde:
          Synchronize(DeletePlayer);
        TelegrammSteinGesetzt:
          Synchronize(SteinGesetzt);
        TelegrammSteinTest:
          Synchronize(SteinTest);
        TelegrammMitspieler:
          Synchronize(AddPlayer);
        TelegrammChat:
          Synchronize(Chat);
        TelegrammNotifyAddress:
          Synchronize(NotifyAddress);
      {$endif}
      TelegrammBereitsBegonnen:
        case SysLocale.PriLangID of
          Lang_German:
            NormalMessage('Auf PC ' + Inet_NtoA(tInAddr(Telegramm.AblehnerIp))
                          + ' ist das Spiel bereits im Gange');
        else
            NormalMessage('The game on PC '
                          + Inet_NtoA(tInAddr(Telegramm.AblehnerIp))
                          + ' is already in progress');
        end;
      TelegrammNochNichtBegonnen:
        case SysLocale.PriLangID of
          Lang_German:
            NormalMessage('Auf PC ' + Inet_NtoA(tInAddr(Telegramm.AblehnerIp))
                          + ' gibt es noch keine Spieler');
        else
          NormalMessage('There are no players on PC '
                        + Inet_NtoA(tInAddr(Telegramm.AblehnerIp)));
        end;
    end;
  end;

procedure tNetworkThread.BlockMessage(s : String);
  begin
    {$ifdef fpc}
      Synchronize(@Blocked);
    {$else}
      Synchronize(Blocked);
    {$endif}
    case SysLocale.PriLangID of
      Lang_German:
        NormalMessage(s + ': ' + WSAErrorString(WsaGetLastError) + A_LF + A_LF
                        + 'Starten Sie nach Beheben des Fehlers das Menue '
                        + '"Einstellung - Netzwerk"');
    else
      NormalMessage(s + ': ' + WSAErrorString(WsaGetLastError) + A_LF + A_LF
                      + 'After fixing the problem call the menu '
                      + '"Settings - Network"');
    end;
  end;

procedure tNetworkThread.NormalMessage(s : String);
  begin
    MessageText := s;
    {$ifdef fpc}
      Synchronize(@ThreadMessage);
    {$else}
      Synchronize(ThreadMessage);
    {$endif}
  end;

procedure tNetworkThread.Execute;
  var     SockName  : tSockAddrIn;
          Action    : Longint;
          i         : Integer;
          b,
          ok        : Boolean;
          TimeVal   : TTimeVal;
          FDs       : tFDSet;
  begin
    {$ifdef fpc}
      Synchronize(@GetParams);
    {$else}
      Synchronize(GetParams);
    {$endif}
    SockName.sin_family      := AF_INET;
    SockName.sin_addr.s_addr := InAddr_any;
    {$ifdef Unix}
      FDs[0]                 := 0;
    {$else}
      FDs.fd_count           := 0;
    {$endif}
    FillChar(SockName.sin_zero, sizeof(SockName.sin_zero), 0);
    While (NetworkState <> NetworkTerminating)
      and (NetworkState <> NetworkTerminated) do
      if NetworkState <> NetworkReady then
        begin
          Sleep(Delay);
          {$ifdef fpc}
            Synchronize(@GetParams);
          {$else}
            Synchronize(GetParams);
          {$endif}
        end
      else
        begin
          SockName.sin_port := UDPPort;
          {$ifdef fpc}  
            UDPSocket := fpSocket(AF_Inet, LongInt(sock_dgram), UDPproto);
          {$else}
            UDPSocket := Socket(AF_Inet, sock_dgram, UDPproto);
          {$endif}
          if UDPSocket = Invalid_Socket then
            BlockMessage('Open socket')
          else
            begin
              {$ifdef fpc}  
                if fpBind(UDPSocket, @SockName, sizeof(SockName)) <> 0 then 
              {$else}  
                if bind(UDPSocket, SockName, sizeof(SockName)) <> 0 then
              {$endif}
                BlockMessage('Binding the socket')
              else
                begin 
                  Action := 1;
                  {$ifdef fpc}  
                    {$ifdef Unix}
                      b := fpioctl(UDPSocket, FIONBIO, @Action) <> 0;
                    {$else}
                      b := IoCtlSocket(UDPSocket, longint(FIONBIO), @Action) <> 0;
                    {$endif}
                  {$else}  
                    b := IoCtlSocket(UDPSocket, fIONBIO, Action) <> 0;
                  {$endif}
                  if b then
                    BlockMessage('Making the socket non blocking')
                  else
                    begin
                      ok := True;
                      {$ifdef fpc}  
                        Synchronize(@Ready);
                      {$else}  
                        Synchronize(Ready);
                      {$endif}
                      while (NetworkState = NetworkReady)
                            and ok
                            and (SockName.sin_port = UDPPort) 
                      do
                        begin
                          {$ifdef Unix}
                            fpfd_zero(FDs);
                            fpfd_set(UDPSocket, FDs);
                          {$else}
                            fd_zero(FDs);
                            fd_set(UDPSocket, FDs);
                          {$endif}
                          TimeVal.tv_sec  := 0;
                          TimeVal.tv_usec := Delay * 1000;
                          {$ifdef Unix}
                            b := fpSelect(UDPSocket + 1, @FDs, Nil, Nil, @TimeVal) > 0;
                          {$else}
                            b := Select(UDPSocket + 1, @FDs, Nil, Nil, @TimeVal) > 0;
                          {$endif}
                          if b then
                            begin
                              {$ifdef fpc}  
                                i := fpRecv(UDPSocket, @Telegramm, sizeof(Telegramm), 0);
                              {$else}
                                i := recv(UDPSocket, Telegramm, sizeof(Telegramm), 0);
                              {$endif}
                              if i < 0 then
                                begin
                                  ok := False;
                                  NormalMessage('Receiving data from socket: '
                                                + WSAErrorString(i));
                                end
                              else
                                if i = sizeof(Telegramm) then
                                  ProcessTelegramm;
                            end;
                          {$ifdef fpc} 
                            Synchronize(@GetParams); 
                          {$else} 
                            Synchronize(GetParams);
                          {$endif} 
                        end;
                    end;
                end;
              {$ifdef fpc}  
                if fpShutdown(UDPSocket, sd_both) <> 0 then
              {$else} 
                if Shutdown(UDPSocket, sd_both) <> 0 then
              {$endif}   
                begin
                  i := WsaGetLastError;
                  case i of
                    WSAENotConn : ;
                  else
                    NormalMessage('Shutdown socket: ' + WSAErrorString(i));
                  end;
                end;
              if CloseSocket(UDPSocket) <> 0 then
                NormalMessage('Close socket: '
                              + WsaErrorString(WsaGetLastError));
            end;
          {$ifdef fpc}
            Synchronize(@GetParams);
          {$else}     
            Synchronize(GetParams);
          {$endif}   
        end;
    {$ifndef fpc}
      if NetworkState = NetworkTerminating then
        if WSACleanup <> 0 then
          NormalMessage('WinSock cleanup' + WSAErrorString(WSAGetLastError));
    {$endif}   
  end;

end.

