unit zusatz;

{ $Id: zusatz.pas 121 2011-12-18 17:42:06Z Heiner $ }

{$ifdef fpc}
  {$mode objfpc}{$H+}
  {$macro on}
{$endif}

{$ifdef MSWINDOWS}
  { ab Delphi 2005 }
  {$define Delphi4}
  {$define Delphi3}
  {$ifndef fpc}
    {$define Delphi}
  {$endif}
{$endif}

{$ifdef Ver140}
  { Kylix 3 }
  {$define Delphi4}
  {$define Delphi3}
{$endif}

{$ifdef Ver130}
  { Delphi 5 }
  {$define Delphi4}
  {$define Delphi3}
  {$define Delphi}
{$endif}

{$ifdef Ver120}
  { Delphi 4 }
  {$define Delphi4}
  {$define Delphi3}
  {$define Delphi}
  {$define DelphiMax4}
{$endif}

{$ifdef Ver100}
  { Delphi 3 }
  {$define Delphi3}
  {$define Delphi}
  {$define DelphiMax3}
  {$define DelphiMax4}
{$endif}

{$ifdef Ver80}
  { Delphi 1 }
  {$define Delphi}
  {$define DelphiMax3}
  {$define DelphiMax4}
{$endif}

{$ifdef FPC}
  {$define Delphi4}
  {$define Delphi3}
{$endif}

interface

uses
  {$ifdef Win32}
    {$ifdef fpc}
      Controls,
      Video,
      Sockets,
      LCLIntf,
      LConvEncoding,
      LCLVersion,
    {$else}
      ActiveX,
      ShlObj,
    {$endif}
    ShellApi,
    Registry,
    Windows,
    ComObj,
    WinSock,
    WinInet,
  {$else}
    {$ifdef Unix}
      LibC,
      {$ifdef FPC}
        LConvEncoding,
        LCLVersion,
        GetText,
        Video,
        Forms,
        Controls,
        Graphics,
        Dialogs,
        Sockets,
        IdHTTP,
        { IdHTTP is in
          http://www.indyproject.org/sockets/fpc/indy-10.2.0.3.tar.gz }
        Unix,
        BaseUnix,
      {$else}
        QForms,
      {$endif}
    {$else}
      WinTypes,
      WinProcs,
      ToolHelp,
    {$endif}
  {$endif}
  {$ifndef Unix}
    Messages,
    Graphics,
    Forms,
    {$ifndef fpc}
      FileCtrl,
      MPlayer,
    {$endif}
  {$endif}
  SysUtils,
  IniFiles,
  WolfUtil,
  Classes;

const
  {$ifndef fpc}
    MaxPalette                  = $100 - COLOR_ENDCOLORS;
    PathDelim                   = '\';
  {$endif}

  {$ifdef Ver80}
    ERROR_INVALID_FUNCTION      =  1;
    ERROR_FILE_NOT_FOUND        =  2;
    ERROR_PATH_NOT_FOUND        =  3;
    ERROR_TOO_MANY_OPEN_FILES   =  4;
    ERROR_ARENA_TRASHED         =  7;
    ERROR_NOT_ENOUGH_MEMORY     =  8;
    ERROR_INVALID_BLOCK         =  9;
    ERROR_BAD_FORMAT            = 11;
    ERROR_NOT_SAME_DEVICE       = 17;
    ERROR_NO_MORE_FILES         = 18;
    ERROR_BAD_COMMAND           = 22;
    ERROR_CRC                   = 23;
    ERROR_BAD_LENGTH            = 24;
    ERROR_SEEK                  = 25;
    ERROR_NOT_DOS_DISK          = 26;
    ERROR_SECTOR_NOT_FOUND      = 27;
    ERROR_OUT_OF_PAPER          = 28;
    ERROR_WRITE_FAULT           = 29;
    ERROR_READ_FAULT            = 30;
    ERROR_GEN_FAILURE           = 31;
  {$endif}

  { TWin32FindData.dwFileAttributes }
  {$ifdef Ver80}
    FILE_ATTRIBUTE_READONLY            = faReadOnly;
    FILE_ATTRIBUTE_HIDDEN              = faHidden;
    FILE_ATTRIBUTE_SYSTEM              = faSysFile;
    FILE_ATTRIBUTE_DIRECTORY           = faDirectory;
    FILE_ATTRIBUTE_ARCHIVE             = faArchive;
    FILE_ATTRIBUTE_NORMAL              =    $80;
    FILE_ATTRIBUTE_TEMPORARY           =   $100;
    FILE_ATTRIBUTE_COMPRESSED          =   $800;
    FILE_ATTRIBUTE_OFFLINE             =  $1000;
  {$endif}
  {$ifndef Ver180}
    FILE_ATTRIBUTE_DEVICE              =    $40;
    FILE_ATTRIBUTE_SPARSE_FILE         =   $200;
    FILE_ATTRIBUTE_REPARSE_POINT       =   $400;
    FILE_ATTRIBUTE_NOT_CONTENT_INDEXED =  $2000;
    FILE_ATTRIBUTE_ENCRYPTED           =  $4000;
  {$endif}
  FILE_ATTRIBUTE_VIRTUAL               = $10000;

  { FileMode }
  fmReadOnly                           = 0;
  fmWriteOnly                          = 1;
  fmReadWrite                          = 2;

  { Sockets }
  sd_Receive                           = 0;
  sd_Send                              = 1;
  sd_Both                              = 2;
  {$ifdef Unix}
    fIoNbio                            = $5421;
    INTERNET_DEFAULT_HTTP_PORT         = 80;
  {$endif}

  { ASCII Codes }
  A_NUL        = #$00;
  A_SOH        = #$01;
  A_STX        = #$02;
  A_ETX        = #$03;
  A_EOT        = #$04;
  A_ENQ        = #$05;
  A_ACK        = #$06;
  A_BEL        = #$07;
  A_BS         = #$08;
  A_HT         = #$09;
  A_LF         = #$0A;
  A_VT         = #$0B;
  A_FF         = #$0C;
  A_CR         = #$0D;
  A_SO         = #$0E;
  A_SI         = #$0F;
  A_DLE        = #$10;
  A_DC1        = #$11;
  A_XON        = #$11;
  A_DC2        = #$12;
  A_DC3        = #$13;
  A_XOF        = #$13;
  A_DC4        = #$14;
  A_NAK        = #$15;
  A_SYN        = #$16;
  A_ETB        = #$17;
  A_CAN        = #$18;
  A_EM         = #$19;
  A_SUB        = #$1A;
  A_ESC        = #$1B;
  A_FS         = #$1C;
  A_GS         = #$1D;
  A_RS         = #$1E;
  A_US         = #$1F;
  A_SP         = #$20;
  A_DEL        = #$7F;

  A_ISO_UPPER_AE = #$C4;
  A_ISO_UPPER_OE = #$D6;
  A_ISO_UPPER_UE = #$DC;
  A_ISO_LOWER_AE = #$E4;
  A_ISO_LOWER_OE = #$F6;
  A_ISO_LOWER_UE = #$FC;
  A_ISO_SS       = #$DF;
  {A_ISO_C        = #$A9; (c) }

  A_IBM_UPPER_AE = #$8E;
  A_IBM_UPPER_OE = #$99;
  A_IBM_UPPER_UE = #$9A;
  A_IBM_LOWER_AE = #$84;
  A_IBM_LOWER_OE = #$94;
  A_IBM_LOWER_UE = #$81;
  A_IBM_SS       = #$E1;

  {$ifdef Win32}
    { Bits fuer tDCB.Flags }
    fBinary                       = $0001;
    fParity                       = $0002;
    fOutxCtsFlow                  = $0004;
    fOutxDsrFlow                  = $0008;
    fDtrControlDisable            = DTR_CONTROL_DISABLE   shl 4;
    fDtrControlEnable             = DTR_CONTROL_ENABLE    shl 4;
    fDtrControlHandshake          = DTR_CONTROL_HANDSHAKE shl 4;
    fDsrSensitivity               = $0040;
    fTXContinueOnXoff             = $0080;
    fOutX                         = $0100;
    fInX                          = $0200;
    fErrorChar                    = $0400;
    fNull                         = $0800;
    fRtsControlDisable            = RTS_CONTROL_DISABLE   shl 12;
    fRtsControlEnable             = RTS_CONTROL_ENABLE    shl 12;
    fRtsControlHandshake          = RTS_CONTROL_HANDSHAKE shl 12;
    fRtsControlToggle             = RTS_CONTROL_TOGGLE    shl 12;
    fAbortOnError                 = $4000;

    { NT Defined Privileges }
    SE_CREATE_TOKEN_NAME          = 'SeCreateTokenPrivilege';
    SE_ASSIGNPRIMARYTOKEN_NAME    = 'SeAssignPrimaryTokenPrivilege';
    SE_LOCK_MEMORY_NAME           = 'SeLockMemoryPrivilege';
    SE_INCREASE_QUOTA_NAME        = 'SeIncreaseQuotaPrivilege';
    SE_UNSOLICITED_INPUT_NAME     = 'SeUnsolicitedInputPrivilege';
    SE_MACHINE_ACCOUNT_NAME       = 'SeMachineAccountPrivilege';
    SE_TCB_NAME                   = 'SeTcbPrivilege';
    SE_SECURITY_NAME              = 'SeSecurityPrivilege';
    SE_TAKE_OWNERSHIP_NAME        = 'SeTakeOwnershipPrivilege';
    SE_LOAD_DRIVER_NAME           = 'SeLoadDriverPrivilege';
    SE_SYSTEM_PROFILE_NAME        = 'SeSystemProfilePrivilege';
    SE_SYSTEMTIME_NAME            = 'SeSystemtimePrivilege';
    SE_PROF_SINGLE_PROCESS_NAME   = 'SeProfileSingleProcessPrivilege';
    SE_INC_BASE_PRIORITY_NAME     = 'SeIncreaseBasePriorityPrivilege';
    SE_CREATE_PAGEFILE_NAME       = 'SeCreatePagefilePrivilege';
    SE_CREATE_PERMANENT_NAME      = 'SeCreatePermanentPrivilege';
    SE_BACKUP_NAME                = 'SeBackupPrivilege';
    SE_RESTORE_NAME               = 'SeRestorePrivilege';
    SE_SHUTDOWN_NAME              = 'SeShutdownPrivilege';
    SE_DEBUG_NAME                 = 'SeDebugPrivilege';
    SE_AUDIT_NAME                 = 'SeAuditPrivilege';
    SE_SYSTEM_ENVIRONMENT_NAME    = 'SeSystemEnvironmentPrivilege';
    SE_CHANGE_NOTIFY_NAME         = 'SeChangeNotifyPrivilege';
    SE_REMOTE_SHUTDOWN_NAME       = 'SeRemoteShutdownPrivilege';
    SE_ENABLE_DELEGATION_NAME     = 'SeEnableDelegationPrivilege';
    SE_IMPERSONATE_PRIVILEGE      = 'SeImpersonatePrivilege';
    SE_MANAGE_VOLUME_NAME         = 'SeManageVolumePrivilege';
    SE_SYNC_AGENT_NAME            = 'SeSyncAgentPrivilege';
    SE_UNDOCK_NAME                = 'SeUndockPrivilege';

    HKCU_ShellFolders             =
      'Software\Microsoft\Windows\CurrentVersion\Explorer\Shell Folders';
    DesktopFolder                 = 'Desktop';
    ProgramsFolder                = 'Programs';
    StartMenuFolder               = 'Start Menu';
    PersonalFolder                = 'Personal';
    LinkExt                       = '.lnk';
  {$else}
    { RS232 Flags }
    fBinary                       = $0001;
    fRtsDisable                   = $0002;
    fParity                       = $0004;
    fOutxCtsFlow                  = $0008;
    fOutxDsrFlow                  = $0010;
    fDtrDisable                   = $0080;
    fOutX                         = $0100;
    fInX                          = $0200;
    fPeChar                       = $0400;
    fNull                         = $0800;
    fChEvt                        = $1000;
    fDtrflow                      = $2000;
    fRtsflow                      = $4000;
  {$endif}

  {$ifdef DelphiMax4}
    { Windows Messages }
    wm_GetObject                       = $003D;
    wm_MenuRButtonUp                   = $0122;
    wm_MenuDrag                        = $0123;
    wm_MenuGetObject                   = $0124;
    wm_UninitMenuPopup                 = $0125;
    wm_MenuCommand                     = $0126;
    wm_ChangeUiState                   = $0127;
    wm_UpdateUiState                   = $0128;
    wm_QueryUiState                    = $0129;
    wm_Ime_Request                     = $0288;
  {$endif}

  {$ifdef Ver80}
    { Delphi 1: Windows Messages }
    wm_SetHotkey                  = $0032;
    wm_GetHotkey                  = $0033;
    wm_CopyData                   = $004A;
    wm_CancelJournal              = $004B;
    wm_Notify                     = $004E;
    wm_InputLangChangeRequest     = $0050;
    wm_InputLangChange            = $0051;
    wm_TCard                      = $0052;
    wm_Help                       = $0053;
    wm_UserChanged                = $0054;
    wm_NotifyFormat               = $0055;
    wm_ContextMenu                = $007B;
    wm_StyleChanging              = $007C;
    wm_StyleChanged               = $007D;
    wm_DisplayChange              = $007E;
    wm_GetIcon                    = $007F;
    wm_SetIcon                    = $0080;
    wm_CtlColorMsgbox             = $0132;
    wm_CtlColorEdit               = $0133;
    wm_CtlColorListbox            = $0134;
    wm_CtlColorBtn                = $0135;
    wm_CtlColorDlg                = $0136;
    wm_CtlColorScrollbar          = $0137;
    wm_CtlColorStatic             = $0138;
    wm_Ime_StartComposition       = $010D;
    wm_Ime_EndComposition         = $010E;
    wm_Ime_Composition            = $010F;
    wm_MouseWheel                 = $020A;
    wm_EnterMenuLoop              = $0211;
    wm_ExitMenuLoop               = $0212;
    wm_NextMenu                   = $0213;
    wm_Sizing                     = $0214;
    wm_CaptureChanged             = $0215;
    wm_Moving                     = $0216;
    wm_PowerBroadcast             = $0218;
    wm_DeviceChange               = $0219;
    wm_Ime_SetContext             = $0281;
    wm_Ime_Notify                 = $0282;
    wm_Ime_Control                = $0283;
    wm_Ime_CompositionFull        = $0284;
    wm_Ime_Select                 = $0285;
    wm_Ime_Char                   = $0286;
    wm_Ime_KeyDown                = $0290;
    wm_Ime_KeyUp                  = $0291;
    wm_EnterSizeMove              = $0231;
    wm_ExitSizeMove               = $0232;
    wm_MdiRefreshMenu             = $0234;
    wm_MouseHover                 = $02A1;
    wm_MouseLeave                 = $02A3;
    wm_Hotkey                     = $0312;
    wm_Print                      = $0317;
    wm_PrintClient                = $0318;
    wm_HandHeldFirst              = $0358;
    wm_HandHeldLast               = $035F;
    wm_App                        = $8000;

    { Virtual Keys }
    vk_LWin                       = $5B;
    vk_RWin                       = $5C;
    vk_Apps                       = $5D;
    vk_LShift                     = $A0;
    vk_RShift                     = $A1;
    vk_LControl                   = $A2;
    vk_RControl                   = $A3;
    vk_LMenu                      = $A4;
    vk_RMenu                      = $A5;
    vk_ProcessKey                 = $E5;
    vk_Attn                       = $F6;
    vk_CrSel                      = $F7;
    vk_ExSel                      = $F8;
    vk_ErEOF                      = $F9;
    vk_Play                       = $FA;
    vk_Zoom                       = $FB;
    vk_NoName                     = $FC;
    vk_Pa1                        = $FD;
    vk_OEM_Clear                  = $FE;
  {$endif}

  {$ifdef Unix}
    wsaEIntr           = EIntr;
    wsaEBadF           = EBadF;
    wsaEAcces          = EAcces;
    wsaEFault          = EFault;
    wsaEInval          = EInval;
    wsaEMFile          = EMFile;
    wsaEWouldBlock     = EWouldBlock;
    wsaEInProgress     = EInProgress;
    wsaEAlready        = EAlready;
    wsaENotSock        = ENotSock;
    wsaEDestAddrReq    = EDestAddrReq;
    wsaEMsgSize        = EMsgSize;
    wsaEPrototype      = EPrototype;
    wsaENoProtoOpt     = ENoProtoOpt;
    wsaEProtoNoSupport = EProtoNoSupport;
    wsaESockTNoSupport = ESockTNoSupport;
    wsaEOpNotSupp      = EOpNotSupp;
    wsaEPFNoSupport    = EPFNoSupport;
    wsaEAFNoSupport    = EAFNoSupport;
    wsaEAddrInUse      = EAddrInUse;
    wsaEAddrNotAvail   = EAddrNotAvail;
    wsaENetDown        = ENetDown;
    wsaENetUnreach     = ENetUnreach;
    wsaENetReset       = ENetReset;
    wsaEConnAborted    = EConnAborted;
    wsaEConnReset      = EConnReset;
    wsaENoBufs         = ENoBufs;
    wsaEIsConn         = EIsConn;
    wsaENotConn        = ENotConn;
    wsaEShutdown       = EShutdown;
    wsaETooManyRefs    = ETooManyRefs;
    wsaETimedOut       = ETimedOut;
    wsaEConnRefused    = EConnRefused;
    wsaELoop           = ELoop;
    wsaENameTooLong    = ENameTooLong;
    wsaEHostDown       = EHostDown;
    wsaEHostUnreach    = EHostUnreach;
    wsaENotEmpty       = ENotEmpty;
    wsaEUsers          = EUsers;
    wsaEDQuot          = EDQuot;
    wsaEStale          = EStale;
    wsaERemote         = ERemote;
    wsaHost_not_found  = Host_not_found;
    wsaTry_again       = Try_again;
    wsaNo_recovery     = No_recovery;
    wsaNo_data         = No_data;
  {$endif}

  GermanLongMonthNames : Array[1 .. 12] of String =
    ('Januar',
     'Februar',
     'Mrz',
     'April',
     'Mai',
     'Juni',
     'Juli',
     'August',
     'September',
     'Oktober',
     'November',
     'Dezember');

  EnglishLongMonthNames : Array[1 .. 12] of String =
    ('January',
     'February',
     'March',
     'April',
     'May',
     'June',
     'July',
     'August',
     'September',
     'October',
     'November',
     'December');

type
  tZusatzSize   = LongestInt;

  tFFcontinue   = function(Name : String; Size : tZusatzSize; error : String) :
                    Boolean of object;

  tFileOrFolder = (ffParam,
                   ffFile,
                   ffFolder,
                   ffNotFound,
                   ffReadOnly,
                   ffDifferent,
                   ffRecursive,
                   ffExists,
                   ffForceDirs,
                   ffCanceled,
                   ffVolumeFull,
                   ffReadError);

  sFileOrFolder  = Set of tFileOrFolder;

  tSetServEnt    = (SetServEntOk,
                    SetServEntNotFound,
                    SetServEntErrCreateTemp,
                    SetServEntErrDelPrev,
                    SetServEntErrRename,
                    SetServEntInvalidPathLen);

  {$ifdef Delphi3}
    tAddrList         = Array[0..0] of pInAddr;
    tNameList         = Array[0..0] of PChar;
    pAddrList         = ^tAddrList;
    pNameList         = ^tNameList;
  {$endif}

  {$ifdef Unix}
    tBitmapFileHeader = packed record
      bfType          : Word;
      bfSize          : DWord;
      bfReserved1     : Word;
      bfReserved2     : Word;
      bfOffBits       : DWord;
    end;

    tBitmapInfoHeader = record
      biSize          : DWORD;
      biWidth         : Longint;
      biHeight        : Longint;
      biPlanes        : WORD;
      biBitCount      : WORD;
      biCompression   : DWORD;
      biSizeImage     : DWORD;
      biXPelsPerMeter : Longint;
      biYPelsPerMeter : Longint;
      biClrUsed       : DWORD;
      biClrImportant  : DWORD;
    end;
  {$endif}

  tBmpHeader        = packed Record
    FileHeader      : tBitmapFileHeader;
    InfoHeader      : tBitmapInfoHeader;
  end;

function  CompilerName : String;
function  Ordinal(i : LongInt) : String;
function  IsIni(s : String) : Boolean;
function  ISOtoIBM(s : String) : String;
function  IBMtoISO(s : String) : String;
Function  IntMin(a, b : LongInt) : LongInt;
Function  IntMax(a, b : LongInt) : LongInt;
function  GetTable(var s : String; var Col, Row : Integer) : String;
function  EscapeAmpersand(s : String) : String;
function  MsgBox(Text, Caption : String; Flags : LongInt) : Integer;

{$ifndef fpc}
procedure GetTasks(List : tStrings);
function  GetHandleFromStr(s : String) : hWnd;
function  GetHandleFromCaption(Caption : String) : HWnd;
function  CreateRegenbogen : pLogPalette;
function  PaletteSize(Entries : Word) : Word;
procedure FreeRegenbogen(var Palette : pLogPalette);
procedure AnimateRegenbogen(var Palette : pLogPalette; Handle : hPalette;
                            Offset : Integer);
function  MessageName(m : Word) : String;
{$endif}

function  CmdCount : Integer;
function  CmdStr(Index : Integer) : String;
function  FontStyleString(FontStyle : tFontStyles) : String;
function  FontStyleSet(FontStyle : String) : tFontStyles;
function  ReplaceEscape(t : String) : String;
function  ASCIIname(c : char) : String;
function  GetMyName : String;
function  MakeShell(s : String) : String;
{ In Win16 kann man keine Parameter z.B "a b" definieren.
  In Win32 kann man zwar Parameter z.B. "a b" definieren, aber nicht ""
}
function  MakePath(Path, FileName : String) : String;
function  MakeFind(Pfad : String) : String;
function  ReadBmpHeader(var Datei : File; var Header : tBmpHeader) : Boolean;
procedure GetBmpDPI(FileName : String; var xDPI, yDPI : Integer);
procedure SetBmpDPI(FileName : String; xDPI, yDPI : Integer);

{$ifndef Unix}
  function  DateiGroesse(SearchRec : tSearchRec) : tZusatzSize;
  function  FileGetSize(FileName : String; var Size : tZusatzSize) : Boolean;
  function  OrdnerGroesse(Ordner : String; FFcontinue : tFFcontinue) :
              tZusatzSize;
  function  CopyFileOrFolder(FromName, ToName : String; Allowed : sFileOrFolder;
                             FFcontinue : tFFcontinue) : sFileOrFolder;
                             { returns error reason }
  function  DelFileOrFolder(fName : String; Allowed : sFileOrFolder;
                            FFcontinue : tFFcontinue) : sFileOrFolder;
                            { returns error reason }
  function  CopyFileEx(FromName, ToName : String;
                       FailIfExists : Boolean;
                       FFcontinue : tFFcontinue) : sFileOrFolder;
  function  FileOrFolderItems(Items : sFileOrFolder) : String;
{$endif}

{$ifdef Delphi4}
{ Delphi 3 hat kein overload und kein int64 }
function  SaveFormToIni(Form : tForm) : Boolean; overload;
function  LoadFormFromIni(Form : tForm) : Boolean; overload;
function  SaveFormToIni(Form : tForm; Name : String) : Boolean; overload;
function  LoadFormFromIni(Form : tForm; Name : String) : Boolean; overload;
{$endif}

{$ifdef Delphi3}
function  SetServEnt(PortName, Protocol : String;
                     PortNum            : Word; { Host Byte order, 0 zum Loeschen }
                     Comment            : String) : tSetServEnt;
function  InetUpdate(var Archiv       : String;
                     Jahr, Monat, Tag : Integer;
                     var NewAvailable : Boolean) : Boolean;
function  GetIpName(Ip : LongInt) : String;
function  GetIpAddr(HostName : String) : LongInt;
function  WSAErrorString(Error : Integer) : String;
{$endif}

{$ifdef DelphiMax3}
  { Delphi 3 hat kein overload,
    in Delphi 1 ist Form.Name ist leer! }
  function SaveFormToIni(Form   : tForm; Name : String) : Boolean;
  function LoadFormFromIni(Form : tForm; Name : String) : Boolean;
{$endif}

{$ifdef Ver80}
  function SysErrorMessage(i : Integer) : String;
  function CompareMem(p1, p2 : Pointer; Len : tZusatzSize) : Boolean;
{$else}
  function ExtGraphicExtension(GraphicClass: TGraphicClass;
                               Index : Integer): string;
{$endif}

{$ifdef Win32}
  {$ifndef  fpc}
    function  GetLink(Link : WideString;
                      var Path, Description, WorkDir : String) : String;
    function  FileTimeToDateTime(FileTime : tFileTime) : tDateTime;
    procedure DateTimeToFileTime(DateTime : tDateTime; var FileTime : tFileTime);
    function  SetFileTimeOnName(DateiName : String;
                                CreationTime, LastAccessTime,
                                LastWriteTime : tFileTime) : Boolean;
    procedure MoveFormTo(Form : tForm; Left, Top, Width, Height : Integer;
                         WindowState : tWindowState);
    function  Exec(Wnd : hWnd; Shell, Params : String; nShow : Integer) : Boolean;
    function  MkLink(Path, Description, WorkDir : String; Link : WideString) :
                String;
    function  ExitWindowsWithMessage(Flags : uInt; var Initiated : Boolean) :
                String;
  {$endif}
  function    SaveFormToReg(Form : tForm) : Boolean;
  function    LoadFormFromReg(Form : tForm) : Boolean;
  function    RegOpen(var Reg : tRegistry; RelativePath : String) : Boolean;
{$else}
  function    GetLastError : Integer;
  {$ifndef FPC}
    function  GetHandleFromTaskName(TaskName : String) : tHandle;
    function  CheckCommErrors(Port : Integer) : String;
    function  OpenCommError(Error : Integer) : String;
  {$endif}
{$endif}

{$ifdef Unix}
  function ReadPipe(Cmd : String; var Output : String) : Boolean;
  function GetProxyParams(var Proxy, User, Password : String;
                          var Port : Integer) : Boolean;
  function WSAGetLastError : Integer;
{$else}
  function WinExecError(Error : Word) : String;
{$endif}

var IniName,
    StartDir : String;

implementation

{$ifdef Win32}
  const Wildcard  = '*';
{$else}
  const Wildcard  = '*.*';
{$endif}

function CompilerName : String;
  {$ifndef fpc}
    const Delphi      = ' Delphi ';
          Kylix       = ' Kylix 3';
          CBuilder    = ' C++ Builder ';
          Borland     = 'Borland';
          CodeGear    = 'CodeGear';
          Embarcadero = 'Embarcadero';
  {$endif}
  var     s           : String;
          {$ifdef MSWindows}
            {$ifndef fpc}
              t       : String;
            {$endif}
          {$endif}
  begin
    s := '';
    {$ifdef Ver80}
      s := Borland + Delphi   + '1';
    {$endif}
    {$ifdef Ver90}
      s := Borland + Delphi   + '2';
    {$endif}
    {$ifdef Ver93}
      s := Borland + CBuilder + '1';
    {$endif}
    {$ifdef Ver100}
      s := Borland + Delphi   + '3';
    {$endif}
    {$ifdef Ver110}
      s := Borland + CBuilder + '3';
    {$endif}
    {$ifdef Ver120}
      s := Borland + Delphi   + '4';
    {$endif}
    {$ifdef Ver125}
      s := Borland + CBuilder + '4';
    {$endif}
    {$ifdef Ver130}
      s := Borland + Delphi   + '5';
    {$endif}
    {$ifdef Ver140}
      case SysLocale.PriLangID of
        Lang_German :
          s := Borland + Delphi + '6 oder' + Kylix;
      else
        s := Borland + Delphi + '6 or' + Kylix;
      end;
    {$endif}
    {$ifdef Ver150}
      s := Borland + Delphi + '7';
    {$endif}
    {$ifdef Ver170}
      s := Borland + Delphi + '2005';
    {$endif}
    {$ifdef Ver180}
      {$ifndef Ver185}
        s := Borland  + Delphi + '2006';
      {$else}
        s := CodeGear + Delphi + '2007';
      {$endif}
    {$endif}
    {$ifdef Ver200}
      s := Embarcadero + Delphi + '2008';
    {$endif}
    {$ifdef Ver210}
      s := Embarcadero + Delphi + '2010';
    {$endif}
    {$ifdef Ver220}
      s := Embarcadero + Delphi + 'XE';
    {$endif}
    {$ifdef fpc}
      s := 'Free Pascal ' + IntToStr(FPC_VERSION) + '.'
                          + IntToStr(FPC_RELEASE) + '.'
                          + IntToStr(FPC_PATCH)
           + '/Lazarus '  + LCL_Version;
    {$endif}
    if s = '' then
      begin
        {$ifdef MSWindows}
          {$ifndef fpc}
            str(CompilerVersion : 1 : 1, t);
            if CompilerVersion > 18.0 then
              if CompilerVersion >= 20.0 then
                s := s + Embarcadero + Delphi + t
              else
                s := s + CodeGear    + Delphi + t
            else
              s :=   s + Borland     + Delphi + t;
          {$endif}
        {$endif}
      end;
    if s = '' then
      case SysLocale.PriLangID of
        Lang_German :
          s := 'unbekannt';
      else
        s := 'unknown';
      end;
    CompilerName := s;
  end;

function SaveFormToIniWithName(Form : tForm; Name : String) : Boolean;
  var    IniFile : tIniFile;
  begin
    SaveFormToIniWithName := True;
    try
      IniFile := tIniFile.Create(IniName);
    except
      IniFile := Nil;
    end;
    if IniFile = Nil then
      SaveFormToIniWithName := False
    else
      try
        if Form.BorderStyle = bssizeable then
          IniFile.WriteInteger(Name, 'WindowState', ord(Form.WindowState));
        if Form.WindowState = wsNormal then
          begin
            IniFile.WriteInteger(Name, 'Top',  Form.Top);
            IniFile.WriteInteger(Name, 'Left', Form.Left);
            if Form.BorderStyle = bsSizeable then
              begin
                IniFile.WriteInteger(Name, 'Height', Form.Height);
                IniFile.WriteInteger(Name, 'Width',  Form.Width);
              end;
          end;
      finally
        IniFile.Free;
      end;
  end;

procedure MoveFormTo(Form : tForm; Left, Top, Width, Height : Integer;
                     WindowState : tWindowState);
  const   Min = 100;
  begin
    Height    := IntMin(Height, Screen.Height);
    Width     := IntMin(Width,  Screen.Width);
    Left      := IntMin(IntMax(Left, Min - Width), Screen.Width  - Min);
    Top       := IntMin(IntMax(Top,  0),           Screen.Height - Min);
    Form.Left := Left;
    Form.Top  := Top;
    if Form.BorderStyle = bsSizeable then
      begin
        Form.Height      := Height;
        Form.Width       := Width;
        if WindowState in [wsNormal, wsMinimized, wsMaximized] then
          Form.WindowState := WindowState;
      end;
  end;

function LoadFormFromIniWithName(Form : tForm; Name : String) : Boolean;
  var    IniFile : tIniFile;
         Left,
         Top     : Integer;
  begin
    LoadFormFromIniWithName := True;
    try
      IniFile := tIniFile.Create(IniName);
    except
      IniFile := Nil;
    end;
    if IniFile = Nil then
      LoadFormFromIniWithName := False
    else
      try
        if form.Parent = Nil then
          begin
            Left := (Screen.Width  - Form.Width)  div 2;
            Top  := (Screen.Height - Form.Height) div 2;
          end
        else
          begin
            Left := (Form.Parent.ClientWidth  - Form.Width)  div 2;
            Top  := (Form.Parent.ClientHeight - Form.Height) div 2;
          end;
        MoveFormTo(Form,
                   IniFile.ReadInteger(Name, 'Left',   Left),
                   IniFile.ReadInteger(Name, 'Top',    Top),
                   IniFile.ReadInteger(Name, 'Width',  Form.Width),
                   IniFile.ReadInteger(Name, 'Height', Form.Height),
                   tWindowState(
                     IniFile.ReadInteger(
                       Name, 'WindowState', ord(Form.WindowState))));
      finally
        IniFile.Free;
      end;
  end;

{$ifdef Ver80}
  { Delphi 1 }
  function SysErrorMessage(i : Integer) : String;
    begin
      SysErrorMessage := 'Error';
    end;

  function DateiGroesse(SearchRec : tSearchRec) : tZusatzSize;
    begin
      DateiGroesse := SearchRec.Size;
    end;

  function CompareMem(p1, p2 : Pointer; Len : tZusatzSize) : Boolean;
    type   a      = Array[0..0] of Byte;
    var    Gleich : Boolean;
           i      : tZusatzSize;
    begin
      Gleich := True;
      i      := 0;
      while (i < Len) and Gleich do
        begin
          {$R-}
            if a(p1^)[i] <> a(p2^)[i] then
              Gleich := False;
          {$R+}
          inc(i);
        end;
      CompareMem := Gleich;
    end;
{$else}
function ExtGraphicExtension(GraphicClass: TGraphicClass;
                             Index : Integer): string;
  const  Stern = '*.';
         Trenn = ';';
  var    s, t,
         u, v  : String;
         i, p  : Integer;
  begin
    if Index = 1 then
      u := GraphicExtension(GraphicClass)
    else
      begin
        i := 1;
        u := '';
        s := GraphicFileMask(GraphicClass);
        t := GraphicExtension(GraphicClass);
        while (u = '') and (s > '') do
          begin
            if Copy(s, 1, length(Stern)) = Stern then
              s := Copy(s, length(Stern) + 1, Length(s));
            p := Pos(Trenn, s);
            if p = 0 then
              begin
                v := s;
                s := '';
              end
            else
              begin
                v := Copy(s, 1, p - 1);
                s := Copy(s, p + 1, length(s));
              end;
            if v > '' then
              if v <> t then
                begin
                  inc(i);
                  if i = Index then
                    u := v;
                end;
          end;
      end;
    ExtGraphicExtension := u;
  end;
{$endif}

{$ifdef DelphiMax3}
function SaveFormToIni(Form : tForm; Name : String) : Boolean;
  begin
    SaveFormToIni := SaveFormToIniWithName(Form, Name);
  end;

function LoadFormFromIni(Form : tForm; Name : String) : Boolean;
  begin
    LoadFormFromIni := LoadFormFromIniWithName(Form, Name);
  end;
{$endif}

{$ifdef Delphi4}
function SaveFormToIni(Form : tForm) : Boolean;
  begin
    SaveFormToIni := SaveFormToIniWithName(Form, Form.Name);
  end;

function LoadFormFromIni(Form : tForm) : Boolean;
  begin
    LoadFormFromIni := LoadFormFromIniWithName(Form, Form.Name);
  end;

function SaveFormToIni(Form : tForm; Name : String) : Boolean;
  begin
    SaveFormToIni := SaveFormToIniWithName(Form, Name);
  end;

function LoadFormFromIni(Form : tForm; Name : String) : Boolean;
  begin
    LoadFormFromIni := LoadFormFromIniWithName(Form, Name);
  end;
{$endif}

{$ifdef Delphi3}
function GetIpName(Ip : LongInt) : String;
  var    HostEnt : pHostEnt;
  begin
    HostEnt := GetHostByAddr(@IP, sizeof(IP), 0);
    if HostEnt = Nil then
      {$ifdef fpc}
        {$ifdef Unix}
          GetIpName := Inet_NtoA(LIBC.tInAddr(Ip))
        {$else}
          GetIpName := Inet_NtoA(tInAddr(Ip))
        {$endif}
      {$else}
        GetIpName := Inet_NtoA(tInAddr(Ip))
      {$endif}
    else
      GetIpName := HostEnt^.h_name;
  end;

function GetIpAddr(HostName : String) : LongInt;
  var    HostEnt : pHostEnt;
  begin
    Hostent := GetHostByName(pChar(HostName));
    if HostEnt = Nil then
      GetIpAddr := longint(InAddr_None)
    else
      GetIpAddr := pInAddr(HostEnt^.h_addr^)^.s_addr;
  end;
{$endif}

{$ifdef Unix}
  function WSAGetLastError : Integer;
    begin
      WSAGetLastError := SocketError;
    end;

  function ReadPipe(Cmd : String; var Output : String) : Boolean;
    var    Pipe : Text;
    begin
      ReadPipe := False;
      Output   := '';
      TTextRec(Pipe).bufend := 0;
      POpen(Pipe, Cmd, 'R');
      if fpgeterrno = 0 then
        begin
          ReadLn(Pipe, Output);
          if PClose(Pipe) = 0 then
            ReadPipe := True;
        end;
    end;

  function GetProxyParams(var Proxy, User, Password : String;
                          var Port : Integer) : Boolean;
    var    s  : String;
           p  : Integer;
           ok : Boolean;
    begin
      Proxy    := '';
      User     := '';
      Password := '';
      Port     := 0;
      ok       := False;
      s        := UpperCase(GetEnvironmentVariable('DESKTOP_SESSION'));
      if s = 'GNOME' then
        begin
          ok := ReadPipe('gconftool-2 -g /system/http_proxy/use_http_proxy', s);
          if ok then
            begin
              s  := UpperCase(s);
              ok := (s = 'TRUE') or (s = 'FALSE');
            end;
          if ok then
            if s = 'TRUE' then
              begin
                if not ReadPipe('gconftool-2 -g /system/http_proxy/host', Proxy) then
                  ok := False;
                if ok then
                  if not ReadPipe('gconftool-2 -g /system/http_proxy/port', s) then
                    ok := False;
                if ok then
                  if s > '' then
                    try
                      Port := StrToInt(s);
                    except
                      ok := False;
                    end;
                if ok then
                  if not ReadPipe(
                    'gconftool-2 -g /system/http_proxy/use_authentication', s)
                  then
                    ok := False;
                if ok then
                  begin
                    s  := UpperCase(s);
                    ok := (s = 'TRUE') or (s = 'FALSE');
                  end;
                if ok then
                  if s = 'TRUE' then
                    begin
                      if not ReadPipe(
                        'gconftool-2 -g /system/http_proxy/authentication_user', User)
                      then
                        ok := False;
                      if not ReadPipe(
                        'gconftool-2 -g /system/http_proxy/authentication_password',
                        Password)
                      then
                        ok := False;
                    end;
              end;
        end
      else
        if s = 'KDE' then
          begin
            ok := ReadPipe('kreadconfig --file kioslaverc '
                           + '--group "Proxy Settings" --key ProxyType', s);
            if ok then
              if s = '1' then
                ok := ReadPipe('kreadconfig --file kioslaverc '
                               + '--group "Proxy Settings" --key httpProxy', s)
              else
                ok := False;
            if ok then
              if s > '' then
                begin
                  p := length(s);
                  while p > 0 do
                    if s[p] = '/' then
                      begin
                        s := Copy(s, 1, p - 1);
                        dec(p);
                      end
                    else
                      p := 0;
                  p := Pos('/', s);
                  while p > 0 do
                    begin
                      s := Copy(s, p + 1, length(s));
                      p := Pos('/', s);
                    end;
                  p := Pos('@', s);
                  if p = 0 then
                    Proxy := s
                  else
                    begin
                      Proxy := Copy(s, p + 1, length(s));
                      User  := Copy(s, 1, p - 1);
                    end;
                  p := Pos(':', Proxy);
                  if p = 0 then
                    ok := False
                  else
                    begin
                      s     := Copy(Proxy, p + 1, length(Proxy));
                      Proxy := Copy(Proxy, 1, p - 1);
                      try
                        Port := StrToInt(s);
                      except
                        ok := False;
                      end;
                    end;
                  if User > '' then
                    begin
                      p := Pos(':', User);
                      if p = 0 then
                        ok := False
                      else
                        begin
                          Password := Copy(User, p + 1, length(User));
                          User     := Copy(User, 1, p - 1);
                        end;
                    end;
                end;
          end;
      if not ok then
        begin
          Proxy    := '';
          User     := '';
          Password := '';
          Port     := 0;
        end;
      GetProxyParams := ok;
    end;

{$else}

function FileGetSize(FileName : String; var Size : tZusatzSize) : Boolean;
  var    SearchRec : tSearchRec;
  begin
    FileGetSize := False;
    Size        := 0;
    if FindFirst(FileName, faAnyFile, SearchRec) = 0 then
      if SearchRec.Name = ExtractFileName(FileName) then
        begin
          FileGetSize := True;
          Size        := DateiGroesse(SearchRec);
        end;
    FindClose(SearchRec);
  end;

function OrdnerGroesse(Ordner : String; FFcontinue : tFFcontinue) : tZusatzSize;
  var    SearchRec : tSearchRec;
         i         : Integer;
         Umfang,
         Groesse   : tZusatzSize;
  begin
    Umfang         := 0;
    SearchRec.Attr := 0;
    FillChar(SearchRec, sizeof(SearchRec), 0);
    if FFcontinue(Ordner, 0, '') then
      if FindFirst(MakeFind(Ordner), faAnyFile, SearchRec) <> 0 then
        begin
          FFcontinue(Ordner, 0, SysErrorMessage(GetLastError)
                                + ' in retrieving size of ' + Ordner);
          FindClose(SearchRec);
        end
      else
        begin
          FindClose(SearchRec);
          if (SearchRec.Attr and (faDirectory or faVolumeID)) = 0 then
            begin
              Groesse := DateiGroesse(SearchRec);
              FFcontinue(Ordner, Groesse, '');
              Inc(Umfang, Groesse);
              FindClose(SearchRec);
            end
          else
            begin
              i := FindFirst(MakePath(Ordner, Wildcard), faAnyFile, SearchRec);
              while (i = 0) and
                    FFContinue(MakePath(Ordner, SearchRec.Name), 0, '')
              do
                begin
                  if SearchRec.Name <> '.' then
                    if SearchRec.Name <> '..' then
                      Inc(Umfang,
                          OrdnerGroesse(MakePath(Ordner, SearchRec.Name),
                                        FFcontinue));
                  i := FindNext(SearchRec);
                end;
              i := abs(i);
              if (i <> 0) and (i <> ERROR_NO_MORE_FILES) then
                FFcontinue(Ordner, 0, SysErrorMessage(GetLastError)
                                      + ' in retrieving size of folder '
                                      + Ordner);
              FindClose(SearchRec);
            end;
        end;
    OrdnerGroesse := Umfang;
  end;

function WinExecError(Error : Word) : String;
  begin
    if Error >= $20 then
      WinExecError := 'Instance = $' + IntToHex(Error, 4)
    else
      case Error of
        0:
          WinExecError := 'system out of memory, executable corrupt, '
                          + 'or relocations invalid';
        ERROR_INVALID_FUNCTION:
          WinExecError := 'invalid function';
        ERROR_FILE_NOT_FOUND:
          WinExecError := 'the system cannot find the file specified';
        ERROR_PATH_NOT_FOUND:
          WinExecError := 'the system cannot find the path specified';
        ERROR_TOO_MANY_OPEN_FILES:
          WinExecError := 'the system cannot open the file';
        5:
          WinExecError := 'conflict on dynamically linking a task or sharing '
                           + 'or network-protection error';
        6: WinExecError := 'library requires separate data segments '
                           + 'for each task';
        ERROR_ARENA_TRASHED:
          WinExecError := 'the storage control blocks were destroyed';
        ERROR_NOT_ENOUGH_MEMORY:
          WinExecError := 'not enough storage is available '
                          + 'to process this command';
        ERROR_INVALID_BLOCK:
          WinExecError := 'the storage control block address is invalid';
        10:
          WinExecError := 'wrong WINDOWS version';
        ERROR_BAD_FORMAT:
          WinExecError := 'an attempt was made to load a program '
                          + 'with an incorrect format';
        12:
          WinExecError := 'wrong operating system';
        13:
          WinExecError := 'DOS 4.0 application';
        14:
          WinExecError := 'unknown executable type';
        15:
          WinExecError := 'cannot load real mode application';
        16:
          WinExecError := 'cannot load 2nd instance with '
                           + 'multiple writeable data segments';
        ERROR_NOT_SAME_DEVICE:
          WinExecError := 'the system cannot move the file '
                          + 'to a different disk drive';
        ERROR_NO_MORE_FILES:
          WinExecError := 'there are no more files';
        19:
          WinExecError := 'cannot load compressed executable';
        20:
          WinExecError := 'invalid DLL';
        21:
          WinExecError := 'application requires 32-bit extensions';
        ERROR_BAD_COMMAND:
          WinExecError := 'the device does not recognize the command';
        ERROR_CRC:
          WinExecError := 'data error (cyclic redundancy check)';
        ERROR_BAD_LENGTH:
          WinExecError := 'the program issued a command '
                          + 'but the command length is incorrect';
        ERROR_SEEK:
          WinExecError := 'the drive cannot locate a specific area '
                          + 'or track on the disk';
        ERROR_NOT_DOS_DISK:
          WinExecError := 'the specified disk or diskette cannot be accessed';
        ERROR_SECTOR_NOT_FOUND:
          WinExecError := 'the drive cannot find the sector requested';
        ERROR_OUT_OF_PAPER:
          WinExecError := 'the printer is out of paper';
        ERROR_WRITE_FAULT:
          WinExecError := 'the system cannot write to the specified device';
        ERROR_READ_FAULT:
          WinExecError := 'the system cannot read from the specified device';
        ERROR_GEN_FAILURE:
          WinExecError := 'a device attached to the system is not functioning';
      else
        WinExecError := 'error $' + IntToHex(Error, 4)
      end;
  end;
{$endif}

{$ifdef  Delphi3}
{ oder hoeher }
{$ifndef Unix}
function DateiGroesse(SearchRec : tSearchRec) : tZusatzSize;
  var    s : tZusatzSize;
  begin
    s            := SearchRec.FindData.nFileSizeHigh;
    s            := s shl 32 + SearchRec.FindData.nFileSizeLow;
    DateiGroesse := s;
  end;
{$endif}

{$ifndef FPC}
function GetLink(Link : WideString; var Path, Description, WorkDir : String) :
           String;
  var    ShellLink    : iShellLink;
         PersistFile  : iPersistFile;
         Buffer       : Array[0..$ff] of char;
         FindData     : TWin32FindData;
  begin
    GetLink     := '';
    Path        := '';
    Description := '';
    WorkDir     := '';
    Coinitialize(nil);
    try
      OLECheck(CoCreateInstance(CLSID_ShellLink, Nil, CLSCTX_INPROC_SERVER,
                                IShellLink, ShellLink));
      PersistFile := ShellLink as IPersistFile;
      OLECheck(PersistFile.Load(PWideChar(Link), 0));
      OLECheck(ShellLink.GetPath(Buffer, sizeof(Buffer), FindData, 0));
      Path := StrPas(Buffer);
      OLECheck(ShellLink.GetDescription(Buffer, sizeof(Buffer)));
      Description := StrPas(Buffer);
      OLECheck(ShellLink.GetWorkingDirectory(Buffer, sizeof(Buffer)));
      WorkDir     := StrPas(Buffer);
      PersistFile := Nil;
      ShellLink   := Nil;
    except
      GetLink := SysErrorMessage(GetLastError);
    end;
  end;

Function FileTimeToDateTime(FileTime : tFileTime) : tDateTime;
  var    SystemTime    : tSystemTime;
         LocalFileTime : tFileTime;
  begin
    FileTimeToLocalFileTime(FileTime, LocalFileTime);
    FileTimeToSystemTime(LocalFileTime, SystemTime);
    FileTimeToDateTime := EncodeDate(SystemTime.wYear,
                                     SystemTime.wMonth,
                                     SystemTime.wDay)
                          + EncodeTime(SystemTime.wHour,
                                       SystemTime.wMinute,
                                       SystemTime.wSecond,
                                       SystemTime.wMilliseconds);
  end;

procedure DateTimeToFileTime(DateTime : tDateTime; var FileTime : tFileTime);
  var     SystemTime    : tSystemTime;
          LocalFileTime : tFileTime;
  begin
    SystemTime.wYear            := 0;
    LocalFileTime.dwLowDateTime := 0;
    FillChar(SystemTime, sizeof(SystemTime), 0);
    DecodeDate(DateTime, SystemTime.wYear, SystemTime.wMonth,
                         SystemTime.wDay);
    DecodeTime(DateTime, SystemTime.wHour, SystemTime.wMinute,
                         SystemTime.wSecond,
                         SystemTime.wMilliSeconds);
    SystemTimeToFileTime(SystemTime, LocalFileTime);
    LocalFileTimeToFileTime(LocalFileTime, FileTime);
  end;

function SetFileTimeOnName(DateiName : String;
                           CreationTime, LastAccessTime,
                           LastWriteTime : tFileTime) : Boolean;
  var    h : Integer;
  begin
    h := FileOpen(Dateiname, fmOpenReadWrite);
    if h < 0 then
      SetFileTimeOnName := False
    else
      begin
        SetFileTimeOnName :=
          SetFileTime(h, @CreationTime, @LastAccessTime, @LastWriteTime);
        FileClose(h);
      end;
  end;

function MkLink(Path, Description, WorkDir : String; Link : WideString) : String;
  var    ShellLink   : iShellLink;
         PersistFile : iPersistFile;
  begin
    MkLink := '';
    Coinitialize(nil);
    try
      OLECheck(CoCreateInstance(CLSID_ShellLink, Nil, CLSCTX_INPROC_SERVER,
                                IShellLink, ShellLink));
      OLECheck(ShellLink.SetPath(PChar(Path)));
      OLECheck(ShellLink.SetDescription(PChar(Description)));
      OLECheck(ShellLink.SetWorkingDirectory(PChar(WorkDir)));
      PersistFile := ShellLink as IPersistFile;
      OLECheck(PersistFile.Save(PWideChar(Link), TRUE));
      PersistFile := Nil;
      ShellLink   := Nil;
    except
      MkLink := SysErrorMessage(GetLastError);
    end;
  end;

function Exec(Wnd : hWnd; Shell, Params : String; nShow : Integer) : Boolean;
  var    Info : tShellExecuteInfo;
  begin
    Info.cbSize         := sizeof(Info);
    Info.fMask          := See_Mask_ConnectNetDrv or See_Mask_DoEnvSubst
                           or See_Mask_NoCloseProcess;
    Info.Wnd            := Wnd;
    Info.lpVerb         := 'Open';
    Info.lpFile         := PChar(Shell);
    if Params = '' then
      Info.lpParameters := Nil
    else
      Info.lpParameters := PChar(Params);
    Info.lpDirectory    := Nil;
    Info.nShow          := nShow;
    Info.hInstApp       := 0;
    Info.lpIDList       := Nil;
    Info.lpClass        := Nil;
    Info.hkeyClass      := 0;
    Info.dwHotKey       := 0;
    Info.hIcon          := 0;
    Info.hProcess       := 0;
    If not ShellExecuteEx(@Info) then
      exec := False
    else
      exec := Info.hInstApp > 32;
  end;

function ExitWindowsWithMessage(Flags : uInt; var Initiated : Boolean) : String;
  var    NewState,
         PreviousState,
         TempState      : tTokenPrivileges;
         TokenHandle    : tHandle;
         PreviousSize,
         TempSize       : DWord;
         Error          : String;
  begin
    Error     := '';
    Initiated := False;
    if Win32Platform = VER_PLATFORM_WIN32_NT then
      begin
        if not OpenProcessToken(GetCurrentProcess,
                                Token_Adjust_Privileges or Token_Query,
                                TokenHandle) then
          Error := 'OpenProcessToken: ' + SysErrorMessage(GetLastError)
        else
          if not LookupPrivilegeValue(Nil, SE_Shutdown_Name,
                                      NewState.Privileges[0].Luid) then
            Error := 'LookupPrivilegeValue: ' + SysErrorMessage(GetLastError)
          else
            begin
              NewState.PrivilegeCount := 1;
              NewState.Privileges[0].Attributes := SE_Privilege_enabled;
              if not AdjustTokenPrivileges(
                       TokenHandle, False, NewState, sizeof(PreviousState),
                       PreviousState, PreviousSize) then
                Error := 'AdjustTokenPrivileges(enable): '
                         + SysErrorMessage(GetLastError);
            end;
      end;
    if Error = '' then
      begin
        Initiated := True;
        if not ExitWindowsEx(Flags, 0) Then
          Error := 'failed: ' + SysErrorMessage(GetLastError);
        if Win32Platform = VER_PLATFORM_WIN32_NT then
          begin
            if not AdjustTokenPrivileges(
                     TokenHandle, False, PreviousState, sizeof(TempState),
                     TempState, TempSize) then
              if Error = '' then
                Error := 'AdjustTokenPrivileges(disable): '
                         + SysErrorMessage(GetLastError);
          end;
      end;
    ExitWindowsWithMessage := Error;
  end;
{$endif}

{$ifdef Win32}
function SaveFormToReg(Form : tForm) : Boolean;
  var    Reg : tRegistry;
  begin
    SaveFormToReg := True;
    try
      reg := tRegistry.Create;
    except
      Reg := Nil;
    end;
    if Reg = Nil then
      SaveFormToReg := False
    else
      try
        if RegOpen(Reg, Form.Name) then
          begin
            if Form.BorderStyle = bsSizeable then
              Reg.WriteInteger('WindowState', ord(Form.WindowState));
            if Form.WindowState = wsNormal then
              begin
                Reg.WriteInteger('Left', Form.Left);
                Reg.WriteInteger('Top',  Form.Top);
                if Form.BorderStyle = bsSizeable then
                  begin
                    Reg.WriteInteger('Width',  Form.Width);
                    Reg.WriteInteger('Height', Form.Height);
                  end;
              end;
            Reg.CloseKey;
          end;
      finally
        Reg.Free;
      end;
  end;

function LoadFormFromReg(Form : tForm) : Boolean;
  var    reg         : tRegistry;
         Left,
         Top,
         Width,
         Height      : Integer;
         WindowState : tWindowState;
  begin
    LoadFormFromReg := True;
    try
      reg := tRegistry.Create;
    except
      Reg := Nil;
    end;
    if reg = Nil then
      LoadFormFromReg := False
    else
      try
        if RegOpen(Reg, Form.Name) then
          begin
            Left        := (Screen.Width  - Form.Width)  div 2;
            Top         := (Screen.Height - Form.Height) div 2;
            Width       := Form.Width;
            Height      := Form.Height;
            WindowState := Form.WindowState;
            try
              Left := reg.ReadInteger('Left');
            except
            end;
            try
              Top  := reg.ReadInteger('Top');
            except
            end;
            if Form.BorderStyle = bsSizeable then
              begin
                try
                  Width  := reg.ReadInteger('Width');
                except
                end;
                try
                  Height := reg.ReadInteger('Height');
                except
                end;
                try
                  WindowState := tWindowState(
                    Reg.ReadInteger('WindowState'));
                except
                end;
              end;
            reg.closekey;
            MoveFormTo(Form, Left, Top, Width, Height, WindowState);
          end;
      finally
        reg.Free;
      end;
  end;
{$endif}

function WSAErrorString(Error : Integer) : String;
  begin
    case Error of
      wsaEIntr :
        wsaErrorString := 'wsaEIntr - Interrupted function call.';
      wsaEBadF :
        WSAErrorString := 'wsaEBadF';
      wsaEAcces :
        WSAErrorString := 'wsaEAcces - Permission denied.';
      wsaEFault :
        WSAErrorString := 'wsaEFault - Bad address.';
      wsaEInval :
        WSAErrorString := 'wsaEInval - Invalid argument.';
      wsaEMFile :
        WSAErrorString := 'wsaEMFile - Too many open files.';
      wsaEWouldBlock :
        WSAErrorString := 'wsaEWouldBlock - Resource temporarily unavailable.';
      wsaEInProgress :
        WSAErrorString := 'wsaEInProgress - Operation now in progress.';
      wsaEAlready :
        WSAErrorString := 'wsaEAlready - Operation already in progress.';
      wsaENotSock :
        WSAErrorString := 'wsaENotSock - Socket operation on non-socket.';
      wsaEDestAddrReq :
        WSAErrorString := 'wsaEDestAddrReq - Destination address required.';
      wsaEMsgSize :
        WSAErrorString := 'wsaEMsgSize - Message too long.';
      wsaEPrototype :
        WSAErrorString := 'wsaEPrototype - Protocol wrong type for socket.';
      wsaENoProtoOpt :
        WSAErrorString := 'wsaENoProtoOpt - Bad protocol option.';
      wsaEProtoNoSupport :
        WSAErrorString := 'wsaEProtoNoSupport - Protocol not supported.';
      wsaESockTNoSupport :
        WSAErrorString := 'wsaESockTNoSupport - Socket type not supported.';
      wsaEOpNotSupp :
        WSAErrorString := 'wsaEOpNotSupp - Operation not supported.';
      wsaEPFNoSupport :
        WSAErrorString := 'wsaEPFNoSupport - Protocol family not supported.';
      wsaEAFNoSupport :
        WSAErrorString := 'wsaEAFNoSupport - Address family not supported '
                          + 'by protocol family.';
      wsaEAddrInUse :
        WSAErrorString := 'wsaEAddrInUse - Address already in use.';
      wsaEAddrNotAvail :
        WSAErrorString := 'wsaEAddrNotAvail - Cannot assign requested address.';
      wsaENetDown :
        WSAErrorString := 'wsaENetDown - Network is down.';
      wsaENetUnreach :
        WSAErrorString := 'wsaENetUnreach - Network is unreachable.';
      wsaENetReset :
        WSAErrorString := 'wsaENetReset - Network dropped connection on reset.';
      wsaEConnAborted :
        WSAErrorString := 'wsaEConnAborted - Software caused connection abort.';
      wsaEConnReset :
        WSAErrorString := 'wsaEConnReset - Connection reset by peer.';
      wsaENoBufs :
        WSAErrorString := 'wsaENoBufs - No buffer space available.';
      wsaEIsConn :
        WSAErrorString := 'wsaEIsConn - Socket is already connected.';
      wsaENotConn :
        WSAErrorString := 'wsaENotConn - Socket is not connected.';
      wsaEShutdown :
        WSAErrorString := 'wsaEShutdown - Cannot send after socket shutdown.';
      wsaETooManyRefs :
        WSAErrorString := 'wsaETooManyRefs';
      wsaETimedOut :
        WSAErrorString := 'wsaETimedOut - Connection timed out.';
      wsaEConnRefused :
        WSAErrorString := 'wsaEConnRefused - Connection refused.';
      wsaELoop :
        WSAErrorString := 'wsaELoop';
      wsaENameTooLong :
        WSAErrorString := 'wsaENameTooLong';
      wsaEHostDown :
        WSAErrorString := 'wsaEHostDown';
      wsaEHostUnreach :
        WSAErrorString := 'wsaEHostUnreach - No route to host.';
      wsaENotEmpty :
        WSAErrorString := 'wsaENotEmpty';
      {$ifndef FPC}
      wsaEProcLim :
        WSAErrorString := 'wsaEProcLim - Too many processes.';
      {$endif}
      wsaEUsers :
        WSAErrorString := 'wsaEUsers';
      wsaEDQuot :
        WSAErrorString := 'wsaEDQuot';
      wsaEStale :
        WSAErrorString := 'wsaEStale';
      wsaERemote :
        WSAErrorString := 'wsaERemote';
      {$ifndef FPC}
      wsaSysNotReady :
        WSAErrorString := 'wsaSysNotReady - Network subsystem is unavailable.';
      wsaVerNotSupported :
        WSAErrorString := 'wsaVerNotSupported - WinSock.DLL version '
                          + 'out of range.';
      wsaNotInitialised :
        WSAErrorString := 'wsaNotInitialised - Successful wsaStartup '
                          + 'not yet performed.';
      wsaEDisCon :
        WSAErrorString := 'wsaEDisCon - Graceful shutdown in progress.';
      {$endif}
      wsaHost_not_found :
        WSAErrorString := 'wsaHost_not_found';
      wsaTry_again :
        WSAErrorString := 'wsaTry_again - Non-authoritative host not found.';
      wsaNo_recovery :
        WSAErrorString := 'wsaNo_recovery - This is a non-recoverable error.';
      {$ifndef FPC}
      wsaNo_data :
        WSAErrorString := 'wsaNo_data - Valid name, no data record '
                          + 'of requested type.';
      {$endif}
    else
      WSAErrorString := 'Error ' + IntToStr(Error) + ' in WinSock call.';
    end;
  end;

function SavePortWithPath(FileName, PortName : String;
                          PortNum            : Word;
                          Protocol, Comment  : String) : tSetServEnt;
  const  Neu              = '.new';
  var    LeseFile,
         SchreibFile      : TextFile;
         ok               : tSetServEnt;
         s, t,
         UpperPortName,
         UpperProtocol,
         CurrentPortNames,
         CurrentProtocol  : String;
         i, j             : Integer;
  begin
    ok := SetServEntOk;
    AssignFile(LeseFile, FileName);
    {$ifdef fpc}
    {$I-}
    reset(LeseFile);
    if ioresult <> 0 then
      ok := SetServEntNotFound;
    {$I+}
    {$else}
    try
      reset(LeseFile);
    except
      ok := SetServEntNotFound;
    end;
    {$endif}
    if Ok = SetServEntOk then
      begin
        AssignFile(SchreibFile, Filename + neu);
        {$ifdef fpc}
        {$I-}
        rewrite(SchreibFile);
        if ioresult <> 0 then
          ok := SetServEntErrCreateTemp;
        {$I+}
        {$else}
        try
          rewrite(SchreibFile);
        except
          ok := SetServEntErrCreateTemp;
        end;
        {$endif}
        if Ok <> SetServEntOk then
          CloseFile(LeseFile)
        else
          begin
            UpperPortName := uppercase(PortName);
            UpperProtocol := uppercase(Protocol);
            While not eof(LeseFile) do
              begin
                ReadLn(LeseFile, s);
                CurrentPortNames := '';
                CurrentProtocol  := '';
                i := pos('#', s);
                if i > 0 then
                  t := trim(uppercase(Copy(s, 1, i - 1)))
                else
                  t := trim(uppercase(s));
                while copy(t, 1, 1) = A_HT do
                  t := Copy(t, 2, length(t));
                i := Pos(' ',  t);
                j := Pos(A_HT, t);
                if i = 0 then
                  i := j
                else
                  if (j > 0) and (j < i) then
                    i := j;
                if i > 0 then
                  begin
                    CurrentPortNames := Trim(Copy(t, 1, i - 1));
                    t                := Copy(t, i + 1, length(t));
                    i                := Pos('/', t);
                    if i = 0 then
                      CurrentPortNames := ''
                    else
                      begin
                        CurrentProtocol := Trim(Copy(t, i + 1, length(t)));
                        i := Pos(' ',  CurrentProtocol);
                        j := Pos(A_HT, CurrentProtocol);
                        if i = 0 then
                          i := j
                        else
                          if (j > 0) and (j < i) then
                            i := j;
                        if i > 0 then
                          begin
                            CurrentPortNames :=
                              CurrentPortNames + ' ' +
                              Copy(CurrentProtocol, i + 1, length(CurrentProtocol));
                            CurrentProtocol := Copy(CurrentProtocol, 1, i - 1);
                          end;
                      end;
                  end;
                if (Pos(UpperPortName, CurrentPortNames) = 0)
                or (CurrentProtocol <> UpperProtocol)
                then
                  WriteLn(SchreibFile, s);
              end;
            if PortNum <> 0 then
              begin
                s := IntToStr(PortNum);
                Write(SchreibFile, PortName, A_HT,
                      StringOfChar(' ', 5 - length(s)), s, '/', Protocol);
                if Comment > '' then
                  Write(SchreibFile, A_HT, Comment);
                writeln(SchreibFile);
              end;
            CloseFile(SchreibFile);
            CloseFile(LeseFile);
            if not DeleteFile(Filename) then
              ok := SetServEntErrDelPrev
            else
              if not RenameFile(Filename + neu, Filename) then
                ok := SetServEntErrRename
          end;
      end;
    SavePortWithPath := ok;
  end;

function SavePortWithoutPath(FileName, PortName : String;
                             PortNum            : Word;
                             Protocol, Comment  : String) : tSetServEnt;
  var    ok          : tSetServEnt;
         {$ifndef Unix}
           Retval    : tSetServEnt;
           i         : Integer;
           Directory : Array[0..Max_Path] of char;
         {$endif}
  begin
    {$ifdef Unix}
      ok := SavePortWithPath(MakePath(PathDelim, FileName),
                             PortName, PortNum, Protocol, Comment);
    {$else}
      ok := SetServEntInvalidPathLen;
      i  := GetWindowsDirectory(Directory, sizeof(Directory));
      if (0 < i) and (i < sizeof(Directory)) then
        ok := SavePortWithPath(MakePath(StrPas(Directory), FileName),
                               PortName, PortNum, Protocol, Comment);
      i := GetSystemDirectory(Directory, sizeof(Directory));
      if (0 < i) and (i < sizeof(Directory)) then
        begin
          RetVal := SavePortWithPath(MakePath(StrPas(Directory), FileName),
                                     PortName, PortNum, Protocol, Comment);
          if ok <> SetServEntOk then
            ok := RetVal;
        end;
    {$endif}
    SavePortWithoutPath := ok;
  end;

function SetServEnt(PortName, Protocol : String;
                    PortNum            : Word;
                    Comment            : String) : tSetServEnt;
  const  Services = 'services';
  var    ok       : tSetServEnt;
         {$ifndef Unix}
           RetVal : tSetServEnt;
         {$endif}
  begin
    {$ifdef Unix}
      ok := SavePortWithoutPath(MakePath('etc', Services),
                                PortName, PortNum, Protocol, Comment);
    {$else}
      ok     := SavePortWithoutPath(Services, PortName, PortNum, Protocol,
                                    Comment);
      RetVal := SavePortWithoutPath(MakePath('drivers\etc', Services),
                                    PortName, PortNum, Protocol, Comment);
      if ok <> SetServEntOk then
        ok := RetVal;
      RetVal := SavePortWithoutPath(MakePath('system32\drivers\etc', Services),
                                    PortName, PortNum, Protocol, Comment);
      if ok <> SetServEntOk then
        ok := RetVal;
    {$endif}
    SetServEnt := ok;
  end;

function InetUpdate(var Archiv       : String;
                    Jahr, Monat, Tag : Integer;
                    var NewAvailable : Boolean) : Boolean;
  const  Domain        = 'www.Wolf-Fuerth.de';
         Datei         = 'downl1de.htm';
         sHttp         = 'http';
         hrefBegin     = '<a';
         hrefLink      = 'href="';
         hrefEnd       = '</a>';
         Spalte        = '<td';
         Pfad          = sHttp + '://' + Domain + '/';
  type   tFortschritt  = (NichtsGefunden, EtwasGefunden,
                          hrefGefunden,   NameGefunden,   DatumGefunden);
  var    Fortschritt   : tFortschritt;
         i, j          : Integer;
         ok            : Boolean;
         s,
         Stream,
         HTML,
         Datum,
         Link,
         Txt           : String;
         {$ifdef Unix}
           HTTP        : tIdHTTP;
           Proxy,
           User,
           Password    : String;
           Port        : Integer;
         {$else}
           Buffer      : Array[0 .. $FF] of char;
           Inet,
           Url         : hInternet;
           d           : DWord;
         {$endif}
  begin
    NewAvailable   := False;
    Datum          := '';
    Link           := '';
    Archiv         := AnsiUpperCase(Archiv);
    Fortschritt    := NichtsGefunden;
    Txt            := '';
    Stream         := '';
    {$ifdef Unix}
      Proxy        := '';
      User         := '';
      Password     := '';
      Port         := 0;
      try
        HTTP       := tIdHttp.Create(Nil);
      except
        HTTP       := Nil;
      end;
      if HTTP <> Nil then
        try
          if GetProxyParams(Proxy, User, Password, Port) then
            if (Proxy > '') and (Port <> 0) then
              begin
                HTTP.ProxyParams.ProxyServer := Proxy;
                HTTP.ProxyParams.ProxyPort   := Port;
                if (User > '') and (Password > '') then
                  begin
                    HTTP.ProxyParams.ProxyUsername := User;
                    HTTP.ProxyParams.ProxyPassword := Password;
                  end;
              end;
          Stream := HTTP.Get(Pfad + Datei);
        finally
          HTTP.free;
        end;
    {$else}
      Inet := InternetOpen('GetHtml', Internet_Open_Type_PreConfig, '', '', 0);
      if Inet <> Nil then
        begin
          d := sizeof(Buffer) - 1;
          if InternetCanonicalizeUrl(Pfad + Datei, Buffer, d, 0) then
            begin
              Buffer[d] := #0;
              Url       := InternetOpenUrl(Inet, Buffer, Nil, DWord(-1), 0, 0);
              if Url <> Nil then
                begin
                  d := 1;
                  while d > 0 do
                    if InternetReadFile(Url, @Buffer, sizeof(Buffer) - 1, d) then
                      begin
                        Buffer[d] := #0;
                        Stream    := Stream + Buffer;
                      end
                    else
                      d := 0;
                  InternetCloseHandle(Url);
                end;
            end;
          InternetCloseHandle(Inet);
        end;
    {$endif}
    Stream := StringReplace(Stream, '&Auml;',  A_ISO_UPPER_AE, [rfReplaceAll]);
    Stream := StringReplace(Stream, '&Ouml;',  A_ISO_UPPER_OE, [rfReplaceAll]);
    Stream := StringReplace(Stream, '&Uuml;',  A_ISO_UPPER_UE, [rfReplaceAll]);
    Stream := StringReplace(Stream, '&auml;',  A_ISO_LOWER_AE, [rfReplaceAll]);
    Stream := StringReplace(Stream, '&ouml;',  A_ISO_LOWER_OE, [rfReplaceAll]);
    Stream := StringReplace(Stream, '&uuml;',  A_ISO_LOWER_UE, [rfReplaceAll]);
    Stream := StringReplace(Stream, '&szlig;', A_ISO_SS,       [rfReplaceAll]);
    Stream := StringReplace(Stream, #$0D,      ' ',            [rfReplaceAll]);
    Stream := StringReplace(Stream, #$0A,      ' ',            [rfReplaceAll]);
    ok := Stream > '';
    i := Pos('>', Stream);
    while (i > 0) and ok do
      begin
        j := Pos('<', Stream);
        if (0 < j) and (j < i) then
          begin
            Txt  := Txt + Copy(Stream, 1, j - 1);
            HTML := Copy(Stream, j, i - j + 1);
            case Fortschritt of
              NichtsGefunden:
                if Copy(HTML, 1, length(hrefBegin)) = hrefBegin then
                  begin
                    j := Pos(hrefLink, HTML);
                    if j > 0 then
                      begin
                        Link := Copy(HTML, j + Length(hrefLink), Length(HTML));
                        j := Pos('"', Link);
                        if j > 0 then
                          Link := Copy(Link, 1, j - 1);
                        if Pos('/', Link) = 0 then
                          Link  := Pfad + Link;
                        Fortschritt := EtwasGefunden;
                      end;
                    Txt := '';
                  end;
              EtwasGefunden:
                begin
                  if Copy(HTML, 1, length(hrefEnd)) = hrefEnd then
                    if Trim(AnsiUpperCase(Txt)) = Archiv then
                      Fortschritt := NameGefunden;
                  if Fortschritt <> NameGefunden then
                    Fortschritt := NichtsGefunden;
                end;
              NameGefunden:
                if Copy(HTML, 1, length(Spalte)) = Spalte then
                  begin
                    Fortschritt := DatumGefunden;
                    Txt         := '';
                  end;
              DatumGefunden:
                if Copy(HTML, 1, length(Spalte)) = Spalte then
                  begin
                    Datum := Txt;
                    Ok    := False;
                  end;
            end;
          end;
        Stream := Copy(Stream, i + 1, length(Stream));
        i      := Pos('>', Stream);
      end;
    ok := True;
    i  := Pos('.', Datum);
    if i = 0 then
      ok := False
    else
      begin
        s     := Copy(Datum, 1, i - 1);
        Datum := Trim(Copy(Datum, i + 1, length(Datum)));
        val(s, i, j);
        if i <> Tag then
          NewAvailable := True;
        i := Pos(' ', Datum);
        if i = 0 then
          ok := False;
      end;
    if ok then
      begin
        s     := Copy(Datum, 1, i - 1);
        Datum := Copy(Datum, i + 1, length(Datum));
        for i := 1 to 12 do
          if s = GermanLongMonthNames[i] then
            if i <> Monat then
              NewAvailable := True;
        val(Datum, i, j);
        if i <> Jahr then
          NewAvailable := True;
      end;
    InetUpdate := Ok;
    Archiv     := Link;
  end;

{$else}

{$ifndef Unix}
{$ifndef Ver100}
function CheckCommErrors(Port : Integer) : String;
  var    Status  : TComStat;
         Errors,
         Bit     : Word;
         s       : String;
  begin
    Errors := GetCommError(Port, Status);
    Bit    := 1;
    s      := '';
    while Bit <> 0 do
      begin
        if (Bit and Errors) <> 0 Then
          case Bit of
            CE_BREAK   : s := s + 'break signal detected' + A_LF;
            CE_CTSTO   : s := s + 'clear-to-send timeout' + A_LF;
            CE_DNS     : s := s + 'parallel port is not selected' + A_LF;
            CE_DSRTO   : s := s + 'data-set-ready timeout' + A_LF;
            CE_FRAME   : s := s + 'framing error' + A_LF;
            CE_IOE     : s := s + 'I/O error on parallel port' + A_LF;
            CE_MODE    : s := s + 'port setting is not supported' + A_LF;
            CE_OOP     : s := s + 'parallel port reports paper out' + A_LF;
            CE_OVERRUN : s := s + 'character overrun' + A_LF;
            CE_PTO     : s := s + 'timeout on parallel port' + A_LF;
            CE_RLSDTO  : s := s + 'receive-line-signal-detect timeout' + A_LF;
            CE_RXOVER  : s := s + 'receive buffer overflow' + A_LF;
            CE_RXPARITY: s := s + 'parity error' + A_LF;
            CE_TXFULL  : s := s + 'send buffer overflow' + A_LF;
          else
            s := s + 'error $' + IntToHex(Bit, 4) + A_LF;
          end;
        Bit := Bit shl 1;
      end;
    CheckCommErrors := s;
  end;

function GetHandleFromTaskName(TaskName : String) : tHandle;
  var    Wnd      : HWnd;
         TaskInfo : tTaskEntry;
  begin
    GetHandleFromTaskName := 0;
    if TaskName <> '' Then
      begin
        Wnd := GetDesktopWindow;
        Wnd := GetWindow(Wnd, gw_Child);
        while Wnd <> 0 do
          begin
            TaskInfo.dwSize := sizeof(TaskInfo);
            if TaskFindHandle(@TaskInfo, GetWindowTask(Wnd)) then
              if StrPas(TaskInfo.szModule) = TaskName then
                begin
                  GetHandleFromTaskName := Wnd;
                  Wnd := 0;
                end
              else
                Wnd := GetWindow(Wnd, gw_HWndNext)
            else
              Wnd := GetWindow(Wnd, gw_HWndNext);
          end;
      end;
  end;
{$endif}

function OpenCommError(Error : Integer) : String;
  begin
    case Error of
      IE_BADID   : OpenCommError := 'port is not supported';
      IE_BAUDRATE: OpenCommError := 'baud rate is not supported';
      IE_BYTESIZE: OpenCommError := 'byte size invalid';
      IE_DEFAULT : OpenCommError := 'error in standard parameters';
      IE_HARDWARE: OpenCommError := 'port is locked';
      IE_MEMORY  : OpenCommError := 'cannot assign buffers';
      IE_NOPEN   : OpenCommError := 'port is not open';
      IE_OPEN    : OpenCommError := 'port is in use';
    else
      OpenCommError := 'error ' + IntToStr(Error);
    end;
  end;
{$endif}
{$endif}

{$ifdef Win32}
  function RegOpen(var Reg : tRegistry; RelativePath : String) : Boolean;
    begin
      RegOpen := reg.OpenKey(
        MakePath(MakePath('Software\Heinrich Wolf', GetMyName), RelativePath),
        True);
    end;
{$endif}

{$ifdef Ver80}
  function GetLastError : Integer;
    begin
      GetLastError := -1;
    end;
{$endif}

{$ifdef Unix}
  function GetLastError : Integer;
    begin
      GetLastError := fpGetErrno;
    end;
{$endif}

function FindQuote(p : pChar) : pChar;
  var    i : Integer;
  begin
    FindQuote := Nil;
    i         := 0;
    while i >= 0 do
      if p[i] = #0 then
        i := -1
      else
        if p[i] = '\' then
          begin
            inc(i);
            if p[i] = #0 then
              i := -1
            else
              inc(i);
          end
        else
          if p[i] = '"' then
            begin
              FindQuote := p + i;
              i         := -1;
            end
          else
            inc(i);
  end;

procedure ParseCmdLine(var Index  : Integer;
                       var CmdStr : String);
  { http://blogs.msdn.com/b/oldnewthing/archive/2010/09/17/10063629.aspx }
  var     Blank,
          QuoteDelim  : Boolean;
          QuoteCount,
          CmdCount,
          Escape,
          i           : Integer;
  begin
    {$ifdef Win32}
      CmdCount := -1;
    {$else}
      CmdCount := 0;
    {$endif}
    CmdStr     := '';
    Blank      := True;
    QuoteDelim := False;
    QuoteCount := 0;
    Escape     := 0;
    if CmdLine <> Nil then
      begin
        i := 0;
        while (CmdLine[i] <> #0) and ((0 = Index) or (CmdCount < Index)) do
          begin
            case CmdLine[i] of
              ' ' : begin
                      CmdStr := CmdStr + StringOfChar('\', Escape)
                                       + StringOfChar('"', QuoteCount div 3);
                      Escape := 0;
                      QuoteCount := 0;
                      if QuoteDelim then
                        CmdStr := CmdStr + ' '
                      else
                        begin
                          if not Blank then
                            inc(CmdCount);
                          if CmdCount < Index then
                            CmdStr := '';
                          Blank := True;
                        end;
                    end;
              '\' : begin
                      Inc(Escape);
                      Blank  := False;
                      CmdStr := CmdStr + StringOfChar('"', QuoteCount div 3);
                      QuoteCount := 0;
                    end;
              '"' : begin
                      Blank  := False;
                      CmdStr := CmdStr + StringOfChar('\', Escape div 2);
                      if Odd(Escape) then
                        CmdStr := CmdStr + '"'
                      else
                        begin
                          if QuoteCount = 0 then
                            QuoteDelim := not QuoteDelim
                          else
                            QuoteDelim := False;
                          inc(QuoteCount);
                        end;
                      Escape := 0;
                    end;
            else
              Blank      := False;
              CmdStr     := CmdStr + StringOfChar('\', Escape)
                                   + StringOfChar('"', QuoteCount div 3)
                                   + CmdLine[i];
              Escape     := 0;
              QuoteCount := 0;
            end;
            inc(i);
          end;
      end;
    if Index = 0 then
      if Blank then
        Index := CmdCount
      else
        Index := CmdCount + 1;
  end;

function CmdCount : Integer;
  var    c        : Integer;
         {$ifndef Unix}
           OldWin : Boolean;
           s      : String;
         {$endif}
         {$ifdef Win32}
           ArgvW  : PPWideChar;
         {$endif}
  begin
    {$ifdef Unix}
      c      := ParamCount;
    {$else}
      c      := 0;
      OldWin := True;
      {$ifdef Win32}
        {$ifdef fpc}
          ArgvW := CommandLineToArgvW(GetCommandLineW, @c);
        {$else}
          ArgvW := CommandLineToArgvW(GetCommandLineW, c);
        {$endif}
        if Assigned(ArgvW) then
          LocalFree(HGlobal(ArgvW));
        if c > 0 then
          begin
            dec(c);
            OldWin := False;
          end;
      {$endif}
      if OldWin then
        begin
          s := '';
          c := 0;
          ParseCmdLine(c, s);
        end;
    {$endif}
    CmdCount := c;
  end;

function CmdStr(Index : Integer) : String;
  {$ifdef Win32}
    type a        = Array[0 .. 0] of PWideChar;
         p        = ^a;
  {$endif}
  var    s        : String;
         {$ifndef Unix}
           OldWin : Boolean;
           c      : Integer;
         {$endif}
         {$ifdef Win32}
           ArgvW  : PPWideChar;
         {$endif}
  begin
    {$ifdef Unix}
      s := ParamStr(Index);
    {$else}
      If Index = 0 then
        s := ParamStr(0)
      else
        if Index > 0 then
          begin
            s      := '';
            c      := 0;
            OldWin := True;
            {$ifdef Win32}
              {$ifdef fpc}
                ArgvW := CommandLineToArgvW(GetCommandLineW, @c);
              {$else}
                ArgvW := CommandLineToArgvW(GetCommandLineW, c);
              {$endif}
              if Assigned(ArgvW) then
                begin
                  if c > 0 then
                    begin
                      OldWin := False;
                      if Index < c then
                        {$R-}
                          s := WideCharToString(p(ArgvW)^[Index]);
                        {$R+}
                    end;
                  LocalFree(HGlobal(ArgvW));
                end;
            {$endif}
            if OldWin then
              ParseCmdLine(Index, s);
          end;
    {$endif}
    CmdStr := s;
  end;

{$ifndef Unix}
function GetHandleFromCaption(Caption : String) : HWnd;
  var    WndName : Array[0..$ff] of char;
  begin
    StrPCopy(WndName, Caption);
    GetHandleFromCaption := FindWindow(Nil, WndName);
  end;

procedure GetTasks(List : tStrings);
  var     s        : String;
          Wnd      : HWnd;
          WndName  : Array[0..$ff] of char;
          {$ifndef Win32}
          TaskInfo : tTaskEntry;
          {$endif}
  begin
    List.Clear;
    Wnd := GetDesktopWindow;
    Wnd := GetWindow(Wnd, gw_Child);
    while Wnd <> 0 do
      begin
        GetWindowText(Wnd, WndName, SizeOf(WndName) - 1);
        s := StrPas(WndName);
        if s > '' then
          if List.IndexOf(s) < 0 then
            List.Add(s);
        {$ifndef Win32}
        TaskInfo.dwSize := sizeof(TaskInfo);
        if TaskFindHandle(@TaskInfo, GetWindowTask(Wnd)) then
          begin
            s := StrPas(TaskInfo.szModule);
            if List.IndexOf(s) < 0 then
              List.Add(s);
          end;
        {$endif}
        Wnd := GetWindow(Wnd, gw_HWndNext);
      end;
  end;

function GetHandleFromStr(s : String) : hWnd;
  var    w : hWnd;
  begin
    w := GetHandleFromCaption(s);
    {$ifndef Win32}
    if w = 0 then
      w := GetHandleFromTaskName(s);
    {$endif}
    GetHandleFromStr := w;
  end;

function DelBeforeCopy(fName : String; FFcontinue : tFFcontinue) : sFileOrFolder;
  var    i         : Integer;
         Status    : sFileOrFolder;
         SearchRec : tSearchRec;
  begin
    Status := [];
    if FindFirst(MakeFind(fName), faAnyFile, SearchRec) <> 0 then
      begin
        if not FFcontinue(fName, 0, SysErrorMessage(GetLastError)
                                    + ' in getting attributes of ' + fName)
        then
          Include(Status, ffCanceled);
      end
    else
      begin
        if FileSetAttr(fName, SearchRec.attr and not (   faHidden
                                                      or faReadOnly
                                                      or faSysFile)) <> 0
        then
          begin
            if (SearchRec.attr and faDirectory) <> 0 then
              Include(Status, ffFolder)
            else
              Include(Status, ffFile);
            if not FFcontinue(
                     fName, 0, SysErrorMessage(GetLastError)
                               + ' in deleting read-only attributes from '
                               + fName)
            then
              Include(Status, ffCanceled);
        end;
      end;
    FindClose(SearchRec);
    if Status = [] then
      if (SearchRec.attr and faDirectory) = 0 then
        begin
          if not DeleteFile(fName) then
            begin
              Include(Status, ffFile);
              if not FFcontinue(fName, 0, SysErrorMessage(GetLastError)
                                          + ' in deleting ' + fName)
              then
                Include(Status, ffCanceled);
            end;
        end
      else
        begin
          i := FindFirst(MakePath(fName, Wildcard), faAnyFile, SearchRec);
          while i = 0 do
            begin
              if not FFcontinue(MakePath(fName, SearchRec.Name), 0, '') then
                Include(Status, ffCanceled);
              if ffCanceled in Status then
                i := ERROR_NO_MORE_FILES
              else
                begin
                  if SearchRec.Name <> '.' then
                    if SearchRec.Name <> '..' then
                      Status := Status + DelBeforeCopy(
                                           MakePath(fName, SearchRec.Name),
                                           FFcontinue);
                  i := FindNext(SearchRec);
                end;
            end;
          i := abs(i);
          if (i <> 0) and (i <> ERROR_NO_MORE_FILES) then
            begin
              Include(Status, ffRecursive);
              if not FFcontinue(fName, 0, SysErrorMessage(GetLastError)
                                          + ' in searching ' + fName)
              then
                Include(Status, ffCanceled);
            end;
          FindClose(SearchRec);
            try
              RmDir(fName);
              { RemoveDir steht unter Delphi 1 nicht zur Verfuegung }
            except
              Include(Status, ffFolder);
              if not FFcontinue(fName, 0, SysErrorMessage(GetLastError)
                                          + ' in deleting ' + fName)
              then
                Include(Status, ffCanceled);
            end;
        end;
    DelBeforeCopy := Status;
  end;

function CopyFileEx(FromName, ToName : String;
                    FailIfExists     : Boolean;
                    FFcontinue       : tFFcontinue) : sFileOrFolder;
  const  IniExt     = '.ini';
         Section    = 'CopyFileEx';
         FromKey    = 'FromName';
         Size       = 'OrigSize';
         Time       = 'Time';
         TmpKey     = 'TmpName';
  var    FromDatei,
         ToDatei,
         LockDatei  : Integer;
         Info       : tSearchRec;
         SeekPos,
         FromBytes,
         ToBytes    : tZusatzSize;
         Status     : sFileOrFolder;
         TmpName,
         IniName    : String;
         IniFile    : tIniFile;
         {$ifdef Ver80}
           s        : String;
           Drive    : Char;
           Buffer   : Array[0 .. $100] of char;
         {$else}
           Buffer   : Array[0 .. $1FFFF] of char;
         {$endif}
  begin
    Status    := [];
    FromDatei := -1;
    ToDatei   := -1;
    LockDatei := -1;
    TmpName   := '';
    SeekPos   := 0;
    IniName   := '';
    IniFile   := Nil;
    FindFirst(FromName, faAnyFile, Info);
    FindClose(Info);
    if FailIfExists and FileExists(ToName) then
      begin
        if not FFcontinue(ToName, 0, ToName + ' exists') then
          Include(Status, ffCanceled);
        Include(Status, ffExists);
      end;
    if Status = [] then
      begin
        FromDatei := FileOpen(FromName, fmOpenRead or fmShareDenyWrite);
        if FromDatei < 0 then
          begin
            if not FFcontinue(FromName, 0, SysErrorMessage(GetLastError)
                                           + ' in opening ' + FromName)
            then
              Include(Status, ffCanceled);
            Include(Status, ffNotFound);
          end;
      end;
    if Status = [] then
      begin
        LockDatei := FileOpen(ToName, fmOpenReadWrite or fmShareExclusive);
        if LockDatei < 0 then
          if FileExists(ToName) then
            begin
              if not FFcontinue(ToName, 0, SysErrorMessage(GetLastError)
                                           + ' in replacing ' + ToName)
              then
                Include(Status, ffCanceled);
              Include(Status, ffFile);
            end;
      end;
    if Status = [] then
      begin
        {$ifdef Ver80}
          s := ExtractFileExt(ToName);
          if s <> IniExt then
            IniName := Copy(ToName, 1, length(ToName) - length(s)) + IniExt;
        {$else}
          IniName := ToName + IniExt;
        {$endif}
        if IniName > '' then
          if FileExists(IniName) then
            try
              IniFile := tIniFile.Create(IniName);
              if IniFile.ReadString(Section, FromKey, '') = FromName then
                begin
                  {$ifdef Ver80}
                     if IniFile.ReadInteger(Section, Size, 0) =
                       DateiGroesse(Info)
                  {$else}
                     if IniFile.ReadFloat(Section, Size, 0) =
                       DateiGroesse(Info)
                  {$endif}
                  then
                    if IniFile.ReadInteger(Section, Time, 0) = Info.Time then
                      begin
                        TmpName := IniFile.ReadString(Section, TmpKey, '');
                        if not FileGetSize(TmpName, SeekPos) then
                          begin
                            SeekPos := 0;
                            TmpName := '';
                          end;
                      end;
                end;
            except
            end;
        if TmpName = '' then
          begin
            try
              IniFile := tIniFile.Create(IniName);
              IniFile.WriteString(Section, FromKey, FromName);
              IniFile.WriteInteger(Section, Time, Info.Time);
              {$ifdef Ver80}
                 IniFile.WriteInteger(Section, Size, DateiGroesse(Info));
              {$else}
                 IniFile.WriteFloat(Section, Size,   DateiGroesse(Info));
              {$endif}
            except
            end;
            {$ifdef Ver80}
              Drive := #0;
              if copy(ToName, 2, 1) = ':' then
                Drive := chr(ord(ToName[1]) or TF_ForceDrive);
              GetTempFileName(Drive, 'tmp', 0, Buffer);
            {$else}
              GetTempFileName(PChar(ExtractFilePath(ToName)),
                              PChar(ExtractFileName(ToName)), 0,
                              Buffer);
            {$endif}
            TmpName := StrPas(Buffer);
          end;
        ToDatei := FileOpen(TmpName, fmOpenWrite or fmShareExclusive);
        if ToDatei < 0 then
          ToDatei := FileCreate(TmpName)
        else
          if FileSeek(ToDatei, SeekPos, 0) < 0 then
            begin
              if not FFcontinue(ToName, SeekPos, 'cannot seek ' + FromName) then
                Include(Status, ffCanceled);
              Include(Status, ffFile);
            end
          else
            FFcontinue(ToName, SeekPos, '');
        if ToDatei < 0 then
          begin
            if not FFcontinue(ToName, 0, 'cannot create ' + ToName) then
              Include(Status, ffCanceled);
            Include(Status, ffFile);
          end
        else
          if IniFile <> Nil then
            IniFile.WriteString(Section, TmpKey, TmpName);
      end;
    IniFile.Free;
    if Status = [] then
      if FileSeek(FromDatei, SeekPos, 0) < 0 then
        begin
          if not FFcontinue(ToName, SeekPos, 'cannot seek ' + FromName) then
            Include(Status, ffCanceled);
          Include(Status, ffFile);
        end;
    if Status = [] then
      repeat
        FromBytes := FileRead(FromDatei, Buffer, sizeof(Buffer));
        if FromBytes < 0 then
          begin
            FromBytes := 0;
            Include(Status, ffReadError);
            if not FFcontinue(FromName, 0, 'read error on ' + FromName) then
              Include(Status, ffCanceled);
          end;
        if FromBytes > 0 then
          begin
            ToBytes := FileWrite(ToDatei, Buffer, FromBytes);
            Inc(SeekPos, ToBytes);
            if ToBytes <> FromBytes then
              begin
                if not FFcontinue(ToName, ToBytes, 'cannot write to ' + ToName)
                then
                  Include(Status, ffCanceled);
                Include(Status, ffVolumeFull);
              end;
           if not FFcontinue(FromName, ToBytes, '') then
             Include(Status, ffCanceled);
          end;
      until (Status <> []) or (FromBytes < sizeof(Buffer));
    if Status = [] then
      begin
        if IniName > '' then
          DeleteFile(IniName);
        {$ifdef Ver80}
          FileSetDate(ToDatei, Info.Time);
        {$else}
          if not SetFileTime(ToDatei,
                             @Info.FindData.ftCreationTime,
                             @Info.FindData.ftLastAccessTime,
                             @Info.FindData.ftLastWriteTime)
          then
            begin
              Include(Status, ffFile);
              if not FFcontinue(ToName, 0, SysErrorMessage(GetLastError)
                                           + ' in setting time on ' + ToName)
              then
                Include(Status, ffCanceled);
            end;
        {$endif}
      end;
    if FromDatei >= 0 then
      FileClose(FromDatei);
    if ToDatei >= 0 then
      FileClose(ToDatei);
    if LockDatei >= 0 then
      FileClose(LockDatei);
    if Status = [] then
      if FileSetAttr(TmpName, Info.attr) <> 0 then
        begin
          if not FFcontinue(ToName, 0, SysErrorMessage(GetLastError)
                                       + ' in setting attributes on '
                                       + TmpName)
          then
            Include(Status, ffCanceled);
          Include(Status, ffFile);
        end;
    if Status = [] then
      begin
        if FileExists(ToName) then
          if not DeleteFile(ToName) then
            begin
              if not FFcontinue(ToName, 0, SysErrorMessage(GetLastError)
                                           + ' in replacing ' + ToName)
              then
                Include(Status, ffCanceled);
              Include(Status, ffFile);
            end;
        if Status = [] then
          if not RenameFile(TmpName, ToName) then
            begin
              if not FFcontinue(ToName, 0, SysErrorMessage(GetLastError)
                                           + ' in renaming ' + TmpName
                                           + ' to ' + ToName)
              then
                Include(Status, ffCanceled);
              Include(Status, ffFile);
            end;
      end;
    CopyFileEx := Status;
  end;

function CopyFileOrFolder(FromName, ToName : String; Allowed : sFileOrFolder;
                          FFcontinue : tFFcontinue)
           : sFileOrFolder; { returns error reason }
  var    Status,
         CallStatus : sFileOrFolder;
         i          : Integer;
         Dir        : String;
         SearchRec  : tSearchRec;
  begin
    Status := [];
    if (FromName = '') or (ToName = '') then
      begin
        Include(Status, ffParam);
        if (FromName = '') and (ToName = '') then
          begin
            if not FFcontinue('', 0, 'neither source nor destination given')
            then
              Include(Status, ffCanceled);
          end
        else
          if FromName = '' then
            begin
              if not FFcontinue(ToName, 0, 'no source given') then
                Include(Status, ffCanceled);
            end
          else
            if not FFcontinue(FromName, 0, 'no destination given') then
              Include(Status, ffCanceled);
      end
    else
      begin
        if FindFirst(MakeFind(FromName), faAnyFile, SearchRec) <> 0 then
          begin
            Include(Status, ffNotFound);
            if not FFcontinue(FromName, 0, FromName + ' not found') then
              Include(Status, ffCanceled);
          end
        else
          begin
            if (SearchRec.attr and faDirectory) <> 0 then
              Include(Status, ffFolder)
            else
              Include(Status, ffFile);
          end;
        FindClose(SearchRec);
      end;
    if Status * [ffNotFound, ffParam] = [] then
      begin
        Dir := ExtractFilePath(ToName);
        if FindFirst(MakeFind(ToName), faAnyFile, SearchRec) <> 0 then
          begin
            if not DirectoryExists(Dir) then
              Include(Status, ffForceDirs);
          end
        else
          begin
            Include(Status, ffExists);
            if (SearchRec.attr and (   faHidden
                                    or faReadOnly
                                    or faSysFile)) <> 0
            then
              Include(Status, ffReadOnly);
            if (SearchRec.attr and faDirectory) <> 0 then
              begin
                if ffFile in Status then
                  Include(Status, ffDifferent)
              end
            else
              if ffFolder in Status then
                Include(Status, ffDifferent);
          end;
        FindClose(SearchRec);
        if Allowed * Status <> Status then
          begin
            if not FFcontinue(FromName, 0,
                              'errors: ' + FileOrFolderItems(Status))
            then
              Include(Status, ffCanceled);
          end
        else
          begin
            if ffDifferent in Status then
              begin
                CallStatus := DelBeforeCopy(ToName, FFcontinue);
                if CallStatus = [] then
                  Status := Status - [ffDifferent, ffExists, ffReadOnly];
              end;
            if not (ffDifferent in Status) then
              begin
                if Dir <> '' then
                  begin
                    { Procedure in Delphi 4; Funktion nur in Delphi 5 }
                    ForceDirectories(Dir);
                    if DirectoryExists(Dir) then
                      Exclude(Status, ffForceDirs)
                    else
                      if not FFcontinue(Dir, 0, SysErrorMessage(GetLastError)
                                                + ' in making directory ' + Dir)
                      then
                        Include(Status, ffCanceled);
                  end;
                If ffFile in Status then
                  begin
                    CallStatus := CopyFileEx(FromName, ToName,
                                             False, FFContinue);
                    if CallStatus = [] then
                      Status := Status - [ffFile, ffExists, ffReadOnly]
                    else
                      Status := Status + CallStatus;
                  end;
                If ffFolder in Status then
                  begin
                    { Procedure in Delphi 4; Funktion nur in Delphi 5 }
                    ForceDirectories(ToName);
                    if DirectoryExists(ToName) then
                      Status := Status - [ffFolder, ffExists, ffReadOnly]
                    else
                      if not FFcontinue(ToName, 0,
                                        SysErrorMessage(GetLastError)
                                        + ' in making directory ' + ToName)
                      then
                        Include(Status, ffCanceled);
                    if not (ffFolder in Status) then
                      if ffRecursive in Allowed then
                        begin
                          i := FindFirst(MakePath(FromName, Wildcard),
                                         faAnyFile, SearchRec);
                          while i = 0 do
                            if not FFcontinue(MakePath(FromName,
                                                       SearchRec.Name), 0, '')
                            then
                              begin
                                i := ERROR_NO_MORE_FILES;
                                Include(Status, ffCanceled);
                              end
                            else
                              begin
                                if SearchRec.Name <> '.' then
                                  if SearchRec.Name <> '..' then
                                    begin
                                      CallStatus :=
                                        CopyFileOrFolder(
                                          MakePath(FromName, SearchRec.Name),
                                          MakePath(ToName,   SearchRec.Name),
                                          Allowed, FFcontinue);
                                      if CallStatus <> [] then
                                        Status := Status + CallStatus
                                                  + [ffRecursive];
                                    end;
                                i := FindNext(SearchRec);
                              end;
                          i := abs(i);
                          if (i <> 0) and (i <> ERROR_NO_MORE_FILES) then
                            begin
                              if not FFContinue(FromName, 0,
                                                SysErrorMessage(GetLastError)
                                                + ' in searching ' + FromName)
                              then
                                Include(Status, ffCanceled);
                              Include(Status, ffRecursive);
                            end;
                          FindClose(SearchRec);
                        end;
                  end;
              end;
          end;
      end;
    CopyFileOrFolder := Status;
  end;

function DelFileOrFolder(fName : String; Allowed : sFileOrFolder;
                         FFcontinue : tFFcontinue) :
         sFileOrFolder; { returns error reason }
  var    Status,
         CallStatus : sFileOrFolder;
         i          : Integer;
         SearchRec  : tSearchRec;
  begin
    Status := [];
    if fName = '' then
      begin
        Include(Status, ffParam);
        if not FFcontinue('', 0, 'no file or folder name given') then
          Include(Status, ffCanceled);
      end
    else
      begin
        if Copy(fName, Length(fName), 1) = PathDelim then
          if Copy(fName, Length(fName) - 1, 1) <> ':' then
            fName := Copy(fName, 1, Length(fName) - 1);
        if FindFirst(MakeFind(fName), faAnyFile, SearchRec) <> 0 then
          begin
            Include(Status, ffNotFound);
            if not FFcontinue(fName, 0, fName + ' not found') then
              Include(Status, ffCanceled);
          end
        else
          begin
            if (SearchRec.attr and faDirectory) <> 0 then
              Include(Status, ffFolder)
            else
              Include(Status, ffFile);
            if (SearchRec.attr and (   faReadOnly
                                    or faHidden
                                    or faSysFile)) <> 0
            then
              Include(Status, ffReadOnly);
          end;
        FindClose(SearchRec);
        if not (ffNotFound in Status) then
          begin
            if Allowed * Status <> Status then
              begin
                if not FFcontinue(fName, 0,
                                  'errors: ' + FileOrFolderItems(Status))
                then
                  Include(Status, ffCanceled);
              end
            else
              begin
                if FileSetAttr(fName,
                               SearchRec.attr and not (   faHidden
                                                       or faReadOnly
                                                       or faSysFile)) = 0
                then
                  Exclude(Status, ffReadOnly);
                If ffFile in Status then
                  begin
                    if DeleteFile(fName) then
                      begin
                        Exclude(Status, ffFile);
                        if not FFcontinue(fName, 0, '') then
                          Include(Status, ffCanceled);
                      end
                    else
                      if not FFcontinue(fName, 0, SysErrorMessage(GetLastError)
                                                  + ' in deleting ' + fName)
                      then
                        Include(Status, ffCanceled);
                  end;
                If ffFolder in Status then
                  begin
                    i := FindFirst(MakePath(fName, Wildcard),
                                   faAnyFile, SearchRec);
                    while i = 0 do
                      if not FFcontinue(MakePath(fName, SearchRec.Name), 0, '')
                      then
                        begin
                          i := ERROR_NO_MORE_FILES;
                          Include(Status, ffCanceled);
                        end
                      else
                        begin
                          if SearchRec.Name <> '.' then
                            if SearchRec.Name <> '..' then
                              if ffRecursive in Allowed then
                                begin
                                  CallStatus :=
                                    DelFileOrFolder(
                                      MakePath(fName, SearchRec.Name),
                                      Allowed, FFcontinue);
                                  if CallStatus <> [] then
                                    Status := Status + CallStatus
                                                     + [ffRecursive];
                                end
                              else
                                begin
                                  Include(Status, ffRecursive);
                                  if not FFcontinue(
                                           MakePath(fName, SearchRec.Name),
                                           0, 'recursive not allowed')
                                  then
                                    Include(Status, ffCanceled);
                                end;
                          i := FindNext(SearchRec);
                        end;
                    i := abs(i);
                    if (i <> 0) and (i <> ERROR_NO_MORE_FILES) then
                      begin
                        Include(Status, ffRecursive);
                        if not FFcontinue(fName, 0,
                                          SysErrorMessage(GetLastError)
                                          + ' in searching ' + fName)
                        then
                          Include(Status, ffCanceled);
                      end;
                    FindClose(SearchRec);
                    Exclude(Status, ffFolder);
                    try
                      RmDir(fName);
                      { RemoveDir steht unter Delphi 1 nicht zur Verfuegung }
                    except
                      Include(Status, ffFolder);
                      if not FFcontinue(fName, 0, SysErrorMessage(GetLastError)
                                                  + ' in deleting ' + fName)
                      then
                        Include(Status, ffCanceled);
                    end;
                  end;
              end;
          end;
      end;
    DelFileOrFolder := Status;
  end;

function FileOrFolderItems(Items : sFileOrFolder) : String;
  var    i : tFileOrFolder;
         s : String;
  begin
    s := '';
    for i := low(tFileOrFolder) to high(tFileOrFolder) do
      if i in Items then
        case i of
          ffParam:
            s := s + ', Param';
          ffFile:
            s := s + ', File';
          ffFolder:
            s := s + ', Folder';
          ffNotFound:
            s := s + ', NotFound';
          ffReadOnly:
            s := s + ', ReadOnly';
          ffDifferent:
            s := s + ', Different';
          ffRecursive:
            s := s + ', Recursive';
          ffExists:
            s := s + ', Exists';
          ffForceDirs:
            s := s + ', ForceDirs';
          ffCanceled:
            s := s + ', Canceled';
          ffVolumeFull:
            s := s + ', VolumeFull';
        else
          s := s + ', ' + IntToStr(ord(i));
        end;
    FileOrFolderItems := Copy(s, 3, length(s));
  end;

{$ifndef fpc}
function Cosinus(Phase : Byte; Verschiebung : extended) : Byte;
  begin
    Cosinus :=
      round((cos(Phase * 8 * ArcTan(1) / MaxPalette + Verschiebung) + 1)
            * 127.5);
  end;

function PaletteSize(Entries : Word) : Word;
  var    Palette : tLogPalette;
  begin
    Palette.palNumEntries := 0;
    PaletteSize := sizeof(Palette)
                   - sizeof(Palette.PalPalEntry)
                   + Entries * sizeof(tPaletteEntry);
  end;

function CreateRegenbogen : pLogPalette;
  var    Color   : Word;
         Palette : pLogPalette;
  begin
    GetMem(Palette, PaletteSize(MaxPalette));
    if Palette <> Nil then
      begin
        Palette^.PalVersion    := $300;
        Palette^.PalNumEntries := MaxPalette;
        for Color := 0 to MaxPalette - 1 do
          begin
            {$R- PalPalEntry[0..0]}
            Palette^.PalPalEntry[Color].peRed   := Cosinus(Color, 1.2);
            Palette^.PalPalEntry[Color].peGreen := Cosinus(Color, 2.9);
            Palette^.PalPalEntry[Color].peBlue  := Cosinus(Color, 4.2);
            Palette^.PalPalEntry[Color].peFlags := pc_Reserved;
            {$R+}
          end;
      end;
    CreateRegenbogen := Palette;
  end;

procedure FreeRegenbogen(var Palette : pLogPalette);
  begin
    if Palette <> Nil then
      begin
        Freemem(Palette, PaletteSize(MaxPalette));
        Palette := Nil;
      end;
  end;

procedure AnimateRegenbogen(var Palette : pLogPalette; Handle : hPalette;
                            Offset : Integer);
  var     Color      : Word;
          NeuPalette : pLogPalette;
  begin
    if (Palette <> Nil) and (Handle <> 0) then
      begin
        GetMem(NeuPalette, PaletteSize(MaxPalette));
        if NeuPalette <> Nil then
          begin
            NeuPalette^ := Palette^;
            if Offset < 0 then
              inc(Offset, ((-Offset div MaxPalette) + 1) * MaxPalette);
            {$R-}
            for Color := 0 to MaxPalette - 1 do
              NeuPalette^.PalPalEntry[Color] :=
                Palette^.PalPalEntry[(Color + Offset) mod MaxPalette];
            {$R+}
            Freemem(Palette, PaletteSize(MaxPalette));
            Palette := NeuPalette;
            {$ifdef Win32}
              AnimatePalette(Handle, 0, MaxPalette, @Palette^.PalPalEntry[0]);
            {$else}
              AnimatePalette(Handle, 0, MaxPalette, Palette^.PalPalEntry[0]);
            {$endif}
          end;
      end;
  end;
{$endif}

function MessageName(m : Word) : String;
  begin
    case m of
      wm_Null:
        MessageName := 'wm_Null';
      wm_Create:
        MessageName := 'wm_Create';
      wm_Destroy:
        MessageName := 'wm_Destroy';
      wm_Move:
        MessageName := 'wm_Move';
      wm_Size:
        MessageName := 'wm_Size';
      wm_Activate:
        MessageName := 'wm_Activate';
      wm_SetFocus:
        MessageName := 'wm_SetFocus';
      wm_KillFocus:
        MessageName := 'wm_KillFocus';
      wm_Enable:
        MessageName := 'wm_Enable';
      wm_SetRedraw:
        MessageName := 'wm_SetRedraw';
      wm_SetText:
        MessageName := 'wm_SetText';
      wm_GetText:
        MessageName := 'wm_GetText';
      wm_GetTextLength:
        MessageName := 'wm_GetTextLength';
      wm_Paint:
        MessageName := 'wm_Paint';
      wm_Close:
        MessageName := 'wm_Close';
      wm_QueryEndSession:
        MessageName := 'wm_QueryEndSession';
      wm_Quit:
        MessageName := 'wm_Quit';
      wm_QueryOpen:
        MessageName := 'wm_QueryOpen';
      wm_EraseBkgnd:
        MessageName := 'wm_EraseBkgnd';
      wm_SysColorChange:
        MessageName := 'wm_SysColorChange';
      wm_EndSession:
        MessageName := 'wm_EndSession';
      {$ifndef fpc}
      wm_SystemError:
        MessageName := 'wm_SystemError';
      {$endif}
      wm_ShowWindow:
        MessageName := 'wm_ShowWindow';
      wm_CtlColor:
        MessageName := 'wm_CtlColor';
      wm_WinIniChange:
        MessageName := 'wm_WinIniChange';
      wm_DevModeChange:
        MessageName := 'wm_DevModeChange';
      wm_ActivateApp:
        MessageName := 'wm_ActivateApp';
      wm_FontChange:
        MessageName := 'wm_FontChange';
      wm_TimeChange:
        MessageName := 'wm_TimeChange';
      wm_CancelMode:
        MessageName := 'wm_CancelMode';
      wm_SetCursor:
        MessageName := 'wm_SetCursor';
      wm_MouseActivate:
        MessageName := 'wm_MouseActivate';
      wm_ChildActivate:
        MessageName := 'wm_ChildActivate';
      wm_QueueSync:
        MessageName := 'wm_QueueSync';
      wm_GetMinMaxInfo:
        MessageName := 'wm_GetMinMaxInfo';
      wm_PaintIcon:
        MessageName := 'wm_PaintIcon';
      wm_IconEraseBkgnd:
        MessageName := 'wm_IconEraseBkgnd';
      wm_NextDlgCtl:
        MessageName := 'wm_NextDlgCtl';
      wm_SpoolerStatus:
        MessageName := 'wm_SpoolerStatus';
      wm_DrawItem:
        MessageName := 'wm_DrawItem';
      wm_MeasureItem:
        MessageName := 'wm_MeasureItem';
      wm_DeleteItem:
        MessageName := 'wm_DeleteItem';
      wm_VKeyToItem:
        MessageName := 'wm_VKeyToItem';
      wm_CharToItem:
        MessageName := 'wm_CharToItem';
      wm_SetFont:
        MessageName := 'wm_SetFont';
      wm_GetFont:
        MessageName := 'wm_GetFont';
      wm_SetHotkey:
        MessageName := 'wm_SetHotkey';
      wm_GetHotkey:
        MessageName := 'wm_GetHotkey';
      wm_QueryDragIcon:
        MessageName := 'wm_QueryDragIcon';
      wm_CompareItem:
        MessageName := 'wm_CompareItem';
      {$ifndef fpc}
      wm_GetObject:
        MessageName := 'wm_GetObject';
      {$endif}
      wm_Compacting:
        MessageName := 'wm_Compacting';
      {$ifndef fpc}
      wm_CommNotify:
        MessageName := 'wm_CommNotify';
      {$endif}
      wm_WindowPosChanging:
        MessageName := 'wm_WindowPosChanging';
      wm_WindowPosChanged:
        MessageName := 'wm_WindowPosChanged';
      wm_Power:
        MessageName := 'wm_Power';
      wm_CopyData:
        MessageName := 'wm_CopyData';
      wm_CancelJournal:
        MessageName := 'wm_CancelJournal';
      wm_Notify:
        MessageName := 'wm_Notify';
      wm_InputLangChangeRequest:
        MessageName := 'wm_InputLangChangeRequest';
      wm_InputLangChange:
        MessageName := 'wm_InputLangChange';
      wm_TCard:
        MessageName := 'wm_TCard';
      wm_Help:
        MessageName := 'wm_Help';
      wm_UserChanged:
        MessageName := 'wm_UserChanged';
      wm_NotifyFormat:
        MessageName := 'wm_NotifyFormat';
      wm_ContextMenu:
        MessageName := 'wm_ContextMenu';
      wm_StyleChanging:
        MessageName := 'wm_StyleChanging';
      wm_StyleChanged:
        MessageName := 'wm_StyleChanged';
      wm_DisplayChange:
        MessageName := 'wm_DisplayChange';
      wm_GetIcon:
        MessageName := 'wm_GetIcon';
      wm_SetIcon:
        MessageName := 'wm_SetIcon';
      wm_NcCreate:
        MessageName := 'wm_NcCreate';
      wm_NcDestroy:
        MessageName := 'wm_NcDestroy';
      wm_NcCalcSize:
        MessageName := 'wm_NcCalcSize';
      wm_NcHitTest:
        MessageName := 'wm_NcHitTest';
      wm_NcPaint:
        MessageName := 'wm_NcPaint';
      wm_NcActivate:
        MessageName := 'wm_NcActivate';
      wm_GetDlgCode:
        MessageName := 'wm_GetDlgCode';
      wm_NcMouseMove:
        MessageName := 'wm_NcMouseMove';
      wm_NcLButtonDown:
        MessageName := 'wm_NcLButtonDown';
      wm_NcLButtonUp:
        MessageName := 'wm_NcLButtonUp';
      wm_NcLButtonDblClk:
        MessageName := 'wm_NcLButtondDlClk';
      wm_NcRButtonDown:
        MessageName := 'wm_NcRButtonDown';
      wm_NcRButtonUp:
        MessageName := 'wm_NcRButtonUp';
      wm_NcRButtonDblClk:
        MessageName := 'wm_NcRButtonDblClk';
      wm_NcMButtonDown:
        MessageName := 'wm_NcMButtonDown';
      wm_NcMButtonUp:
        MessageName := 'wm_NcMButtonUp';
      wm_NcMButtonDblClk:
        MessageName := 'wm_NcMButtonDblClk';
      wm_KeyDown:
        MessageName := 'wm_KeyDown';
      wm_KeyUp:
        MessageName := 'wm_KeyUp';
      wm_Char:
        MessageName := 'wm_Char';
      wm_DeadChar:
        MessageName := 'wm_DeadChar';
      wm_SysKeyDown:
        MessageName := 'wm_SysKeyDown';
      wm_SysKeyUp:
        MessageName := 'wm_SysKeyUp';
      wm_SysChar:
        MessageName := 'wm_SysChar';
      wm_SysDeadChar:
        MessageName := 'wm_SysDeadChar';
      wm_KeyLast:
        MessageName := 'wm_KeyLast';
      wm_InitDialog:
        MessageName := 'wm_InitDialog';
      wm_Command:
        MessageName := 'wm_Command';
      wm_SysCommand:
        MessageName := 'wm_SysCommand';
      wm_Timer:
        MessageName := 'wm_Timer';
      wm_HScroll:
        MessageName := 'wm_HScroll';
      wm_VScroll:
        MessageName := 'wm_VScroll';
      wm_InitMenu:
        MessageName := 'wm_InitMenu';
      wm_InitMenuPopup:
        MessageName := 'wm_InitMenuPopup';
      wm_MenuSelect:
        MessageName := 'wm_MenuSelect';
      wm_MenuChar:
        MessageName := 'wm_MenuChar';
      wm_EnterIdle:
        MessageName := 'wm_EnterIdle';
      {$ifndef fpc}
      wm_MenuRButtonUp:
        MessageName := 'wm_MenuRButtonUp';
      wm_MenuDrag:
        MessageName := 'wm_MenuDrag';
      wm_MenuGetObject:
        MessageName := 'wm_MenuGetObject';
      wm_UninitMenuPopup:
        MessageName := 'wm_UninitMenuPopup';
      wm_MenuCommand:
        MessageName := 'wm_MenuCommand';
      {$endif}
      wm_ChangeUiState:
        MessageName := 'wm_ChangeUiState';
      wm_UpdateUiState:
        MessageName := 'wm_UpdateUiState';
      wm_QueryUiState:
        MessageName := 'wm_QueryUiState';
      wm_CtlColorMsgbox:
        MessageName := 'wm_CtlColorMsgbox';
      wm_CtlColorEdit:
        MessageName := 'wm_CtlColorEdit';
      wm_CtlColorListbox:
        MessageName := 'wm_CtlColorListbox';
      wm_CtlColorBtn:
        MessageName := 'wm_CtlColorBtn';
      wm_CtlColorDlg:
        MessageName := 'wm_CtlColorDlg';
      wm_CtlColorScrollbar:
        MessageName := 'wm_CtlColorScrollbar';
      wm_CtlColorStatic:
        MessageName := 'wm_CtlColorStatic';
      wm_MouseMove:
        MessageName := 'wm_MouseMove';
      wm_LButtonDown:
        MessageName := 'wm_LButtonDown';
      wm_LButtonUp:
        MessageName := 'wm_LButtonUp';
      wm_LButtonDblClk:
        MessageName := 'wm_LButtonDblClk';
      wm_RButtonDown:
        MessageName := 'wm_RButtonDown';
      wm_RButtonUp:
        MessageName := 'wm_RButtonUp';
      wm_RButtonDblClk:
        MessageName := 'wm_RButtonDblClk';
      wm_MButtonDown:
        MessageName := 'wm_MButtonDown';
      wm_MButtonUp:
        MessageName := 'wm_MButtonUp';
      wm_MButtonDblClk:
        MessageName := 'wm_MButtonDblClk';
      wm_MouseWheel:
        MessageName := 'wm_MouseWheel';
      wm_ParentNotify:
        MessageName := 'wm_ParentNotify';
      wm_EnterMenuLoop:
        MessageName := 'wm_EnterMenuLoop';
      wm_ExitMenuLoop:
        MessageName := 'wm_ExitMenuLoop';
      {$ifndef fpc}
      wm_NextMenu:
        MessageName := 'wm_NextMenu';
      {$endif}
      wm_Sizing:
        MessageName := 'wm_Sizing';
      wm_CaptureChanged:
        MessageName := 'wm_CaptureChanged';
      wm_Moving:
        MessageName := 'wm_Moving';
      wm_PowerBroadcast:
        MessageName := 'wm_PowerBroadcast';
      wm_DeviceChange:
        MessageName := 'wm_DeviceChange';
      wm_Ime_StartComposition:
        MessageName := 'wm_Ime_StartComposition';
      wm_Ime_EndComposition:
        MessageName := 'wm_Ime_EndComposition';
      wm_Ime_Composition:
        MessageName := 'wm_Ime_Composition';
      wm_Ime_SetContext:
        MessageName := 'wm_Ime_SetContext';
      wm_Ime_Notify:
        MessageName := 'wm_Ime_Notify';
      wm_Ime_Control:
        MessageName := 'wm_Ime_Control';
      wm_Ime_CompositionFull:
        MessageName := 'wm_Ime_CompositionFull';
      wm_Ime_Select:
        MessageName := 'wm_Ime_Select';
      wm_Ime_Char:
        MessageName := 'wm_Ime_Char';
      {$ifndef fpc}
      wm_Ime_Request:
        MessageName := 'wm_Ime_Request';
      {$endif}
      wm_Ime_KeyDown:
        MessageName := 'wm_Ime_KeyDown';
      wm_Ime_KeyUp:
        MessageName := 'wm_Ime_KeyUp';
      wm_MdiCreate:
        MessageName := 'wm_MdiCreate';
      wm_MdiDestroy:
        MessageName := 'wm_MdiDestroy';
      wm_MdiActivate:
        MessageName := 'wm_MdiActivate';
      wm_MdiRestore:
        MessageName := 'wm_MdiRestore';
      wm_MdiNext:
        MessageName := 'wm_MdiNext';
      wm_MdiMaximize:
        MessageName := 'wm_MdiMaximize';
      wm_MdiTile:
        MessageName := 'wm_MdiTile';
      wm_MdiCascade:
        MessageName := 'wm_MdiCascade';
      wm_MdiIconArrange:
        MessageName := 'wm_MdiIconArrange';
      wm_MdiGetActive:
        MessageName := 'wm_MdiGetActive';
      wm_MdiSetMenu:
        MessageName := 'wm_MdiSetMenu';
      wm_EnterSizeMove:
        MessageName := 'wm_EnterSizeMove';
      wm_ExitSizeMove:
        MessageName := 'wm_ExitSizeMove';
      wm_DropFiles:
        MessageName := 'wm_DropFiles';
      wm_MdiRefreshMenu:
        MessageName := 'wm_MdiRefreshMenu';
      wm_MouseHover:
        MessageName := 'wm_MouseHover';
      wm_MouseLeave:
        MessageName := 'wm_MouseLeave';
      wm_Cut:
        MessageName := 'wm_Cut';
      wm_Copy:
        MessageName := 'wm_Copy';
      wm_Paste:
        MessageName := 'wm_Paste';
      wm_Clear:
        MessageName := 'wm_Clear';
      wm_Undo:
        MessageName := 'wm_Undo';
      wm_RenderFormat:
        MessageName := 'wm_RenderFormat';
      wm_RenderAllFormats:
        MessageName := 'wm_RenderAllFormats';
      wm_DestroyClipboard:
        MessageName := 'wm_DestroyClipboard';
      wm_DrawClipboard:
        MessageName := 'wm_DrawClipboard';
      wm_PaintClipboard:
        MessageName := 'wm_PaintClipboard';
      wm_VScrollClipboard:
        MessageName := 'wm_VScrollClipboard';
      wm_SizeClipboard:
        MessageName := 'wm_SizeClipboard';
      wm_AskCbFormatName:
        MessageName := 'wm_AskCbFormatName';
      wm_ChangeCbChain:
        MessageName := 'wm_ChangeCbChain';
      wm_HScrollClipboard:
        MessageName := 'wm_HScrollClipboard';
      wm_QueryNewPalette:
        MessageName := 'wm_QueryNewPalette';
      wm_PaletteIsChanging:
        MessageName := 'wm_PaletteIsChanging';
      wm_PaletteChanged:
        MessageName := 'wm_PaletteChanged';
      wm_Hotkey:
        MessageName := 'wm_Hotkey';
      wm_Print:
        MessageName := 'wm_Print';
      wm_PrintClient:
        MessageName := 'wm_PrintClient';
      {$ifndef fpc}
      wm_HandHeldFirst:
        MessageName := 'wm_HandHeldFirst';
      wm_HandHeldLast:
        MessageName := 'wm_HandHeldLast';
      {$endif}
      wm_PenWinFirst:
        MessageName := 'wm_PenWinFirst';
      wm_PenWinLast:
        MessageName := 'wm_PenWinLast';
      {$ifndef fpc}
      wm_Coalesce_first:
        MessageName := 'wm_Coalesce_first';
      wm_Coalesce_last:
        MessageName := 'wm_Coalesce_last';
      wm_DDE_Initiate:
        MessageName := 'wm_DDE_Initiate';
      wm_DDE_Terminate:
        MessageName := 'wm_DDE_Terminate';
      wm_DDE_Advise:
        MessageName := 'wm_DDE_Advise';
      wm_DDE_UnAdvise:
        MessageName := 'wm_DDE_UnAdvise';
      wm_DDE_Ack:
        MessageName := 'wm_DDE_Ack';
      wm_DDE_Data:
        MessageName := 'wm_DDE_Data';
      wm_DDE_Request:
        MessageName := 'wm_DDE_Request';
      wm_DDE_Poke:
        MessageName := 'wm_DDE_Poke';
      wm_DDE_Execute:
        MessageName := 'wm_DDE_Execute';
      {$endif}
      wm_App:
        MessageName := 'wm_App';
      wm_User:
        MessageName := 'wm_User';
    else
      MessageName := '$' + IntToHex(m, 4);
    end;
  end;

{$endif}

const fsNames : Array[TFontStyle] of String =
{$ifdef fpc}
  ('Bold', 'Italic', 'StrikeOut', 'Underline');
{$else}
  ('Bold', 'Italic', 'Underline', 'StrikeOut');
{$endif}

function FontStyleSet(FontStyle : String) : tFontStyles;
  var    Style : tFontStyles;
         s     : tFontStyle;
  begin
    Style := [];
    for s := Low(tFontStyle) to High(tFontStyle) do
      if pos(fsNames[s], FontStyle) > 0 then
        Include(Style, s);
    FontStyleSet := Style;
  end;

function FontStyleString(FontStyle : tFontStyles) : String;
  var    s : tFontStyle;
         t : String;
  begin
    t := '';
    for s := Low(tFontStyle) to High(tFontStyle) do
      if s in FontStyle then
        t := t + ',' + fsNames[s];
    FontStyleString := Copy(t, 2, length(t));
  end;

function ReplaceEscape(t : String) : String;
  var    s      : String;
         Escape : Boolean;
  begin
    s      := '';
    Escape := False;
    while t > '' do
      begin
        if t[1] = '\' then
          if Escape then
            begin
              s      := s + '\';
              Escape := False;
            end
          else
            Escape := True
        else
          if Escape then
            begin
              case t[1] of
                'n': s := s + A_LF;
                't': s := s + A_HT;
                'A': s := s + A_ISO_UPPER_AE;
                'O': s := s + A_ISO_UPPER_OE;
                'U': s := s + A_ISO_UPPER_UE;
                'a': s := s + A_ISO_LOWER_AE;
                'o': s := s + A_ISO_LOWER_OE;
                'u': s := s + A_ISO_LOWER_UE;
                's': s := s + A_ISO_SS;
                '\': s := s + '\';
              else
                s := s + '\' + t[1];
              end;
              Escape := False;
            end
          else
            s := s + t[1];
        t := Copy(t, 2, length(t));
      end;
    ReplaceEscape := s;
  end;

function Ordinal(i : LongInt) : String;
  var    s : String;
  begin
    str(i, s);
    case abs(i) mod 100 of
      11, 12, 13 : Ordinal := s + 'th';
    else
      case abs(i) mod 10 of
        1: Ordinal := s + 'st';
        2: Ordinal := s + 'nd';
        3: Ordinal := s + 'rd';
      else
        Ordinal := s + 'th';
      end;
    end;
  end;

function GetMyName : String;
  var    s, t : String;
  begin
    s         := ExtractFileName(Application.ExeName);
    t         := ExtractFileExt(s);
    GetMyName := copy(s, 1, length(s) - length(t));
  end;

function MakeShell(s : String) : String;
  { Windows verkraftet zwar Parameter
    z.B. 'a', '"a b"', 'a""b' und '" "',
    aber nicht '"a"'.
    Aus 'a""b'  wird 'a"b'.
    Statt '"a"' muss 'a' verwendet werden.
  }
  var t : String;
      p : Integer;
  begin
    t := '';
    p := Pos('"', s);
    while p > 0 do
      begin
        t := t + copy(s, 1, p - 1) + '""';
        s := Copy(s, p + 1, length(s));
        p := Pos('"', s);
      end;
    t := t + s;
    if t = '' then
      t := '""'
    else
      if pos(' ', t) > 0 then
        t := '"' + t + '"';
    MakeShell := t;
  end;

function ASCIIname(c : char) : String;
  begin
    case c of
      A_NUL: ASCIIname := 'NUL';
      A_SOH: ASCIIname := 'SOH';
      A_STX: ASCIIname := 'STX';
      A_ETX: ASCIIname := 'ETX';
      A_EOT: ASCIIname := 'EOT';
      A_ENQ: ASCIIname := 'ENQ';
      A_ACK: ASCIIname := 'ACK';
      A_BEL: ASCIIname := 'BEL';
      A_BS : ASCIIname := 'BS';
      A_HT : ASCIIname := 'HT';
      A_LF : ASCIIname := 'LF';
      A_VT : ASCIIname := 'VT';
      A_FF : ASCIIname := 'FF';
      A_CR : ASCIIname := 'CR';
      A_SO : ASCIIname := 'SO';
      A_SI : ASCIIname := 'SI';
      A_DLE: ASCIIname := 'DLE';
      A_DC1: ASCIIname := 'DC1=XON';
      A_DC2: ASCIIname := 'DC2';
      A_DC3: ASCIIname := 'DC3=XOF';
      A_DC4: ASCIIname := 'DC4';
      A_NAK: ASCIIname := 'NAK';
      A_SYN: ASCIIname := 'SYN';
      A_ETB: ASCIIname := 'ETB';
      A_CAN: ASCIIname := 'CAN';
      A_EM : ASCIIname := 'EM';
      A_SUB: ASCIIname := 'SUB';
      A_ESC: ASCIIname := 'ESC';
      A_FS : ASCIIname := 'FS';
      A_GS : ASCIIname := 'GS';
      A_RS : ASCIIname := 'RS';
      A_US : ASCIIname := 'US';
      A_SP : ASCIIname := 'SP';
      A_DEL: ASCIIname := 'DEL';
    else
      ASCIIName := '';
    end;
  end;

function MakePath(Path, FileName : String) : String;
  begin
    if copy(FileName, 1, 1) = PathDelim then
      FileName := copy(FileName, 2, length(FileName));
    if (Path = '') or (Copy(Path, length(Path), 1) = PathDelim) then
      MakePath := Path + FileName
    else
      MakePath := Path + PathDelim + FileName;
  end;

function MakeFind(Pfad : String) : String;
  begin
    if Copy(Pfad, length(Pfad), 1) <> PathDelim then
      MakeFind := Pfad
    else
      if Copy(Pfad, length(Pfad) - 1, 1) = ':' then
        MakeFind := MakePath(Pfad, Wildcard)
      else
        MakeFind := Copy(Pfad, 1, length(Pfad) - 1);
  end;

function IsIni(s : String) : Boolean;
  var    i  : Integer;
         ok : Boolean;
  begin
    Ok := True;
    i  := 1;
    while (i <= length(s)) and ok do
      begin
        if not (   ('A' <= s[i]) and (s[i] <= 'Z')
                or ('a' <= s[i]) and (s[i] <= 'z')
                or ('0' <= s[i]) and (s[i] <= '9')
                or (s[i] = '_')
                or (s[i] = ' ')) then
          Ok := False;
        inc(i);
      end;
    IsIni := Ok;
  end;

function ISOtoIBM(s : String) : String;
  begin
    s := StringReplace(s, A_ISO_UPPER_AE, A_IBM_UPPER_AE, [rfReplaceAll]);
    s := StringReplace(s, A_ISO_UPPER_OE, A_IBM_UPPER_OE, [rfReplaceAll]);
    s := StringReplace(s, A_ISO_UPPER_UE, A_IBM_UPPER_UE, [rfReplaceAll]);
    s := StringReplace(s, A_ISO_LOWER_AE, A_IBM_LOWER_AE, [rfReplaceAll]);
    s := StringReplace(s, A_ISO_LOWER_OE, A_IBM_LOWER_OE, [rfReplaceAll]);
    s := StringReplace(s, A_ISO_LOWER_UE, A_IBM_LOWER_UE, [rfReplaceAll]);
    s := StringReplace(s, A_ISO_SS,       A_IBM_SS,       [rfReplaceAll]);
    ISOtoIBM := s;
  end;

function IBMtoISO(s : String) : String;
  begin
    s := StringReplace(s, A_IBM_UPPER_AE, A_ISO_UPPER_AE, [rfReplaceAll]);
    s := StringReplace(s, A_IBM_UPPER_OE, A_ISO_UPPER_OE, [rfReplaceAll]);
    s := StringReplace(s, A_IBM_UPPER_UE, A_ISO_UPPER_UE, [rfReplaceAll]);
    s := StringReplace(s, A_IBM_LOWER_AE, A_ISO_LOWER_AE, [rfReplaceAll]);
    s := StringReplace(s, A_IBM_LOWER_OE, A_ISO_LOWER_OE, [rfReplaceAll]);
    s := StringReplace(s, A_IBM_LOWER_UE, A_ISO_LOWER_UE, [rfReplaceAll]);
    s := StringReplace(s, A_IBM_SS,       A_ISO_SS,       [rfReplaceAll]);
    IBMtoISO := s;
  end;

Function IntMin(a, b : LongInt) : LongInt;
  Begin
    If a < b Then
      IntMin := a
    else
      IntMin := b;
  End;

Function IntMax(a, b : LongInt) : LongInt;
  Begin
    If a > b Then
      IntMax := a
    else
      IntMax := b;
  End;

function GetTable(var s : String; var Col, Row : Integer) : String;
  var    ht,
         cr,
         lf,
         nl,
         tr : Integer;
  begin
    ht := Pos(A_HT, s);
    cr := Pos(A_CR, s);
    lf := Pos(A_LF, s);
    if cr = 0 then
      nl := lf
    else
      if lf = 0 then
        nl := cr
      else
        if cr < lf then
          nl := cr
        else
          nl := lf;
    if nl = 0 then
      begin
        tr := ht;
        inc(col);
      end
    else
      if ht = 0 then
        begin
          tr  := nl;
          col := 0;
          inc(row);
        end
      else
        if nl < ht then
          begin
            tr  := nl;
            col := 0;
            inc(row);
          end
        else
          begin
            tr := ht;
            inc(col);
          end;
    if tr = 0 then
      begin
        GetTable := s;
        s        := '';
      end
    else
      begin
        GetTable := Copy(s, 1,      tr - 1);
        s        := Copy(s, tr + 1, length(s));
        if tr = cr then
          if copy(s, 1, 1) = A_LF then
            s := Copy(s, 2, length(s));
      end;
  end;

function EscapeAmpersand(s : String) : String;
  var    t : String;
         p : Integer;
  begin
    t := '';
    while s > '' do
      begin
        p := Pos('&', s);
        if p > 0 then
          begin
            t := t + Copy(s, 1, p) + '&';
            s := Copy(s, p + 1, length(s));
          end
        else
          begin
            t := t + s;
            s := '';
          end;
      end;
    EscapeAmpersand := t;
  end;

function MsgBox(Text, Caption : String; Flags : LongInt) : Integer;
  var    TextBuffer,
         CaptionBuffer : PChar;
  begin
    TextBuffer    := StrAlloc(Length(Text)    + 1);
    CaptionBuffer := StrAlloc(Length(Caption) + 1);
    StrPCopy(TextBuffer,    Text);
    StrPCopy(CaptionBuffer, Caption);
    MsgBox := Application.MessageBox(TextBuffer, CaptionBuffer, Flags);
    StrDispose(TextBuffer);
    StrDispose(CaptionBuffer);
  end;

function ReadBmpHeader(var Datei : File; var Header : tBmpHeader) : Boolean;
  const  BM    : Array[0..1] of Char = 'BM';
  var    Count : Integer;
  begin
    ReadBmpHeader            := False;
    Header.FileHeader.bfSize := 0;
    Count                    := 0;
    BlockRead(Datei, Header, sizeof(Header), Count);
    if Count = sizeof(Header) then
      if    (Header.FileHeader.bfType = word(BM))
        and (Header.InfoHeader.biSize >= sizeof(Header.InfoHeader))
      then
        ReadBmpHeader := True;
  end;

procedure GetBmpDPI(FileName : String; var xDPI, yDPI : Integer);
  var     Header : tBmpHeader;
          Datei  : File;
          ok     : Boolean;
  begin
    Header.FileHeader.bfSize := 0;
    xDPI := 0;
    yDPI := 0;
    ok   := True;
    AssignFile(Datei, FileName);
    try
      Reset(Datei, 1);
    except
      ok := False;
    end;
    if ok then
      begin
        if ReadBmpHeader(Datei, Header) then
          begin
            xDPI := ((Header.InfoHeader.biXPelsPerMeter * 254)
                     div 1000 + 5) div 10;
            yDPI := ((Header.InfoHeader.biYPelsPerMeter * 254)
                     div 1000 + 5) div 10;
          end;
        CloseFile(Datei);
      end;
  end;

procedure SetBmpDPI(FileName : String; xDPI, yDPI : Integer);
  var     Header : tBmpHeader;
          Datei  : File;
          ok     : Boolean;
          Count  : Integer;
  begin
    Header.FileHeader.bfSize := 0;
    Count := 0;
    ok    := True;
    AssignFile(Datei, FileName);
    try
      Reset(Datei, 1);
    except
      ok := False;
    end;
    if ok then
      begin
        if ReadBmpHeader(Datei, Header) then
          begin
            Header.InfoHeader.biXPelsPerMeter :=
              ((longint(xDPI) * 100000) div 254 + 5) div 10;
            Header.InfoHeader.biYPelsPerMeter :=
              ((longint(yDPI) * 100000) div 254 + 5) div 10;
            Seek(Datei, 0);
            BlockWrite(Datei, Header, sizeof(Header), Count);
          end;
        CloseFile(Datei);
      end;
  end;

initialization;
  GetDir(0, StartDir);
  IniName := MakePath(StartDir, GetMyName + '.ini');
end.

